/* --------------------------------------------------------------------------
**	S3 86c964/S3 86c968 chipset driver
** --------------------------------------------------------------------------
**
**	Copyright (C) 1995-1997 Steffen Seeger
**	Copyright (C) 1998 Jan Kneschke
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
**
**	This code was derived from the following source of information:
**	[1]	S3 Vision964 Data Book, (c) 1994 S3 Incorporated
**      [2]	S3 Vision968 Multimedia Accelerator, (c) 1995 S3 Incorporated
**
** !!!	[1] states that DCLK must be greater than max(10 MHz, 0.5*SystemCLK),
** !!!	which might be up to 25 MHz on VL bus, to ensure valid writes to the
** !!!	clock chip. To allow for 320 pixel horizontal (ca. 400 pixel htotal) 
** !!!	resolution on a standard VGA monitor (25.175 kHz) we need DCLK down to
** !!!	10 MHz. My miroCrystal20SV didn't seem to have any problems so 
** !!!	chipset_info.dclk.min is set to 10 MHz. Hopefully all is well... 
*/



#include <ggi/maintainers.h>
#define	MAINTAINER	Jan_Kneschke
#define	DRIVER_NAME	"S3 86c964/86c968 chipset driver"
#define	DRIVER_REV	"$Revision: 1.10 $"

/*	WARNING: Don't set DEBUG_LEVEL higher than 0 unless you have a second
**	WARNING: display or klogd running to catch the output!
*/
#define	DEBUG_LEVEL		0
#define	HARDCODED_DEFAULTS	FALSE
#define	HARDWARE_TEXT_CURSOR	1

#include <ggi/system.h>
#include <ggi/module.h>

#define	S3_86C964
#define	S3_86C968
#include "S3/Vision.h"
#define	IO_VGA_BASE		0x03C0
#define	IO_VGA_SIZE		0x20
#define	IO_ACCEL_BASE		0x02E8
#define	IO_ACCEL_SIZE		4

#define TEXT_MEMORY_SIZE	(64 KB)

enum s3_sid_design
{
	SID_32S,
	SID_32P,
	SID_64S,
	SID_64P,
	SID_128S,
	SID_SPEA_MERCURY_P64,

	SID_LAST
};

/*	We keep the driver specific state per chipset handled in a
**	chipset_private struct. As s3_mmio.inc and s3_text16.inc rely on it,
**	it must be declared early.
*/
struct chipset_private
{
	/*	internal stuff
	*/
	mem_vaddr	mem_vga;
	mem_vaddr	mem_accel;	/* memory mapped accel ports	*/

	ggi_uint		frame;
	ggi_uint		split;
	ggi_uint		origin, org_xdot, org_ydot, org_offs;
	ggi_uint		fb_size;

	uint32		ptr_old, cur_old;
	uint16		ptr_and, cur_and;
	uint16		ptr_xor, cur_xor;

	/*	exported stuff
	*/
	struct kgi_mmio_region

	lram,		/* on-board RAM		little endian	*/
	lpdt,		/* pixel data transfer	little endian	*/
	lpfd,		/* pixel formatter data	little endian	*/
	lpfm,		/* pixel formatter data	little endian	*/
	
	bram,		/* on-board RAM		big endian	*/
	bpdt,		/* pixel data transfer	big endian	*/
	bpfd,		/* pixel formatter data	big endian	*/
	bpfm;		/* pixel formatter mask	big endian	*/

	/*	administrative stuff
	*/
	struct io_region	io_vga_region;
	struct io_region	io_accel_region;
	struct mem_region	mem_accel_region;
	struct mem_region	mem_law_region;

	struct kgi_chipset	chipset_info;

	enum s3_sid_design 	sid;
};

#define GET_PRIVATE \
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset)

#define	IO_VGA			(private->io_vga_region.baseptr-IO_VGA_BASE)
#define	IO_ACCEL		(private->io_accel_region.baseptr-IO_ACCEL_BASE)
#define	MEM_LAW			(private->mem_law_region.baseptr)

#define	MEM_ACCEL		(private->mem_accel)
#define	MEM_VGA			(private->mem_vga)

#define	IO_VGA_REGION		(private->io_vga_region)
#define	IO_ACCEL_REGION		(private->io_accel_region)
#define	MEM_ACCEL_REGION	(private->mem_accel_region)
#define	MEM_LAW_REGION		(private->mem_law_region)

#define	CHIPSET_INFO		(private->chipset_info)

#define	TEXTMODE_MMIO_REGIONS	NULL
#define	GRAPHMODE_MMIO_REGIONS	NULL


#define	NoSEQRegs	0x0B
#define	NoCRTRegs	0x80
#define	NoGRCRegs	0x09
#define	NoATCRegs	0x15
#include "s3_mmio.inc"
#include "s3_text16.inc"


static const struct kgi_chipset s3_96x_info =
{
	"S3 Incorporated",		/* manufacturer		*/
	"86c964 (Vision)",		/* model (detected)	*/
	GM_TEXT16 | /* GM_4BIT | */	/* supported modes	*/
	GM_8BIT | GM_15BIT |
	GM_16BIT| GM_32BIT,
	0,
	{ 2048, 2048 },			/* maximum resolution	*/

	0,				/* RAM size (detected)	*/
	{ 10000000, 95000000 }		/* DCLK limits		*/
};

/*	This table holds the default bits for text & graphics mode and
**	a copy of it will be filled in completely during mode_init().
*/
static struct statetable default_state =
{
	0x23, 0x00,					/* MISC, FCTRL	*/
	{
		0x03, 0x00, 0x0F, 0x20, 0x02, 0x00, 0x00, 0x00,	/* SEQ	*/
		0x06, 0x00, 0x00
	},
	{
		0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x10,	/* CRT	*/
		0x00, 0x40, 0x00, 0x1F, 0x00, 0x00, 0xFF, 0xFF,
		0x00, 0x00, 0x00, 0x00, 0x1F, 0x00, 0x00, 0xE3,	/* 0x1x	*/
		0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x2x	*/
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x01, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00,	/* 0x3x	*/
		0x48, 0xA5, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x4x	*/
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x38, 0x00, 0x00, 0x00,	/* 0x5x	*/
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x6x	*/
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x7x	*/
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	},
	{
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* GRC	*/
		0xFF
	},
	{
		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,	/* ATC	*/
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
		0x84, 0x00, 0x0F, 0x00, 0x00
	}
};

/*
**	ressources used by the S3 Vision 86c964
*/
static const struct io_region s3_io_vga =
{
	IO_NULL,
	IO_VGA_BASE,
	IO_VGA_SIZE,
	0xFFFF,
	"S3 VGA chipset"
};

static const struct io_region s3_io_accel =
{
	IO_NULL,
	IO_ACCEL_BASE,
	IO_ACCEL_SIZE,
	0x03FF,
	"S3 accelerator"
};

static const struct mem_region s3_mem_accel =
{
	MEM_NULL,
	0xA0000,
	0x10000,
	-1,
	"S3 accelerator"
};

static const struct mem_region s3_964_mem_law =
{
	MEM_NULL,
	0,
	0x00800000,
	-1,
	"S3 Vision 86c964 linear addressing window"
};

/*	ressources used by the S3 Vision 86c968.
**
**	After detection, we enable the new MMIO scheme and disable the
**	possibly active io regions. Thus only the MMIO regions are needed
**	for proper operation.
*/

static const struct mem_region s3_968_mem_law =
{
	MEM_NULL,
	0,
	0x04000000,
	-1,
	"S3 86c968 chipset LAW"
};

/*	s3_examine_chipset() does either read the chipset registers to collect
**	hardware dependent defaults (assuming the BIOS set them properly) or
**	uses hardcoded defaults.
*/
static void s3_examine_chipset(struct kgi_display *dpy)
{
	#define COPYDEFAULT(X) (default_state.CRT[X] = CRT_in8(dpy, X))
	register uint8 foo;
	const char	*chipname = KGIM_OPTIONS(dpy, chipset)->chipset;
	GET_PRIVATE;

	/*	copy default values
	*/
	#if HARDCODED_DEFAULT

		#error need default register values!
	#else
		COPYDEFAULT(CRT_CHIPID);
		COPYDEFAULT(CRT_CONFIG2);
		COPYDEFAULT(CRT_CONFIG3);
		COPYDEFAULT(CRT_MEMCONTROL3);
		COPYDEFAULT(CRT_MEMCONTROL4);
		COPYDEFAULT(CRT_MEMCONTROL5);

		default_state.CRT[0x3A] |= CRT_in8(dpy, 0x3A) &
			(CR3A_PCIMASK | CR3A_REFRESHMASK);

		default_state.CRT[0x40] |= CRT_in8(dpy,0x40) &
			CR40_BUSMASK;

		default_state.CRT[0x58] |= CRT_in8(dpy,0x58) &
			CR58_VL_T1_LATCH;

		default_state.CRT[0x66] |= CRT_in8(dpy, 0x66) &
			CR66_PCI_DISCONNECT;

		#if DEBUG_LEVEL > 0

			#define DEFAULT(x)	default_state.CRT[x] 
			DEBUG1("default values: %.2x  %.2x  %.2x  %.2x "\
				" %.2x  %.2x  %.2x  %.2x  %.2x", \
				DEFAULT(CRT_CONFIG2), \
				DEFAULT(CRT_CONFIG3), \
				DEFAULT(CRT_MEMCONTROL3), \
				DEFAULT(CRT_MEMCONTROL4), \
				DEFAULT(CRT_MEMCONTROL5), \
				DEFAULT(0x3A), \
				DEFAULT(0x40), \
				DEFAULT(0x58), \
				DEFAULT(0x66) \
			);

			#undef DEFAULT
		#endif
	#endif

	private->sid = SID_32P;
	if (chipname) {

#		define	SID_OPTION(x,y)		\
		if (strcmp(chipname, x) == 0) {	\
						\
			private->sid = y;	\
		}

		SID_OPTION("SID_32S", SID_32S);
		SID_OPTION("SID_32P", SID_32P);
		SID_OPTION("SID_64S", SID_64S);
		SID_OPTION("SID_64P", SID_64P);
		SID_OPTION("SID_128S", SID_128S);
		SID_OPTION("spea_mercury_p64", SID_SPEA_MERCURY_P64);
#		undef	SID_OPTION
	}

	/*	detect amount, type of installed memory and bus type
	*/
	foo = COPYDEFAULT(0x36);
	switch(foo & CR36_MEMSIZE_MASK) {

	case CR36_1MB:
		CHIPSET_INFO.ramsize = 1024 KB;
		MEM_LAW_REGION.size  = 1024 KB;
		default_state.CRT[0x58] |= CR58_LAW_1MB;
		break;

	case CR36_2MB:
		CHIPSET_INFO.ramsize = 2048 KB;
		MEM_LAW_REGION.size  = 2048 KB;
		default_state.CRT[0x58] |= CR58_LAW_2MB;
		break;

	case CR36_3MB:
		CHIPSET_INFO.ramsize = 3071 KB;
		MEM_LAW_REGION.size  = 8192 KB;
		default_state.CRT[0x58] |= CR58_LAW_8MB;
		break;

	case CR36_4MB:
		CHIPSET_INFO.ramsize = 4096 KB;
		MEM_LAW_REGION.size  = 8192 KB;
		default_state.CRT[0x58] |= CR58_LAW_8MB;
		break;

	case CR36_6MB:
		CHIPSET_INFO.ramsize = 6144 KB;
		MEM_LAW_REGION.size  = 8192 KB;
		default_state.CRT[0x58] |= CR58_LAW_8MB;
		break;

	case CR36_8MB:
		CHIPSET_INFO.ramsize = 8192 KB;
		MEM_LAW_REGION.size  = 8192 KB;
		default_state.CRT[0x58] |= CR58_LAW_8MB;
		break;

	default:
		CHIPSET_INFO.ramsize = 1024 KB;
		MEM_LAW_REGION.size  = 1024 KB;
		default_state.CRT[0x58] |= CR58_LAW_1MB;
		NOTICE("unknown VRAM size, assume 1024 KB.");
	}
	DEBUG1("%i KB frame buffer RAM detected.", CHIPSET_INFO.ramsize/1024);
	MEM_LAW_REGION.size = CHIPSET_INFO.ramsize;

	#if DEBUG_LEVEL > 0

		switch(foo & CR36_RAMMASK) {

		case CR36_FASTPAGE:
			NOTICE1("fast page mode RAM");
			break;

		case CR36_EDO:
			NOTICE1("extended data out RAM");
			break;

		case CR36_EDO_1CYC:
			NOTICE2("1cycle extended data out RAM");
			break;

		default:
			NOTICE("unknown RAM type");
		}

		switch(foo & CR36_BUSMASK) {

		case CR36_PCI:
			NOTICE2("PCI bus");
			break;

		case CR36_VESALOCAL:
			NOTICE2("VESA Local bus");
			break;

		default:
			NOTICE("unknown bus type");
		}
	#endif

	/*
	**	set miscellaneous stuff
	*/
	private->ptr_and = 0x00FF;
	private->ptr_xor = 0x6F00;
	private->cur_and = 0x00FF;
	private->cur_xor = 0x7F00;

	TRACE2(dump_state(&default_state));

	private->lram.next	= NULL;
	private->lram.name	= "video ram";
	private->lram.type	= MMAP_TYPE_MMIO|MMAP_PUBLIC | MMAP_FRAMEBUFFER;
	private->lram.prot	= MMIO_RDWR | MMIO_SHARED;
	private->lram.psize	= private->lram.size = CHIPSET_INFO.ramsize;
	private->lram.pstart	= MEM_LAW_REGION.base;

	#undef COPYDEFAULT
}


static int s3_verify_86c964(struct kgi_display *dpy)
{
	GET_PRIVATE;
	ggi_uint    foo;
	ggi_sint   err;
	uint16  cmd = 0;
	mem_addr base0 = KGIM_OPTIONS(dpy, misc)->law_base;
	struct kgim_options_pci *pci = KGIM_OPTIONS(dpy, pci);
	struct kgim_options_misc *misc = KGIM_OPTIONS(dpy, misc);

	DEBUG2("testing for S3 Vision 86c964, dev = %.8x", pci->dev);

	memset(private, 0, sizeof(*private));
	private->mem_vga   = MEM_NULL;
	private->mem_accel = MEM_NULL;

	IO_VGA_REGION      = s3_io_vga;
	IO_ACCEL_REGION    = s3_io_accel;
	MEM_ACCEL_REGION   = s3_mem_accel;
	MEM_LAW_REGION     = s3_964_mem_law;
	CHIPSET_INFO       = s3_96x_info;

	/*	PCI verification is easiest. However, we may have to enable it.
	*/
	if (pci->dev != PCICFG_NULL) {

		if ((pcicfg_in16(pci->dev+PCI_VENDOR_ID) != PCI_VENDOR_ID_S3) ||
			((pcicfg_in16(pci->dev+PCI_DEVICE_ID) & 0xFFF0) !=
			0x88D0)) {

			NOTICE2("  PCI device is no S3 80c964.");
			return -E(CHIPSET, UNKNOWN);
		}

		/*	enable PCI device
		*/
		cmd = pcicfg_in16(pci->dev + PCI_COMMAND);
		pcicfg_out16(PCI_COMMAND_IO | PCI_COMMAND_MEMORY,
			pci->dev + PCI_COMMAND);
	}

	/*	make sure noone else is serving this card.
	*/
	if (io_check_region(& IO_VGA_REGION) || 
		io_check_region(& IO_ACCEL_REGION)) {

		ERROR("io_accel or io_vga ressource busy.");
		return -E(CHIPSET, UNKNOWN);
	}
	io_claim_region(& IO_VGA_REGION);
	io_claim_region(& IO_ACCEL_REGION);

	/* NOTE	We assume the chip is woken up already, If not, add wakeup!
	*/
	s3_unlock(dpy);

	foo = CRT_in8(dpy, CRT_CHIPID);
	if ((foo & CR30_CHIPID) == CR30_964_ID) {

		/*	determine frame buffer window address
		*/
		base0 = (CRT_in8(dpy, CRT_LAW_POS_H) << 24) |
			(CRT_in8(dpy, CRT_LAW_POS_L) << 16);
		MEM_LAW_REGION.base = (misc->law_base
			? misc->law_base : base0) & ~(MEM_LAW_REGION.size - 1);

		if (! MEM_LAW_REGION.base ||
			mem_check_region(&MEM_LAW_REGION) ||
			mem_check_region(&MEM_ACCEL_REGION)) {

			ERROR("mem_law region busy");
			err = -EBUSY;

		} else {

			MEM_LAW = mem_claim_region(& MEM_LAW_REGION);
			MEM_ACCEL = mem_claim_region(&  MEM_ACCEL_REGION);
		
			default_state.CRT[0x53] |= CR53_ENB_MMIO;
			default_state.CRT[0x58] |= CR58_LAW_ENABLE;
			default_state.CRT[0x59] = (MEM_LAW_REGION.base >> 24) &
				0xFF;
			default_state.CRT[0x5A] = (MEM_LAW_REGION.base >> 16) &
				0xFF;

			s3_examine_chipset(dpy);

			err = EOK;
		}

	} else {

		err = -E(CHIPSET, UNKNOWN);
	}

	if (err) {

		s3_lock(dpy);

		if (pci->dev != PCICFG_NULL) {

			pcicfg_out16(cmd, pci->dev + PCI_COMMAND);
		}

		io_free_region(& IO_VGA_REGION);
		io_free_region(& IO_ACCEL_REGION);
		NOTICE2("no S3 80c964 detected");
		return -E(CHIPSET, UNKNOWN);
	}

	DEBUG1("%s rev %i detected.\n", CHIPSET_INFO.model, \
		CRT_in8(dpy, CRT_CHIPID));

	dpy->card->chipset = & CHIPSET_INFO;
	return EOK;
}


static int s3_verify_86c968(struct kgi_display *dpy)
{
	GET_PRIVATE;
	mem_addr base0 = 0;
	struct kgim_options_pci *pci = KGIM_OPTIONS(dpy, pci);
	struct kgim_options_misc *misc = KGIM_OPTIONS(dpy, misc);

	DEBUG2("testing for S3 Vision 86c968, dev = %.8x", pci->dev);

	memset(private, 0, sizeof(*private));
	private->mem_vga   = MEM_NULL;
	private->mem_accel = MEM_NULL;

	IO_VGA_REGION      = s3_io_vga;
	IO_ACCEL_REGION    = s3_io_accel;
	MEM_ACCEL_REGION   = s3_mem_accel;
	MEM_LAW_REGION     = s3_968_mem_law;
	CHIPSET_INFO       = s3_96x_info;
	CHIPSET_INFO.model[8] = '8';

	if (pci->dev != PCICFG_NULL) {

		uint16 cmd;
		#define IO_ENABLED      (cmd & PCI_COMMAND_IO)
		#define MEM_ENABLED     (cmd & PCI_COMMAND_MEMORY)
		#define ENABLED         (cmd & (PCI_COMMAND_IO | \
						PCI_COMMAND_MEMORY))

		if ((pcicfg_in16(pci->dev + PCI_VENDOR_ID) != 
			PCI_VENDOR_ID_S3) || 
			((pcicfg_in16(pci->dev + PCI_DEVICE_ID) & 0xFFF0) !=
			PCI_DEVICE_ID_S3_968)) {

			NOTICE1("  PCI device is not a Vision968");
			return -E(CHIPSET, UNKNOWN);                    
		}

		/*	We check now if the specified LAW window is available,
		**	thus we don't have to restore modifications in case
		**	it is not.
		*/
		cmd   = pcicfg_in16(pci->dev + PCI_COMMAND);
		base0 = pcicfg_in32(pci->dev + PCI_BASE_ADDRESS_0);

		MEM_LAW_REGION.base = (misc->law_base 
			? misc->law_base : base0) & 0xFF800000;
		MEM_LAW_REGION.size = 64 MB;
		if ((MEM_LAW_REGION.base == 0) ||
			mem_check_region(& MEM_LAW_REGION)) {

			ERROR("new mem_law region busy.");
			return -EBUSY;
		}

		/*	Ok, there _is_ a PCI Vision968. This one might still
		**	be sleeping, so we may have to wake it up. If it does
		**	not sleep, it might be the primary display or it is
		**	already served by another driver.
		*/
		if (ENABLED) {

			/* !!!	remember for restore ?
			**	was_primary = TRUE;
			*/

			/*	It is active, so it may be the primary VGA,
			**	but it may also be served already. So first 
			**	make sure noone else is serving it.
			*/
			MEM_LAW_REGION.base = base0;
			MEM_LAW_REGION.size = 4 KB;
			if (base0 && MEM_ENABLED && 
				mem_check_region(& MEM_LAW_REGION)) {

				ERROR("old mem_law region busy.");
				return -EBUSY;
			}
			
			if (IO_ENABLED && 
				(io_check_region(& IO_VGA_REGION) ||
				io_check_region(& IO_ACCEL_REGION))) {

				ERROR("io_vga or io_accel ressources busy.");
				return -EBUSY;
			}

			/*	All right, enable new MMIO.
			*/
			io_claim_region(& IO_VGA_REGION);
			io_claim_region(& IO_ACCEL_REGION);

			s3_unlock(dpy);

			CRT_out8(dpy, CRT_in8(dpy,0x53) | CR53_ENB_MMIO |
				CR53_NEW_MMIO, 0x53);
			CRT_out8(dpy, CRT_in8(dpy,0x58) | CR58_LAW_8MB | 
				CR58_LAW_ENABLE, 0x58);

			/*	Now that the 968 is responding to MMIO,
			**	disable I/O and release I/O regions.
			*/
			NOTICE1("I/O ports switched off!");
			SEQ_out8(dpy, SR09_MMIO_ONLY, 0x09);
			default_state.SEQ[0x09] |= SR09_MMIO_ONLY;

			io_free_region(& IO_ACCEL_REGION);
			io_free_region(& IO_VGA_REGION);

		} else {        /* if (ENABLED) */

			/*	It's there, but disabled.
			*/
			pcicfg_out16(PCI_COMMAND_MEMORY,
				pci->dev + PCI_COMMAND);
		}

		/*	We checked the MEM_LAW_REGION before, so now let's
		**	claim it.
		*/
		MEM_LAW_REGION.size = 64 MB;
		MEM_LAW_REGION.base = (misc->law_base
			? misc->law_base : base0) & ~(MEM_LAW_REGION.size - 1);
		mem_claim_region(& MEM_LAW_REGION);
		pcicfg_out32(MEM_LAW_REGION.base, pci->dev +
			PCI_BASE_ADDRESS_0);

		/*	Setting SR08 doesn't seem to be enough. So we just
		**	disable it via PCI configuration space (again).
		*/
		pcicfg_out16(PCI_COMMAND_MEMORY, pci->dev + PCI_COMMAND);

		MEM_VGA   = MEM_LAW + 0x1008000;
		MEM_ACCEL = MEM_LAW + 0x1000000;

		s3_unlock(dpy);

		/*	verify it again
		*/
		#if DEBUG_LEVEL > 0

			if ((CRT_in8(dpy, 0x30) != CR30_968ID) ||
				(CRT_in8(dpy, CRT_CHIPID_HIGH) != 0x88) ||
				(CRT_in8(dpy, CRT_CHIPID_LOW) != 0xF0)) {

				ERROR("!! 80c968 MMIO verification failed !!");
				MEM_LAW = mem_free_region(& MEM_LAW_REGION);

				io_free_region(& IO_ACCEL_REGION);
				io_free_region(& IO_VGA_REGION);
				return -E(CHIPSET, UNKNOWN);
			}
		#endif

		default_state.CRT[0x58] |= CR58_LAW_ENABLE | CR58_LAW_8MB;
		default_state.CRT[0x53] |= CR53_ENB_MMIO | CR53_NEW_MMIO;
		default_state.CRT[0x59] = (MEM_LAW_REGION.base >> 24) & 0xFF;
		default_state.CRT[0x5A] = (MEM_LAW_REGION.base >> 16) & 0xFF;

		s3_examine_chipset(dpy);

		return EOK;

	} else {	/* if (pci->dev != PCICFG_NULL)	*/

		uint8 CR53, CR58, CR59, CR5A;

		/*	We don't know if there actually is one, so we have to
		**	verify it is a Vision968.
		*/
		if (io_check_region(& IO_VGA_REGION) ||
			io_check_region(& IO_ACCEL_REGION)) {

			ERROR("io_vga or io_accel region busy.");
			return -EBUSY;
		}
		io_claim_region(& IO_ACCEL_REGION);
		io_claim_region(& IO_VGA_REGION);
		
		s3_unlock(dpy);

		if ((CRT_in8(dpy, 0x30) != CR30_968ID) ||
			(CRT_in8(dpy, CRT_CHIPID_HIGH) != 0x88) ||
			(CRT_in8(dpy, CRT_CHIPID_LOW) != 0xF0)) {

			NOTICE2("S3 Vision 80c968 I/O-verification failed.");
			io_free_region(& IO_ACCEL_REGION);
			io_free_region(& IO_VGA_REGION);
			return -E(CHIPSET, UNKNOWN);
		}

		/*	Bingo. Now make sure that we can operate it in new
		**	MMIO mode.
		*/
		CR53 = CRT_in8(dpy, 0x53);
		CR58 = CRT_in8(dpy, 0x58);
		CR59 = CRT_in8(dpy, 0x59);
		CR5A = CRT_in8(dpy, 0x5A);

		base0 = (CR59 << 24) | (CR5A << 16);
		MEM_LAW_REGION.size = 64 MB;
		MEM_LAW_REGION.base = (misc->law_base 
			? misc->law_base : base0) & ~(MEM_LAW_REGION.size - 1);
		if (! MEM_LAW_REGION.base ||
			mem_check_region(& MEM_LAW_REGION)) {

			ERROR("(new) mem_law region busy.");
			io_free_region(& IO_ACCEL_REGION);
			io_free_region(& IO_VGA_REGION);
			return -EBUSY;                 
		}

		default_state.CRT[0x53] |= CR53_ENB_MMIO | CR53_NEW_MMIO;
		default_state.CRT[0x58] |= CR58_LAW_ENABLE | CR58_LAW_8MB;
		default_state.CRT[0x59] = (MEM_LAW_REGION.base >> 24) & 0xFF;
		default_state.CRT[0x5A] = (MEM_LAW_REGION.base >> 16) & 0xFF;

		CRT_out8(dpy, default_state.CRT[0x59], 0x59);
		CRT_out8(dpy, default_state.CRT[0x5A], 0x5A);
		CRT_out8(dpy, CRT_in8(dpy, 0x58) | CR58_LAW_8MB |
			CR58_LAW_ENABLE, 0x58);
		CRT_out8(dpy, CRT_in8(dpy, 0x53) | CR53_ENB_MMIO |
			CR53_NEW_MMIO, 0x53);

		mem_claim_region(& MEM_LAW_REGION);
		MEM_VGA   = MEM_LAW + 0x1008000;
		MEM_ACCEL = MEM_LAW + 0x1000000;

		/*	It should be responding to new MMIO and I/O, so we
		**	release the io_regions and move to MMIO only mode.
		*/
		#if DEBUG_LEVEL > 0

			if ((CRT_in8(dpy, 0x30) != CR30_968ID) ||
				(CRT_in8(dpy, CRT_CHIPID_HIGH) != 0x88) ||
				(CRT_in8(dpy, CRT_CHIPID_LOW) != 0xF0)) {

				ERROR("!! 80c968 MMIO verification failed !!");
				MEM_LAW = mem_free_region(& MEM_LAW_REGION);

				CRT_out8(dpy, CR53, 0x53);
				CRT_out8(dpy, CR58, 0x58);
				CRT_out8(dpy, CR59, 0x59);
				CRT_out8(dpy, CR5A, 0x5A);
				
				io_free_region(& IO_ACCEL_REGION);
				io_free_region(& IO_VGA_REGION);
				return -E(CHIPSET, UNKNOWN);
			}
		#endif

		NOTICE1("I/O ports switched off!");
		SEQ_out8(dpy, SR09_MMIO_ONLY, 0x09);
		default_state.SEQ[0x09] |= SR09_MMIO_ONLY;

		/* ???	Disabling I/O using SR08 does not seem to work as
		** ???	intended. Thus we use 'internal ' PCICFG space.
		*/
		mem_out8(mem_in8(MEM_VGA + 4) & ~1, MEM_VGA + 4);

		io_free_region(& IO_ACCEL_REGION);
		io_free_region(& IO_VGA_REGION);
	}

	DEBUG2("%s rev %i detected.", CHIPSET_INFO.model, \
		CRT_in8(dpy, CRT_CHIP_REV));
	s3_examine_chipset(dpy);

	return EOK;
}

/*
**	KGI module interface
*/
int kgim_chipset_init(struct kgi_card *card, struct kgi_display *dpy)
{
	ggi_uint foo;
	struct chipset_private *private;

	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

	private = kgi_malloc_small(sizeof(*private), GFP_KERNEL);
	if (private == NULL) {

		NOTICE2("failed to allocate private data");
		return -ENOMEM;
	}

	memset(private, 0, sizeof(*private));
	KGIM_PRIVATE(dpy, chipset) = private;
	card->chipset = &CHIPSET_INFO;

	if (s3_verify_86c964(dpy) && s3_verify_86c968(dpy)) {

		ERROR("failed to detect Vision964 or Vision968.");
		kgi_free_small(private);
		KGIM_PRIVATE(dpy, chipset) = NULL;
		card->chipset = NULL;
		return -E(CHIPSET, UNKNOWN);
	}

	/*      Select CR42 to drive the clock select lines.
	*/
	foo = MISC_in8(dpy);
	if ((foo & MISC_CLOCK_MASK) != MISC_CLOCK_MASK) {
 
		CRT_out8(dpy, (foo & MISC_CLOCK_MASK) >> 2, 0x42);
		MISC_out8(dpy, (foo & ~MISC_CLOCK_MASK) | MISC_EXT_CLOCK);
	}

	return EOK;
}

void kgim_chipset_done(struct kgi_display *dpy)
{
	GET_PRIVATE;

	if (IO_VGA_REGION.baseptr) {

		io_free_region(& IO_VGA_REGION);
	}
	if (IO_ACCEL_REGION.baseptr) {

		io_free_region(& IO_ACCEL_REGION);
	}
	if (MEM_ACCEL_REGION.baseptr) {

		mem_free_region(& MEM_ACCEL_REGION);
	}
	if (MEM_LAW_REGION.baseptr) {

		mem_free_region(& MEM_LAW_REGION);
	}

	kgi_free_small(private);
	TRACE(KGIM_PRIVATE(dpy, chipset) = NULL);
	NOTICE(DRIVER_NAME " rev " DRIVER_REV " removed.");
}

static void adjust_timing(struct kgi_mode *mode)
{
	register int foo = (mode->request.graphtype == GT_TEXT16) 
				? mode->request.dpp.x : 8;

	#define ADJUST(X) mode->x.X -= (mode->x.X % foo)

	ADJUST(width);
	ADJUST(blankstart);
	ADJUST(syncstart);
	ADJUST(syncend);
	ADJUST(blankend);

	if (mode->x.total % foo) {

		mode->x.total += foo - (mode->x.total % foo);
	}

	if (((mode->x.syncend / foo) & 0x3F) ==
		((mode->x.syncstart / foo) & 0x3F)) {

		mode->x.syncend -= foo;
	}

	if (((mode->x.blankend / foo) & 0x7F) ==
		((mode->x.blankstart / foo) & 0x7F)) {

		mode->x.blankend -= foo;
	}

	if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F)) {

		mode->y.syncend--;
	}

	if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF)) {

		mode->y.blankend--;
	}

	#undef ADJUST
}

int kgim_chipset_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
			    enum kgi_tm_cmd cmd)
{
	GET_PRIVATE;
	int Bpp = 1;
	
	if (mode->flags & FL_MON_INTERLACE) {

		NOTICE1("interlaced modes not supported (yet).\n");
		return -E(CHIPSET, UNKNOWN); 
	}

	switch(mode->request.graphtype) {

	case GT_32BIT:  Bpp +=4;        /* (4+2+1+1)/2 = 4 Bpp  */
	case GT_15BIT:
	case GT_16BIT:  Bpp +=2;        /*  (2+1+1)/2 =   2 Bpp */
	case GT_8BIT:   Bpp +=1;        /*    (1+1)/2 =   1 Bpp */
	case GT_4BIT:                   /*        1/2 = 0.5 Bpp */
	case GT_TEXT16: break;

	default:
		DEBUG2("graphtype %i not supported", mode->request.graphtype);
		return -E(CHIPSET, UNKNOWN);
	}

	if (mode->lclk.mul != 1) {

		NOTICE1("invalid lclk multiplier");
		return -E(CHIPSET, UNKNOWN);
	}

	switch (mode->lclk.div) {

	case 1:
	case 2:
	case 4:
	case 8:
	case 16:
	case 32:
		break;

	default:
		DEBUG1("invalid CLK to LCLK divider %i", mode->lclk.div);
		return -E(CHIPSET, UNKNOWN);
	}

	#define	XVIRTUAL	(mode->request.virt.x)
	#define	YVIRTUAL	(mode->request.virt.y)
	#define	XVISIBLE	(mode->request.visible.x)
	#define	YVISIBLE	(mode->request.visible.y)
	#define	XDPP		(mode->request.dpp.x)
	#define	YDPP		(mode->request.dpp.y)

	switch (cmd) {

	case CMD_PROPOSE:

		if (mode->request.graphtype == GT_TEXT16) {
		
			int textx = mode->request.dpp.x;
			int texty = mode->request.dpp.y;

			mode->magnify.x = mode->magnify.y = 1;
			if ((textx < 7) || (9 < textx)) {

				DEBUG1("invalid char width %i", textx);
				return -EINVAL;
			}
			
			if (XVISIBLE == GGI_AUTO)	XVISIBLE = 80;
			if (YVISIBLE == GGI_AUTO)	YVISIBLE = 25;
			if (XVIRTUAL == GGI_AUTO)	XVIRTUAL = XVISIBLE;
			if (YVIRTUAL == GGI_AUTO)	YVIRTUAL = YVISIBLE;
			if (XDPP     == GGI_AUTO)	XDPP     =  8;
			if (YDPP     == GGI_AUTO)	YDPP     = 16;

			if (!(XVIRTUAL || YVIRTUAL)) {

				NOTICE1("need x or y virtual!");
				return -EINVAL;
			}

			if (!XVIRTUAL) {

				XVIRTUAL = TEXT_MEMORY_SIZE / YVIRTUAL;
				DEBUG1("setting xvirtual = %i", XVIRTUAL);
			}

			if (!YVIRTUAL) {

				YVIRTUAL = TEXT_MEMORY_SIZE / XVIRTUAL;
				DEBUG1("setting yvirtual = %i", YVIRTUAL);
			}

			if ((XVIRTUAL * YVIRTUAL) > TEXT_MEMORY_SIZE) {

				NOTICE1("not enough text memory");
				return -ENOMEM;
			} 
			
			mode->x.width = mode->request.visible.x * textx;
			mode->y.width = mode->request.visible.y * texty;

		} else {        /* mode->request.graphtype == GT_TEXT16 */

			mode->magnify.x = mode->magnify.y = 1;
			
			if (XVISIBLE == GGI_AUTO)	XVISIBLE = 640;
			if (YVISIBLE == GGI_AUTO)	YVISIBLE = 400;
			if (XVIRTUAL == GGI_AUTO)	XVIRTUAL = XVISIBLE;
			if (YVIRTUAL == GGI_AUTO)	YVIRTUAL = YVISIBLE;
			if (XDPP     == GGI_AUTO)	XDPP     =   1;
			if (YDPP     == GGI_AUTO)	YDPP     =   1;
			
			if      (XVIRTUAL <=  640) XVIRTUAL =  640;
			else if (XVIRTUAL <=  800) XVIRTUAL =  800;
			else if (XVIRTUAL <= 1024) XVIRTUAL = 1024;
			else if (XVIRTUAL <= 1152) XVIRTUAL = 1152;
			else if (XVIRTUAL <= 1280) XVIRTUAL = 1280;
			else if (XVIRTUAL <= 1600) XVIRTUAL = 1600;
			else if (XVIRTUAL <= 2048) XVIRTUAL = 2048;
			else {

				DEBUG2("%i hpixels are too many", XVIRTUAL);
				return -E(CHIPSET, UNKNOWN);
			}
	
			if ((XVIRTUAL * YVIRTUAL * Bpp / 2) >
				CHIPSET_INFO.ramsize) {

				DEBUG2("not enough memory for %ix%i@%ibpp",\
					XVIRTUAL, YVIRTUAL, Bpp * 4);
				return -ENOMEM;
			}

			mode->request.visible.x &= 0x7FFFFFF8;
			mode->request.visible.y &= 0x7FFFFFF8;
			
			mode->x.width = mode->request.visible.x * mode->magnify.x;
			mode->y.width = mode->request.visible.y * mode->magnify.y;
		}



		return EOK;

	case CMD_LOWER:

		if (mode->dclk < CHIPSET_INFO.dclk.min) {

			DEBUG2("%i Hz DCLK are too low", mode->dclk);
			return -E(CHIPSET, UNKNOWN);
		}

		mode->dclk = (mode->dclk > CHIPSET_INFO.dclk.max)
			? CHIPSET_INFO.dclk.max : mode->dclk;

		//adjust_timing(mode);
		return EOK;

	case CMD_RAISE:

		if (mode->dclk > CHIPSET_INFO.dclk.max) {

			DEBUG2("%i Hz DCLK are too high", mode->dclk);
			return -E(CHIPSET, UNKNOWN);
		}

		mode->dclk = (mode->dclk < CHIPSET_INFO.dclk.min)
			? CHIPSET_INFO.dclk.min : mode->dclk;

		//adjust_timing(mode);
		return EOK;

	case CMD_CHECK:

		if (mode->request.graphtype == GT_TEXT16) {

			if ((mode->magnify.x != 1) || (mode->magnify.y != 1) /* driver*/ 
			    || (mode->request.dpp.x < 7) /* hardware */
			    || (9 < mode->request.dpp.x) /* hardware */
			    || (mode->request.dpp.y < 1) /* hardware */
			    || (mode->request.dpp.y > 32) /* hardware */
			    || ((XVIRTUAL * YVIRTUAL) > TEXT_MEMORY_SIZE)     /* hardware */
			    || (mode->dclk > CHIPSET_INFO.dclk.max) /* hardware */
			    || (mode->dclk < CHIPSET_INFO.dclk.min)) /*hardware*/ {

				return -E(CHIPSET, UNKNOWN);
			}

			mode->flags |= FL_CHIP_COLORTEXT;
			mode->fb      = MEM_LAW;
			mode->fb_size = CHIPSET_INFO.ramsize;
			mode->fsize   = (8 * mode->request.virt.x * mode->request.virt.y);
			mode->io  = &text16_color_io;
			mode->mem = &text16_color_mem;
			mode->mmio = TEXTMODE_MMIO_REGIONS;

		} else {

			switch (XVIRTUAL) {

			case 640:
			case 800:
			case 1024:
			case 1152:
			case 1280:
			case 1600:
			case 2048:
				break;

			default:
				return -E(CHIPSET, UNKNOWN);
			}

			if (((XVIRTUAL * YVIRTUAL * Bpp / 2) > 
				CHIPSET_INFO.ramsize) ||
				(mode->magnify.x != 1)/*!!! temp */ ||
				(mode->magnify.y != 1)/*!!! temp */ ||
				(mode->clk.mul != 1) || (mode->clk.div != 1) ||
				(mode->dclk < CHIPSET_INFO.dclk.min) ||
				(CHIPSET_INFO.dclk.max < mode->dclk)) {

				return -E(CHIPSET, UNKNOWN);
			}

			mode->fb      = MEM_LAW;
			mode->fb_size = CHIPSET_INFO.ramsize;
			mode->fsize   = mode->request.virt.x *
				mode->request.virt.y * Bpp / 2;
			mode->io  = NULL;
			mode->mem = NULL;
			mode->mmio = GRAPHMODE_MMIO_REGIONS;
		}

		#define TM(X) (((unsigned) CHIPSET_INFO.maxdots.x < mode->x.X)\
			|| ((unsigned) CHIPSET_INFO.maxdots.y<mode->y.X))

		if (TM(width) || TM(blankstart) || TM(syncstart) ||
			TM(syncend) || TM(blankend) || TM(total)) {

			return -E(CHIPSET, UNKNOWN);
		}

		#undef TM

		//adjust_timing(mode);
		mode->mmio = & (private->lram);
		return EOK;

	default:
		INTERNALERROR;
		return -E(CHIPSET, UNKNOWN);
	}

	#undef XVIRTUAL 
	#undef YVIRTUAL
}

void kgim_chipset_set_mode (struct kgi_display *dpy, struct kgi_mode *mode)
{
	IS_CRITICAL;
	GET_PRIVATE;
	struct statetable setup = default_state;
	struct kgi_ramdac *dac;
	register int foo;
	int DCLKdiv, HSYNCadjust;

	DEBUG1("setting mode %i, %ix%i.%ix%i.%ix%i", \
		mode->request.graphtype,\
		mode->request.visible.x, mode->request.visible.y, \
		mode->request.virt.x, mode->request.virt.y, \
		mode->request.dpp.x, mode->request.dpp.y); 

	/*	Setup bits per pixel and logical screen width
	*/

	switch(mode->request.graphtype) {

	case GT_TEXT16:
		foo = 2 * mode->request.virt.x;
		foo /= 4; /* using doubleword mode */ 
		break;

	case GT_4BIT:
		setup.CRT[0x50] |= CR50_4BPP;
		foo = mode->request.virt.x >> 4;
		break;

	case GT_8BIT:
		setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
		setup.CRT[0x50] |= CR50_8BPP;
		foo = mode->request.virt.x >> 3;
		break;

	case GT_15BIT:
	case GT_16BIT:
		setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
		setup.CRT[0x50] |= CR50_16BPP;
//		setup.CRT[0x67] |= CR67_NO_INVERT_VCLK;
		foo = mode->request.virt.x >> 2;
		break;

	case GT_32BIT:
		setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
		setup.CRT[0x50] |= CR50_32BPP;
		foo = mode->request.virt.x >> 1;
		break;

	default:
		INTERNALERROR;
		return;
	}

	setup.CRT[CRT_LOGICALWIDTH]  |= foo & 0xFF;
	setup.CRT[0x51] |= (foo & 0x100) ? CR51_LOGICALWIDTH_B8 : 0;
	setup.CRT[0x51] |= (foo & 0x200) ? CR51_LOGICALWIDTH_B9 : 0;

	if(mode->request.graphtype == GT_TEXT16) {

		setup.CRT[0x17] = CR17_CGA_ADDR_WRAP | CR17_ENABLE_SYNC |
				CR17_CGA_BANKING | CR17_HGC_BANKING;

		setup.CRT[0x31] |= CR31_HISPEEDTXT;

		setup.SEQ[0x02] = SR02_PLANE0 | SR02_PLANE1;

		setup.GRC[0x05] |= GR5_ODD_EVEN;
		setup.GRC[0x06] |= GR6_MAP_B8_32 | GR6_CHAIN_OE;

		/* ???  can we leave this enabled?
		** ???  Miro20SV, V7 Mercury P-64 can. 
		** ???
		** ???  setup.CRT[0x34] &= ~CR34_ENB_FIFO; 
		*/
	
		switch((DCLKdiv = mode->request.dpp.x)) {

		case 7:
			setup.CRT[0x32] |= CR32_7DOT_CHCLK;
			break;

		case 8:
			setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
			break;

		case 9:
			break;

		default:
			INTERNALERROR;
			return;
		}

		setup.CRT[0x09] |= mode->request.dpp.y - 1;
		setup.ATC[0x13] = (mode->request.dpp.x == 9) ? 0x08 : 0x00;

		HSYNCadjust = 0;

	} else {        /* mode->request.graphtype == GT_TEXT16 */

		DCLKdiv = 8;
		HSYNCadjust = -2;

		setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
		setup.SEQ[0x02] |= SR02_PLANE_MASK;	/* ???  needed ? */ 
		setup.ATC[0x10] |= AR10_GRAPHICS;
		setup.GRC[0x05] |= GR5_8BIT_COLOR;	/* ???  needed ? */
		setup.GRC[0x06] |= GR6_GRAPHMODE | GR6_MAP_A0_64 /*??? needed? */;
		setup.CRT[0x31] |= CR31_ENHANCED;
		setup.CRT[0x33] |= CR33_NOBORDER;
		setup.CRT[0x58] |= CR58_LAW_ENABLE;

		setup.CRT[0x09] |= 0;
			/* !!! (TM->ymagnify - 1) & CR09_CHARHEIGHT_MASK */

		setup.CRT[0x09] |= (mode->flags & FL_CHIP_DOUBLESCAN) 
					? CR09_DOUBLESCAN : 0;

		switch(mode->request.virt.x) { 

		case  640:      setup.CRT[0x50] |= CR50_640;    break;
		case  800:      setup.CRT[0x50] |= CR50_800;    break;
		case 1024:      setup.CRT[0x50] |= CR50_1024;   break;
		case 1152:      setup.CRT[0x50] |= CR50_1152;   break;
		case 1280:      setup.CRT[0x50] |= CR50_1280;   break;
		case 1600:      setup.CRT[0x50] |= CR50_1600;   break;

		case 2048:
			setup.CRT[0x50] |= CR50_2048;
			setup.CRT[0x31] |= CR31_2PAGE_SCRN; 
			break;

		default:
			INTERNALERROR;
			return;
		}

		dac = dpy->card->ramdac;
		if (dac->dactype & DAC_VRAM) {

			setup.CRT[0x66] |= CR66_PIXELBUS_OFF;

			switch (private->sid) {

			case SID_32S:
				setup.CRT[0x66] |= CR66_32S_SID;
				setup.CRT[0x58] |= CR58_SAM_256;
				setup.CRT[0x67] &= ~CR67_NO_INVERT_VCLK;
				break;

			case SID_64S:
				setup.CRT[0x66] |= CR66_64S_SID;
				setup.CRT[0x58] |= CR58_SAM_256;
				break;

			case SID_SPEA_MERCURY_P64:
				setup.CRT[0x65] |= CR65_ADJ_512SAM;
				setup.CRT[0x66] 
				    |= ((1 << mode->request.graphtype) 
					& (GM_15BIT | GM_16BIT | GM_32BIT))
					? CR66_32P_SID : CR66_64S_SID;
				break;

			case SID_32P:
				setup.CRT[0x58] |= CR58_SAM_256;
				setup.CRT[0x66] |= CR66_32P_SID;
				break;
   
			case SID_128S:
				setup.CRT[0x65]
				    |= (setup.CRT[0x58] & CR58_SAM_256) 
					? CR65_ADJ_256SAM
					: CR65_ADJ_512SAM;
				setup.CRT[0x66] |= CR66_128S_SID;
				NOT_TESTED;
				break;

			default:
				INTERNAL_ERROR;
			}

		} else {	/* if (dac->dactype & DAC_VRAM)	*/ 

			/*	strange: a 964 with a RAMDAC that doesn't
			**	support VRAM SCLK...
			*/
			ERROR("DAC without VRAM shift clock support");
			INTERNALERROR;
			return;
		}
	}


	/*
	**	setup horizontal timing
	*/

	/*	horizontal display end
	*/
	foo = (mode->x.width / DCLKdiv) - 1;
	DEBUG2("horizontal disp end: %i",foo);
	setup.CRT[0x01] |= foo & 0xFF;
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HDISPLAYEND_B8 : 0;

	/*	horizontal blank start
	*/
	foo = (mode->x.blankstart / DCLKdiv);
	DEBUG2("horizontal blank start: %i",foo);
	setup.CRT[0x02] |= foo & 0xFF;
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HBLANKSTART_B8 : 0;

	/*	horizontal sync start
	*/
	foo = (mode->x.syncstart / DCLKdiv) + HSYNCadjust;
	DEBUG2("horizontal sync start: %i",foo);
	setup.CRT[0x04] |= foo & 0xFF;
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HSYNCSTART_B8 : 0;

	/*	horizontal sync end
	*/
	foo = (mode->x.syncend / DCLKdiv) + HSYNCadjust;
	DEBUG2("horizontal sync end: %i",foo);
	setup.CRT[0x5D] |= ((foo- (mode->x.syncstart/DCLKdiv)) > 0x20)
		? CR5D_HSYNCEND_B5 : 0;

	/*	horizontal blank end
	*/
	foo = (mode->x.blankend / DCLKdiv) - 1;
	DEBUG2("horizontal blank end: %i",foo);
	setup.CRT[0x03] |= foo & CR03_HBE_MASK;
	setup.CRT[0x05] |= (foo & 0x20) ? CR05_HBLANKEND_B5 : 0;
	setup.CRT[0x5D] |= ((foo - (mode->x.blankstart/DCLKdiv)) > 0x40)
		? CR5D_HBLANKEND_B6 : 0;
  
	/*	horizontal total
	*/
	foo = (mode->x.total / DCLKdiv) - 5;
	DEBUG2("horizontal total: %i",foo);
	setup.CRT[0x00] |= foo & 0xFF;
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HTOTAL_B8 : 0;


	/*
	**	setup vertical timing
	*/

	/*	vertical display end
	*/
	foo = mode->y.width - 1;
	DEBUG2("vertical disp end: %i",foo);
	setup.CRT[0x12] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VDISPLAYEND_B8  : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VDISPLAYEND_B9  : 0;
	setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VDISPLAYEND_B10 : 0;


	/*	vertical blank start
	*/
	foo = mode->y.blankstart - 1;
	DEBUG2("vertical blank start: %i",foo);
	setup.CRT[0x15] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VBLANKSTART_B8  : 0;
	setup.CRT[0x09] |= (foo & 0x200) ? CR09_VBLANKSTART_B9  : 0;
	setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VBLANKSTART_B10 : 0;

	/*	vertical sync start
	*/    
	foo = mode->y.syncstart;
	DEBUG2("vertical sync start: %i",foo);
	setup.CRT[0x10] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VSYNCSTART_B8  : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VSYNCSTART_B9  : 0;
	setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VSYNCSTART_B10 : 0;

	/*	vertical sync end
	*/
	foo = mode->y.syncend;
	DEBUG2("vertical sync end: %i",foo);
	setup.CRT[0x11] |= foo & CR11_VSYNCEND_MASK;
 
	/*	vertical blank end
	*/
	foo = mode->y.blankend - 1;
	DEBUG2("vertical blank end: %i",foo);
	setup.CRT[0x16] |= foo & 0xFF;
  
	/*	vertical total
	*/
	foo = mode->y.total - 2;
	DEBUG2("vertical total: %i",foo);
	setup.CRT[0x06] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VTOTAL_B8  : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VTOTAL_B9  : 0;
	setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VTOTAL_B10 : 0;


	/* !!! FIXME! */
	setup.CRT[CRT_INTERLACESTART]|= mode->y.total >> 1;
	if(mode->flags & FL_MON_INTERLACE) {

		NOTICE1("no interlaced modes supported (yet).\n");
		setup.CRT[0x42] |= CR42_INTERLACED;
		INTERNALERROR;
		return;
	}


	setup.MISC |= (mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC;
	setup.MISC |= (mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC;


	for (foo = 0; !((1 << foo) & mode->lclk.div); foo++) {
	}
	setup.CRT[0x66] |= foo & CR66_VCLK_DIV_MASK;


	/*	Setup the FIFO control parameters.
	*/
	foo = (mode->x.total + mode->x.syncstart)/2 / DCLKdiv;
	setup.CRT[0x3B] |= foo & 0xFF;
	setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_FIFOSTART_B8 : 0;

	TRACE2(dump_state(&setup));
	/*	Ok. We'll do the programming now.
	*/
	ENTER_CRITICAL;

		/*	wait for the accelerator to finish and disable it.
		*/
		if (dpy->mode.request.graphtype != GT_TEXT16) {

			foo = 10000;
			while ((ACCEL_in16(dpy, ECR_GP_STATUS) & 0xF2FF) &&
				foo) {

				foo--;
			}
			AC_out16(dpy, 0, AC_ADVFUNC_CNTL);
		}

		/*	switch sync signals off 
		*/
		CRT_out8(dpy, CRT_in8(dpy, 0x56) | CR56_SYNC_OFF, 0x56);

		/*	prepare subsystems
		*/
		s3_unlock(dpy);
		SEQ_out8(dpy, SEQ_in8(dpy, 0x01) | SR01_DISPLAY_OFF, 0x01);

		NOTICE2("setting up hardware");

		kgim_monitor_set_mode(dpy, mode);
		NOTICE2("monitor set");
		kgim_ramdac_set_mode(dpy, mode);
		NOTICE2("ramdac set");
		kgim_clock_set_mode(dpy, mode);
		NOTICE2("clock set");
		s3_unlock(dpy);
		restore_state(dpy, &setup);
		NOTICE2("chipset set");

		#ifdef MERCURY_P64              /* only SPEA knows why... */

			CRT_out8(dpy, CRT_in8(dpy, 0x42) | 4, 0x42);
		#endif

	LEAVE_CRITICAL;

	memcpy(&(dpy->mode), mode, sizeof(dpy->mode));

	NOTICE2("setting up display...");

	private->split = -1;
	private->org_xdot = private->org_ydot = private->org_offs = -1;

	if (mode->request.graphtype == GT_TEXT16) {

		dpy->set_font   = s3_text_set_font;
		dpy->set_split  = s3_set_split;
		dpy->set_frame  = s3_set_frame;
		dpy->set_origin = s3_text_set_origin;

		/*      provide hardware cursor in text mode
		*/
		if (HARDWARE_TEXT_CURSOR) {

			dpy->show_cursor = s3_show_hw_text_cursor;
			dpy->hide_cursor = s3_hide_hw_text_cursor;
			dpy->undo_cursor = NULL;

		} else {

			dpy->show_cursor = s3_show_sw_text_cursor;
			dpy->hide_cursor = s3_hide_sw_text_cursor;
			dpy->undo_cursor = s3_undo_sw_text_cursor;
		}

		/*      provide software cursor too in case DAC didn't already.
		*/
		if (! dpy->show_pointer) {

			dpy->show_pointer = s3_show_sw_text_pointer;
			dpy->hide_pointer = s3_hide_sw_text_pointer;
			dpy->undo_pointer = s3_undo_sw_text_pointer;
		}

		dpy->fb.size.x = mode->request.virt.x;
		dpy->fb.size.y = mode->request.virt.y;
		dpy->fb.bpp    = MAP_16_BPP;
		dpy->fb.pbuf   = MEM_LAW;
		dpy->fb.zbuf   = NULL;
		dpy->fb.ops    = &s3_text16;

		private->fb_size = dpy->fb.size.x * dpy->fb.size.y;

	} else {

		dpy->fb.size.x	= mode->request.virt.x;
		dpy->fb.size.y	= mode->request.virt.y;
		#warning set dpy->fb.bpp!
		dpy->fb.pbuf	= MEM_LAW;
		dpy->fb.zbuf	= NULL;
		dpy->fb.ops	= NULL;

		private->fb_size = dpy->fb.size.x * dpy->fb.size.y;
		
		dpy->show_pointer	= NULL;
		dpy->hide_pointer	= NULL;
		dpy->undo_pointer	= NULL;

		dpy->show_cursor	= NULL;
		dpy->hide_cursor	= NULL;
		dpy->undo_cursor	= NULL;

		dpy->set_font		= NULL;
		dpy->set_split		= s3_set_split;
		dpy->set_frame		= s3_set_frame;
		dpy->set_origin 	= s3_set_origin;

#if 1
		kgim_graphics_set_mode(dpy, mode);
		NOTICE2("graphics set");

#else
		/* !!! temporary, this will be moved to an extra source file.
		*/

		/* Reset the Accelerator Engine */
		AC_out16(dpy, ((mode->request.graphtype == GT_4BIT) 
			? EC01_4BPP : EC01_ENH_ENB) | EC01_LAW_ENB |
			EC01_MMIO_ENB, AC_ADVFUNC_CNTL);

		AC_out16(dpy, EC00_RESET,  AC_SUBSYS_CNTL);
		AC_out16(dpy, EC00_ENABLE, AC_SUBSYS_CNTL);

		ACCEL_out16(dpy, MF_MULT_MISC, ECR_MULTIFUNC); /* 16 bit I/O registers */

		/* Now set some basic accelerator registers */
		wait_fifo(dpy, 6);
		ACCEL_out16(dpy, SRC_FOREGROUND | MIX_SRC, ECR_FOREGR_MIX);
		ACCEL_out16(dpy, SRC_BACKGROUND | MIX_SRC, ECR_BACKGR_MIX);/* direct color*/

		ACCEL_out16(dpy, MF_T_CLIP | 0, ECR_MULTIFUNC);     /* clip virtual area  */
		ACCEL_out16(dpy, MF_L_CLIP | 0, ECR_MULTIFUNC);
		ACCEL_out16(dpy, MF_R_CLIP | (mode->request.virt.x-1), ECR_MULTIFUNC);
		ACCEL_out16(dpy, MF_B_CLIP | (mode->request.virt.y-1), ECR_MULTIFUNC);

		wait_fifo(dpy, 5);
		ACCEL_out16(dpy, ~0, ECR_WRITE_MASK);       /* Enable all planes */
		ACCEL_out16(dpy, ~0, ECR_WRITE_MASK);       /* Enable all planes */
		ACCEL_out16(dpy, MF_PIX_CONTROL | MFA_SRC_FOREGR_MIX, ECR_MULTIFUNC);

#endif
		/* !!! reset other registers too. */
	}
	
	NOTICE2("mode set.");
	
	DEBUG1("%.3i MHz %.3i kHz %.3i Hz", \
		mode->dclk / 1000000, \
		(mode->dclk / ((mode->x.total ) - 5)) / 1000, \
		(mode->dclk / ((mode->x.total ) - 5)) / (mode->y.total - 2) );
}

static int chipset_command(struct kgi_graphic *state,ggi_uint cmd,ggi_uint arg)
{
	ggi_coord  *coord;
	struct kgi_display *dpy;

	dpy = state->dev.dpy;
	switch (cmd) {
		case CHIP_SETSPLITLINE:
			if (arg > state->dev.mode.virt.y)
				return -E(CHIPSET,BOGUS);

			kgi_set_split(&(state->dev),arg*dpy->mode.magnify.y);
			return 0;
		case CHIP_SETVISFRAME:
			coord = (ggi_coord *)(state->io_buf);

			if (coord->x < 0 || 
			    coord->x > state->dev.mode.virt.x)
				return -E(CHIPSET,BOGUS);

			if (coord->y < 0 || 
			    coord->y > state->dev.mode.virt.y)
				return -E(CHIPSET,BOGUS);

			kgi_set_origin(&(state->dev), 0, coord->x, coord->y);
			return 0;
		default:
			break;
	}

	return -E(CHIPSET,UNKNOWN);
}

int kgim_chipset_command(struct kgi_graphic *graph, ggi_uint cmd, ggi_uint arg)
{
        return chipset_command(graph, cmd, arg);
}
