/* ----------------------------------------------------------------------------
**	generic PLL routines
** ----------------------------------------------------------------------------
**
**	Copyright	1997		Steffen Seeger
**
**	Generic routines to calculate PLL parameter settings. The formula
**	implemented is 
**
**	fout = (fref * a.mul/a.div * mul/div) / 2^p
**
**	check your manual for further reference.
**
** ----------------------------------------------------------------------------
**
**	MAINTAINER	Steffen_Seeger
*/

/*	find the best paramter set to produce a certain frequency f
*/
static ggi_sint pll_best_state(ggi_uint f, enum kgi_tm_cmd cmd,
	struct pll_state *s, const struct pll_limits *pll)
{
	ggi_ratio	ratio;
	ggi_uint	shift;

	ASSERT(0 < pll->fvco.min);
	ASSERT(pll->fvco.min <= pll->fvco.max);

	ASSERT(pll->p.min <= pll->p.max);

	ASSERT(0 < pll->mul.min);
	ASSERT(pll->mul.min <= pll->mul.max);

	ASSERT(0 < pll->div.min);
	ASSERT(pll->div.min <= pll->div.max);

	/*	clear state
	*/
	memset(s, 0, sizeof(*s));

	/*	determine post-VCO divider/multiplier
	*/
	if ((f < pll->fvco.min) || (pll->fvco.max < f)) {

		if (f < pll->fvco.min) {

			s->p = (pll->p.min < 0) ? 0 : pll->p.min;
			s->fvco = f << s->p;

			while ((s->fvco < pll->fvco.min) &&
				(s->p < pll->p.max)) {

				s->p++;
				s->fvco = f << s->p;
			}

		} else {

			s->p = (pll->p.max > 0) ? 0 : pll->p.max;
			s->fvco = f >> (- s->p);

			while ((s->fvco > pll->fvco.min) &&
				(s->p > pll->p.min)) {

				s->p--;
				s->fvco = f >> (- s->p);
			}
		}

		if ((s->fvco < pll->fvco.min) || (pll->fvco.max < s->fvco) ||
			(s->p < pll->p.min) || (pll->p.max < s->p)) {

			DEBUG1("fvco = %i Hz, p = %i is out of range!", \
				s->fvco, s->p);
			return -EINVAL;
		}

	} else {

		s->fvco = f;
		s->p = 0;
	}

	if (((s->fvco << 1) < pll->fvco.max) && (s->p < pll->p.max)) {

		s->p++;
		s->fvco = f << s->p;
	}

	ratio.mul = s->fvco * pll->a.div;
	ratio.div = pll->fref * pll->a.mul;

	ASSERT(ratio.mul > 0);
	ASSERT(ratio.div > 0);

	/*	avoid integer overflow
	*/
	shift = 0;
	while (((ratio.mul >> shift) > 0x00100000) &&
		((ratio.div >> shift) > 0x00100000)) {

		shift++;
	}

	if (shift) {

		if (cmd == CMD_LOWER) {
	
			ratio.mul = ratio.mul >> shift;
			ratio.div = (ratio.div - ((1 << shift) - 1)) >> shift;

		} else {

			ratio.mul = (ratio.mul - ((1 << shift) - 1)) >> shift;
			ratio.div = ratio.div >> shift;
		}
	}

	if (0 == ratio.mul  ||  0 == ratio.div) {

		DEBUG0("bad parameter set!!");
		return -EINVAL;
	}

	/*	now determine the best mul and div values.
	*/
	s->div = s->mul = -1;
	if ((pll->mul.max - pll->mul.min) < (pll->div.max - pll->div.min)) {

		ggi_sint best_err = 0x7fffffff;
		ggi_sint mul = pll->mul.min;
		ggi_sint div = ((ratio.div * mul) / ratio.mul) +
			(cmd == CMD_LOWER ? 1 : 0);

		while (mul < pll->mul.max) {

			register ggi_sint err = (cmd == CMD_LOWER)
				? div * ratio.mul - mul * ratio.div
				: mul * ratio.div - div * ratio.mul;

			if (0 <= err && (err < best_err)) {

				best_err = err;
				s->mul = mul;
				s->div = div;
			}

			mul++;
			div = ((ratio.div * mul) / ratio.mul) +
				(cmd == CMD_LOWER ? 1 : 0);

			if (pll->div.max < div) {

				break;
			}
		}
		
	} else {

		ggi_sint best_err = 0x7fffffff;
		ggi_sint div = pll->div.min;
		ggi_sint mul = ((ratio.mul * div) / ratio.div) +
			(cmd == CMD_LOWER ? 0 : 1);

		while (div < pll->div.max) {

			register ggi_sint err = (cmd == CMD_LOWER)
				? div * ratio.mul - mul * ratio.div
				: mul * ratio.div - div * ratio.mul;

			if ((0 <= err) && (err < best_err)) {

				best_err = err;
				s->mul = mul;
				s->div = div;
			}

			div++;
			mul = ((ratio.mul * div) / ratio.div) +
				(cmd == CMD_LOWER ? 0 : 1);

			if (pll->mul.max < mul) {

				break;
			}
		}
	}

	if (s->mul < pll->mul.min  ||  pll->mul.max < s->mul ||
		s->div < pll->div.min || pll->div.max < s->div) {

		DEBUG1("parameters out of range");
		return -EINVAL;
	}

	s->f = pll->fref * s->mul / s->div;
	s->f = s->f * pll->a.mul / pll->a.div;
	s->f = (s->p >= 0)
		? s->f >> s->p : s->f << (-s->p);

	DEBUG1("fout = %i, fvco = %i, mul = %i, div = %i, p = %i", \
		s->f, s->fvco, s->mul, s->div, s->p);
	return EOK;
}

/*
**	generic KGI module interface
*/

#ifndef	NO_KGI_INIT_DONE_COMMAND

#ifndef	NO_KGIM_CLOCK_INIT
ggi_sint kgim_clock_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

	card->clock = &clock;
	return EOK;
}
#endif

#ifndef	NO_KGIM_CLOCK_DONE
void kgim_clock_done(struct kgi_display *dpy)
{
	dpy->card->clock = NULL;
	NOTICE(DRIVER_NAME " removed");
}
#endif

#ifndef	NO_KGIM_CLOCK_COMMAND
ggi_sint kgim_clock_command(struct kgi_graphic *state, ggi_uint cmd, ggi_uint arg)
{
	return -E(CLOCK, UNKNOWN);
}
#endif

#endif	/* #ifndef NO_KGI_INIT_DONE_COMMANS*/

#ifndef	NO_KGIM_CLOCK_CHECK_MODE

#define	MIN_FREQ clock.clock.range[0].min
#define	MAX_FREQ clock.clock.range[0].max

ggi_sint kgim_clock_check_mode(struct kgi_display *dpy, struct kgi_mode *mode, 
	enum kgi_tm_cmd cmd)
{
	struct pll_state state;
	ggi_uint newfreq = (cmd == CMD_PROPOSE) 
		? 0 : mode->dclk * mode->clk.mul / mode->clk.div;

	switch (cmd) {

	case CMD_PROPOSE:
		mode->clock = INVALID_COMMAND;
		return EOK;

	case CMD_LOWER:

		if (newfreq < MIN_FREQ) {

			DEBUG1("DCLK = %i Hz, CLK = %i Hz is too low",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		if (newfreq > MAX_FREQ) {

			newfreq = MAX_FREQ;
		}

		if (pll_best_state(newfreq, cmd, &state, &dclk_pll)) {

			DEBUG1("invalid request for DCLK = %i Hz, CLK = %i Hz",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		mode->dclk = state.f * mode->clk.div / mode->clk.mul;
		return EOK;

	case CMD_RAISE:

		if (MAX_FREQ < newfreq) {

			DEBUG1("DCLK = %i Hz, CLK = %i Hz is too high", \
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		if (newfreq < MIN_FREQ) {

			newfreq = MIN_FREQ;
		}

		if (pll_best_state(newfreq, cmd, &state, &dclk_pll)) {

			DEBUG1("invalid request for DCLK = %i Hz, CLK = %i Hz",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		mode->dclk = state.f * mode->clk.div / mode->clk.mul;
		return EOK;

	case CMD_CHECK:

		mode->clock = INVALID_COMMAND;
		if (pll_best_state(newfreq, CMD_LOWER, &state, &dclk_pll)) {

			DEBUG1("invalid request for DCLK = %i Hz, CLK = %i Hz",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		mode->clock = pll_command(&state, mode);
		return EOK;

	default:
		INTERNAL_ERROR;
		return -E(CLOCK, UNKNOWN);
	}
}
#endif

#ifndef	NO_KGIM_CLOCK_SET_MODE
void kgim_clock_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	if (mode->clock != INVALID_COMMAND) {

		pll_set(dpy, mode->clock);

	} else {

		INTERNAL_ERROR;
	}
}
#endif
