/*
 * This header file contains the definitions for the new evstack 
 * coding scheme.
 */

#ifndef _EVSTACK_H_
#define _EVSTACK_H_

#ifndef MODULE
#include <stdio.h>
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#else
#include <linux/module.h>
#include <linux/kernel.h>
#endif

#include <ggi/errno.h>
#include <ggi/debug.h>
#include <ggi/system.h>
#include <ggi/events.h>
#include <ggi/ggi.h>

#ifdef USE_GGI 	/* testing purposes */
# include <ggi/libggi.h>
# define MASTER_VIS _evstack_master_visual
extern ggi_visual_t MASTER_VIS;
#endif /* USE_GGI */

/* Configuration area
 */

#ifndef MODULE
#define printk(form,args...) fprintf(stderr,form,##args);
#define KERN_ERR   "Kernel Error: "
#define KERN_DEBUG "Kernel Debug: "
#endif

/* Maximum size of an Event-Page. Set something resonable.
 * It must be big enough to hold all events for a run through the most
 * complex handler.
 */
#define EVPAGESIZE 8192

#define EV_CLASS_MASK	0x7FFF0000
#define EV_ID_MASK	0x0000FFFF

#define EV_CLASS_MKID(x) (((x)<<16)&EV_CLASS_MASK)
#define EV_MKID(x,i)	 (((x)&EV_CLASS_MASK) | ((i)&EV_ID_MASK))

#define EV_NAME_LEN	64

typedef int ev_id;	/* Class | Unique id */

/* The evstack structure. The core of it all :
 */

typedef struct evstack
{
	ev_id id;		/* ID # for this VC */
	char name[EV_NAME_LEN];	/* How is this device/stack called ? */
	const char *maintainer;	/* Who is in charge of changig me ? */
	uid_t uid;		/* UID for /proc */
	gid_t gid;		/* GID for /proc */
	int perm;		/* permissions for /proc */
	struct evstack_operations *ops;/* Operations for this stack */
	void *private;
	struct evstack *report_to;	/* Normally the parent. */
} evstack;

/* The evstack_operations structure */
typedef struct evstack_operations {
	int (*open)(struct evstack *blank);	/* Create a new stack */
	int (*close)(struct evstack *slf);	/* Close the stack */
	void (*handler)(struct evstack *slf,	/* Handle an event */
		       ggi_event *first_event,int *space_left);
	int (*uses)(struct evstack *slf);	/* Get # users of the stack */
	int (*write)(struct evstack *slf,
			char *string);		/* configure stack. string may be  modified for parsing purposes */
	int (*read)(struct evstack *slf,
			int lineno,int maxlen,
			char *string);		/* get stack config */
} evstack_operations;

/* Inlined functions */

/*none yet*/

/* Prototypes for registering/unregistering items */

/* When you register you get a unique id for your `class' of stack/head
 * Use this id for de-registering a class of stacks 
 */
ev_id ev_register_class(const char *classname,evstack_operations *stack);
int   ev_unregister_class(ev_id stack_class); 

evstack *ev_create_stack(ev_id stack_class);
evstack *ev_clone_stack(evstack *stack);
int	 ev_destroy_stack(evstack *stack);

ev_id ev_lookup_class(const char *name);	/* Get classid from name */
evstack *ev_lookup_stack(const char *name,ev_id id); /* name==NULL - use id 
						* Id lowbits = 0, get first
						*   stack found
						*/

/* Handling EventPages 
 */
ggi_event *ev_evpage_alloc(void);
void       ev_evpage_free(ggi_event *evpage);
ggi_event *ev_evpage_next(ggi_event *currev);
ggi_event *ev_evpage_add (ggi_event *currev,int *len,ggi_event *what_to_add);
ggi_event *ev_evpage_tail(ggi_event *one_event_in_there);

/* Sending Events
 */
int ev_send(ggi_event *evpage,int *left,ggi_event *event_to_send, 
	    evstack *whereto);

/* modifying events
 */
static inline void ev_consume(ggi_event *event) 
{ event->any.type|=0x80; }

static inline int ev_is_consumed(ggi_event *event) 
{ return !!(event->any.type&0x80); }

static inline int ev_unconsumed_type(ggi_event *event) 
{ return event->any.type&0x7f; }

/*\
|*| Misc thing that don't really belong here.
\*/

void *ev_malloc(size_t sz);
void  ev_free(void *ptr);

enum Commands { CmdTtyOutWriteChar  =0x00010000,	/* Chars from /dev/tty */
		CmdTtyInWriteChar   =0x00010010,	/* Chars to   /dev/tty */

		CmdScrollWriteChar  =0x00011000,	/* Scroller events */
		CmdScrollScroll     =0x00011010,
		CmdScrollGotoXY     =0x00011020,
		CmdScrollOrigin	    =0x00011030,
		CmdScrollReadChar   =0x00011040,

		CmdVtSwitch	    =0x00012000,	/* Administrative commands */
		CmdVtActivate	    =0x00012010,
		CmdStkReprocess	    =0x00012020,
		CmdEventLost	    =0x00012100,	/* Set at places where overflows occurred */

		CmdMakeSound	    =0x00020000,

                CmdPrivate          =0x80000000,
              };


typedef struct { unsigned char size ; char data[128]; 
		} DataTtyWriteChar;

typedef struct { ggi_color fg_col,bg_col;
		 int attrib;
#define AT_BOLD		0x0001
#define AT_UNDERLINE	0x0002
#define AT_BLINK	0x0004
#define AT_ITALIC	0x0008
#define AT_SUBSCRIPT	0x0010
#define AT_SUPERSCRIPT	0x0020
		 char dx,dy;	/* drawing direction */
		 unsigned char numchr;
		 unicode chr[100];
		} DataScrollWriteChar;

typedef struct { int x,y;
		 int width,height;
		 int dx,dy;
		} DataScrollScroll;

typedef struct { int x,y;
		 int curnum;
#define CUR_NORM	0x0000
#define CUR_MOUSE	0x0001
#define CUR_SELECTSTART	0x0002
#define CUR_SELECTEND	0x0003
		 char fixedpointshift;
		} DataScrollGotoXY;

typedef struct { int x,y;
		} DataScrollOrigin;

typedef struct { int newvt;
		} DataVtSwitch;

typedef struct { int onoff;
		} DataVtActivate;

typedef struct { int frequency;	/* In Hz */
		 int duration;	/* milliseconds */
		 char type[32];	/* descriptive string */
		} DataMakeSound;

extern inline void build_cmd_event(void *buffer, uint32 command, void *data, size_t length)
{
	ggi_cmd_event *cmd=buffer;
	cmd->type=evCommand;
	cmd->size=sizeof(*cmd)+length;
	cmd->code=command;
	if (length) memcpy(cmd->data,data,length);
	ev_evpage_next((ggi_event *)buffer)->size=0;
}

#endif
