/*
 * This file contains the functions for a head.
 *
 * A head is a container which can have up to MAXSUBSTACK clients.
 * Events are broadcast to all clients. All clients see all events.
 *
 * A head understands the following commands :
 *
 * Command Short Parameters
 * attach  +     stack_to_attach
 * detach  -     stack_to_detach
 */


#undef DEBUG
#ifdef DEBUG
#define DEBUG_LEVEL 1
#endif

#include "evstack.h"
#include "parsearg.h"

#define MAXSUBSTACK 16

typedef struct head_state
{
	int refcount;
	evstack *substacks[MAXSUBSTACK];
} head_state;

static head_state defaultstate = {
	1,
	{ NULL, }
};

#define STACKTYPEDEF head_state
#include "substack.inc"

#define CLASSNAME "head"

static ev_id head_id;
static int head_use=0;

static int head_open(evstack *blank);
static int head_close(evstack *slf);
static void head_handler(evstack *slf,ggi_event *ev,int *left);
static int head_uses(evstack *slf);
static int head_write(evstack *slf, char *str);
static int head_read(evstack *slf, int lineno, int maxlen, char *str);

static evstack_operations head_ops = {
	head_open,
	head_close,
	head_handler,
	head_uses,
	head_write,
	head_read
};

static int head_open(evstack *blank)
{
	head_state *dat;

	sprintf(blank->name,"Head v1.0 (instance %.4x)",
		blank->id & EV_ID_MASK);

	blank->maintainer="becka@sunserver1.rz.uni-duesseldorf.de";

	if ((dat=blank->private))
	{
		dat->refcount++;
		DEBUG2("Stack cloned (rc %d).",dat->refcount);
	}
	else
	{	
		blank->private=dat=(head_state *)ev_malloc(sizeof(head_state));
		memcpy(dat,&defaultstate,sizeof(head_state));
		DEBUG2("Stack created (rc %d).",dat->refcount);
	}

        MOD_INC_USE_COUNT;
	head_use++;

	return 0;
}

static int head_close(evstack *slf)
{
	head_state *des=slf->private;
        
	if (--des->refcount==0)		/* No more instances left */
	{
		ev_free(des);
		DEBUG2("Stack destroyed.",0);
	}

        MOD_DEC_USE_COUNT;
	head_use--;
	
	return 0;
}


/* Event handler code.
 *
 * Switch-behaviour. Only first member matching the current vt-num will be fed.
 */

static inline void tosubstack(struct head_state *ss,ggi_event *ev,int *left)
{
  ggi_event *havetosee[MAXSUBSTACK];
  int x,oldleft;

  for(x=0;x<MAXSUBSTACK;x++) havetosee[x]=ev;
          
  do {
    /* Keep for seeing, if the page has grown */
    oldleft=*left;

    for(x=0;x<MAXSUBSTACK&&ss->substacks[x];x++)
    {
      /* If not all events in the queue have been seen by this driver,
         send the remaining ones. */
      if (havetosee[x]< ev_evpage_tail(havetosee[x]))
      { 
	ev_send(havetosee[x],left,NULL,ss->substacks[x]);

        /* Now it has seen them. Remember this. */
        havetosee[x]= ev_evpage_tail(havetosee[x]);
      }
    }

  /* See if we have new events : */
  } while( oldleft != *left );
}

static void head_handler(evstack *slf,ggi_event *ev,int *left)
{ 
	struct head_state *ss=slf->private;

	/* now send it to the selected of the attached substacks. */
	tosubstack(ss,ev,left);

}

static int head_uses(evstack *slf)
{
	return head_use;
}

static int head_write(evstack *slf, char *str)
{
	int rc,help;
	char buffer[256];
	struct head_state *ss=slf->private;
	
	DEBUG1("%s->write:%s",slf->name,str);

	if ( 1 != sscanf(str,"%s%n",buffer,&help) ) return 0;	/* empty */
	str+=help;
	
	if (strcmp(buffer,"attach")==0 ||
	    strcmp(buffer,"+")==0 ) 
	{
		evstack *hlpstk;
		if ( 1 != sscanf(str,"%s%n",buffer,&help) ) 
		{
			DEBUG1("Invalid attach command.\n",0);return -EINVAL;
		}
		if (!(hlpstk=par_string_to_stack(buffer))) return -EINVAL;
		rc=attach_ev_stack(slf,hlpstk,NULL);
    		if (rc<0) return rc;

		hlpstk->report_to=slf;
    		return 0;
	}
	else if (strcmp(buffer,"detach")==0 ||
		 strcmp(buffer,"-")==0 ) 
	{
		evstack *hlpstk;
		if ( 1 != sscanf(str,"%s%n",buffer,&help) ) 
		{
			DEBUG1("Invalid detach command.\n",0);return -EINVAL;
		}
		if (!(hlpstk=par_string_to_stack(buffer))) return -EINVAL;
		rc=detach_ev_stack(slf,hlpstk);
    		if (rc<0) return rc;

		hlpstk->report_to=NULL;
    		return 0;
	}
	return -EINVAL;
}

static int head_read(evstack *slf,int lineno,int maxlen,char *str)
{
	struct head_state *ss=slf->private;
	if (lineno==0) 
	{
		sprintf(str,"# Head %s consists of :",par_stack_to_string(slf)); return 0;
	} else
	if (lineno<=MAXSUBSTACK)
	{
		if (ss->substacks[lineno-1])
		{	sprintf(str,"+ %s   #%s",par_stack_to_string(ss->substacks[lineno-1]),
				ss->substacks[lineno-1]->name); 
			return 0;
		}
	}
	return -EINVAL;
}

#ifdef MODULE
#define init_head init_module
#endif

int init_head(void)
{ 
	head_id=ev_register_class(CLASSNAME,&head_ops);
	printk("Registered %s.\n",CLASSNAME);
	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{ 
	ev_unregister_class(head_id);
	printk("Unregistered %s.\n",CLASSNAME);
}
#endif
