/*
 * File to be included for simple access to display scrolling commands.
 *
 * Scroll_scroll,scroll_gotoxy and scroll_write can be called without further 
 * precautions. However if you use them, you _must_ call scroll_flush()
 * before returning, or weird things might happen.
 * 
 * This is because we do some caching here to speed things up.
 *
 * All other commands rely on the preprocessor macroc SCROLLERSTATE to be 
 * defined as the type which contains the following information fields :
 * curx,cury, (current cursor position)
 * width,height (size of the screen as the terminal parser thinks it is)
 *
 * These fields are accessed as SC_CURX,SC_CURY,... via macros which 
 * default to #define SC_CURX ps->curx, ... . You can however override these 
 * by defining such makros yourself. Scroll.inc will NOT override them.
 * The pointer to the parser-state is always called ps.
 */

union cache {
	char outbuf[256];
	DataScrollWriteChar dswc;
	DataScrollGotoXY    dsg;
	DataScrollOrigin    dso;
};

static DataScrollWriteChar scroll_write_buffer = {
	{0xffff,0xffff,0xffff},{0x0000,0x0000,0x0000},
	0,
	1,0,
	0,
	{0,}
};

#define scroll_write(x,y,z) _scroll_write(x,y,z,0)

static inline void _scroll_write(ggi_event *ev,int *left,unicode chr,int flush)
{
	if ( !flush ) scroll_write_buffer.chr[scroll_write_buffer.numchr++]=chr;
	if ( flush ||
	     scroll_write_buffer.numchr 
		    >=  sizeof(scroll_write_buffer.chr)/
			sizeof(scroll_write_buffer.chr[0]) )
	{
		char outbuf[256];
		build_cmd_event(outbuf,CmdScrollWriteChar,&scroll_write_buffer,sizeof(scroll_write_buffer));
		DEBUG1("ScrollScroll\n",0);
		ev_evpage_add(ev,left,(ggi_event *)&outbuf);
		scroll_write_buffer.numchr=0;
	}
}

static inline void scroll_scroll(ggi_event *ev,int *left,int x,int y,int width,int height,int dx,int dy)
{
	DataScrollScroll dss;
	char outbuf[256];

	dss.x=x;dss.y=y;dss.width=width;dss.height=height;
	dss.dx=dx;dss.dy=dy;

	build_cmd_event(outbuf,CmdScrollScroll,&dss,sizeof(dss));
	DEBUG1("ScrollScroll\n",0);
	ev_evpage_add(ev,left,(ggi_event *)&outbuf);
}

static inline void scroll_gotoxy(ggi_event *ev,int *left,int x,int y,int curnum,int fpshift)
{
	DataScrollGotoXY dss;
	char outbuf[256];

	dss.x=x;dss.y=y;dss.curnum=curnum;dss.fixedpointshift=fpshift;

	build_cmd_event(outbuf,CmdScrollGotoXY,&dss,sizeof(dss));
	DEBUG2("ScrollGotoXY\n",0);
	ev_evpage_add(ev,left,(ggi_event *)&outbuf);
}

static void inline scroll_flush(ggi_event *ev,int *left)
{
	_scroll_write(ev,left,0,1);
}

static void inline scroll_mksound(ggi_event *ev, int *left, int pitch, int duration, char *class)
/* This produces a sound of the given pitch in Hz for the given duration
 * in ms. Class is used to distinguish different sound for advanced
 * sound subsystems
*/
{ 
	DataMakeSound ms;
	char outbuf[256];

	ms.frequency=pitch;
	ms.duration =duration;
	strcpy(ms.type,class);

	build_cmd_event(outbuf,CmdMakeSound,&ms,sizeof(ms));
	DEBUG1("ScrollMakeSound\n",0);
	ev_evpage_add(ev,left,(ggi_event *)&outbuf);
}

#ifdef SCROLLERSTATE

#ifndef SC_CURX
#define SC_CURX ps->curx
#endif

#ifndef SC_CURY
#define SC_CURY ps->cury
#endif

#ifndef SC_WIDTH
#define SC_WIDTH ps->width
#endif

#ifndef SC_HEIGHT
#define SC_HEIGHT ps->height
#endif

#ifndef SC_TOP
#define SC_TOP 0
#endif

#ifndef SC_BOTTOM
#define SC_BOTTOM (ps->height-1)
#endif

#ifndef SC_LEFT
#define SC_LEFT 0
#endif

#ifndef SC_RIGHT
#define SC_RIGHT (ps->width-1)
#endif


static void inline scroll_lf(SCROLLERSTATE *ps,ggi_event *ev, int *left)
/* Perform a line feed operation, that is move the writing position down
 * one line or scroll up line if at the SC_BOTTOM line of the current scrolling
 * region. Will not work correctly, if out of scrolling region.
 */
{
	if (SC_CURY==SC_BOTTOM)
	{	scroll_scroll(ev,left,SC_LEFT,SC_TOP+1,
				      (SC_RIGHT)-(SC_LEFT)+1,
				      (SC_BOTTOM)-(SC_TOP),
					0,-1);
	}
	else
	{	SC_CURY++;
		scroll_gotoxy(ev,left,SC_CURX,SC_CURY, CUR_NORM, 0);
	}
}

static void inline scroll_reverse_lf(SCROLLERSTATE *ps,ggi_event *ev, int *left)
/* Perform a reverse linefeed operation. Inverse of the above.
*/
{
	if (SC_CURY==SC_TOP)
	{	scroll_scroll(ev,left,SC_LEFT,SC_TOP,
				      (SC_RIGHT)-(SC_LEFT)+1,
				      (SC_BOTTOM)-(SC_TOP),
					0,1);
	}
	else
	{	SC_CURY--;
		scroll_gotoxy(ev,left,SC_CURX,SC_CURY, CUR_NORM, 0);
	}
}

static void inline scroll_cr(SCROLLERSTATE *ps,ggi_event *ev, int *left)
{ 
	scroll_gotoxy(ev,left,SC_CURX=0,SC_CURY, CUR_NORM, 0);
}

static void inline scroll_bs(SCROLLERSTATE *ps,ggi_event *ev, int *left)
{ 
	if (SC_CURX) 
	{	SC_CURX--;
		scroll_gotoxy(ev,left,SC_CURX,SC_CURY, CUR_NORM, 0);
	}
}

static void inline scroll_up(SCROLLERSTATE *ps, ggi_event *ev, int *left, ggi_uint t, uintf b, uintf n)
/* Scroll the area between top and bottom up by n lines.
*/
{
	scroll_scroll(ev,left,SC_LEFT,t+n,
			     (SC_RIGHT)-(SC_LEFT)+1,
			     b-t+1-n,
				0,-n);
}

static void inline scroll_down(SCROLLERSTATE *ps, ggi_event *ev, int *left,ggi_uint t, uintf b, uintf n)
/* Scroll the area between top and bottom down by n lines.
*/
{ 
	scroll_scroll(ev,left,SC_LEFT,t,
			     (SC_RIGHT)-(SC_LEFT)+1,
			     b-t+1-n,
				0,n);
}

#if 0

void scroll_write(SCROLLERSTATE *cons, unicode c);
/* Put the font representation of the character c at the current
 * writing position and move the writing position one position to the right.
 * This does not mark the modified area using scroll_modified(),
 * because usually you have to write lots of characters and don't change the
 * writing position against the normal direction. Thus the intended use is
 * to keep the current writing position POS before a series of scroll_write() and
 * call scroll_modified(oldpos, POS); whenever moving the current
 * writing direction against the normal writing direction (which is left to
 * right, top to down). Also, scroll_write() does not perform
 * a CR LF sequence or change the writing position when the
 * current writing position reaches the leftmost column. Instead it sets the
 * flag need_wrap which can be checked using FLAG(need_wrap)
 * by the parser.
 */


void scroll_insert_chars(SCROLLERSTATE *cons, ggi_uint n)
{}

/*
This inserts n characters at the current writing position in
the current line. The 'fresh' area is cleared using the ERASE
attributes. The modified area is marked properly using scroll_modified().
*/

void scroll_delete_chars(SCROLLERSTATE *cons, ggi_uint n);
/*
This deletes n characters at the current writing position in
the current line. The 'fresh' area is cleared using the ERASE
attributes. The modified area is marked properly using scroll_modified().
*/

void scroll_erase_chars(SCROLLERSTATE *cons, ggi_uint n)
{}
/*
This erases (overwrites)&nbsp;n characters starting at the
current writing position in the current line. The erased area is cleared
using the ERASE attributes. The modified area is marked properly
using scroll_modified().
*/

void scroll_backward(SCROLLERSTATE *cons, int lines)
{}
/*
Increase the scrollback-offset by half the screen size (lines &lt;=
0) or the specified amount of lines (lines &gt; 0). This
also marks the given console to need a scroll_sync() operation.
*/

void scroll_forward(SCROLLERSTATE *cons, int lines);
{}
/*
Decrease the scrollback-offset by half the screen size (lines &lt;=
0) or the specified amount of lines (lines &gt; 0). This
also marks the given console to need a scroll_sync() operation.
*/

#endif
#endif
