/*
 * This file contains the functions for a vtswitch.
 *
 * A vtswitch is a container which can have up to MAXSUBSTACK clients
 * events are always dispatched to only one client CmdVtSwitch events 
 * cause the target client to change.
 *
 * A vtswitch understands the following commands :
 *
 * Command Short Parameters
 * attach  +     stack_to_attach VT_number_to_use
 * detach  -     stack_to_detach
 * setvt   =     stack_to_change new_VT_number
 */

#undef DEBUG
#ifdef DEBUG
#define DEBUG_LEVEL 2
#endif

#include "evstack.h"
#include "parsearg.h"

#define MAXSUBSTACK 16

typedef struct vtswitch_state
{
	evstack *substacks[MAXSUBSTACK];
	int vtnumber[MAXSUBSTACK];
	int current_vt;
	int refcount;
} vtswitch_state;

static vtswitch_state defaultstate = {
	{ NULL, },
	{0, },
	0,
	1
};

#define STACKTYPEDEF vtswitch_state
#include "substack.inc"

#define CLASSNAME "vtswitch"

static ev_id vtswitch_id;
static int vtswitch_use=0;
static int vtswitch_open(evstack *blank);
static int vtswitch_close(evstack *slf);
static void vtswitch_handler(evstack *slf,ggi_event *ev,int *left);
static int vtswitch_uses(evstack *slf);
static int vtswitch_write(evstack *slf, char *str);
static int vtswitch_read(evstack *slf, int lineno, int maxlen, char *str);

static evstack_operations vtswitch_ops = {
	vtswitch_open,
	vtswitch_close,
	vtswitch_handler,
	vtswitch_uses,
	vtswitch_write,
	vtswitch_read
};

static int vtswitch_open(evstack *blank)
{
	vtswitch_state *dat;

	sprintf(blank->name,"VtSwitch v1.0 (instance %.4x)",
		blank->id & EV_ID_MASK);

	blank->maintainer="becka@sunserver1.rz.uni-duesseldorf.de";

	if ((dat=blank->private))
	{
		dat->refcount++;
		DEBUG2("Stack cloned (rc %d).",dat->refcount);
	}
	else
	{	
		blank->private=dat=(vtswitch_state *)ev_malloc(sizeof(vtswitch_state));
		memcpy(dat,&defaultstate,sizeof(vtswitch_state));
		DEBUG2("Stack created (rc %d).",dat->refcount);
	}

        MOD_INC_USE_COUNT;
	vtswitch_use++;

	return 0;
}

static int vtswitch_close(evstack *slf)
{
	vtswitch_state *des=slf->private;
        
	if (--des->refcount==0)		/* No more instances left */
	{
		ev_free(des);
		DEBUG2("Stack destroyed.",0);
	}

        MOD_DEC_USE_COUNT;
	vtswitch_use--;
	
	return 0;
}


static inline void setreport(struct vtswitch_state *ss,evstack *stack)
{
	int x;
	
	for(x=0;x<MAXSUBSTACK&&ss->substacks[x];x++)
	{
  	  	if (ss->vtnumber[x]==ss->current_vt)
		{
			ss->substacks[x]->report_to=stack;
			break;/* Only send to first matching stack.*/
		}
	}
}
/* Event handler code.
 *
 * Switch-behaviour. Only first member matching the current vt-num will be fed.
 */

static inline void tosubstack(struct vtswitch_state *ss,ggi_event *ev,int *left)
{
	int x;
	
	for(x=0;x<MAXSUBSTACK&&ss->substacks[x];x++)
	{
  	  	if (ss->vtnumber[x]==ss->current_vt)
		{
	  	  	DEBUG1("sending to %s",ss->substacks[x]->name);
			ev_send(ev,left,NULL,ss->substacks[x]);
			break;/* Only send to first matching stack.*/
		}
	}
}

static void vtswitch_handler(evstack *slf,ggi_event *ev,int *left)
{ 
	ggi_cmd_event *evdat;
	struct vtswitch_state *ss=slf->private;

	/* now send it to the selected of the attached substacks. */
	tosubstack(ss,ev,left);

	/* now check for VT-switch */
	for(;ev->size;ev=ev_evpage_next(ev))
	{
		switch(ev->any.type)
		{ 
		  case evCommand:
		  	evdat=(ggi_cmd_event *)ev;
		  	switch(evdat->code)
		  	{
		  	  case CmdVtSwitch:
		  	  	ev_consume(ev);
				DEBUG1("got event %x.\n",evdat->code);
				{ DataVtSwitch *dt=(DataVtSwitch *)evdat->data;
				  
				  ggi_event *evpage;
				  char outbuf[256];
				  DataVtActivate vta;
				  int pagesize;

				  setreport(ss,NULL);
				  evpage=ev_evpage_alloc();
				  evpage->size=0;
				  pagesize=EVPAGESIZE;
				  vta.onoff=0;
				  build_cmd_event(outbuf,CmdVtActivate,&vta,sizeof(vta));
				  ev_evpage_add(evpage,&pagesize,(ggi_event *)outbuf);
				  tosubstack(ss,evpage,&pagesize);

				  DEBUG1("Switchto %d.\n",dt->newvt);
				  ss->current_vt=dt->newvt;

				  setreport(ss,slf);
				  evpage->size=0;
				  pagesize=EVPAGESIZE;
				  vta.onoff=1;
				  build_cmd_event(outbuf,CmdVtActivate,&vta,sizeof(vta));
				  ev_evpage_add(evpage,&pagesize,(ggi_event *)outbuf);
				  tosubstack(ss,evpage,&pagesize);

				  ev_evpage_free(evpage);

				} break;
		  	  default:
		  	  	break;
			}
			break;
		  default: 
		  	break;
		}
	}

}

static int vtswitch_uses(evstack *slf)
{
	return vtswitch_use;
}

static int vtswitch_write(evstack *slf, char *str)
{
	int rc,help;
	char buffer[256];
	struct vtswitch_state *ss=slf->private;
	
	DEBUG1("%s->write:%s\n",slf->name,str);

	DEBUG2("%d",sscanf(str,"%s%n",buffer,&help));
	if ( 1 != sscanf(str,"%s%n",buffer,&help) ) return 0;	/* empty */
	str+=help;
	DEBUG2("arg0:%s,-%s\n",buffer,str);
	
	if (strcmp(buffer,"attach")==0 ||
	    strcmp(buffer,"+")==0 ) 
	{
		evstack *hlpstk;
		int vtnum;	
		if ( 2 != sscanf(str,"%s%d%n",buffer,&vtnum,&help) ) 
		{
			DEBUG1("Invalid attach command.\n",0);return -EINVAL;
		}
		if (vtnum < 0 || !(hlpstk=par_string_to_stack(buffer))) return -EINVAL;
		rc=attach_ev_stack(slf,hlpstk,NULL);
    		if (rc<0) return rc;
    		ss->vtnumber[rc]=vtnum;
    		return 0;
	}
	else if (strcmp(buffer,"detach")==0 ||
		 strcmp(buffer,"-")==0 ) 
	{
		evstack *hlpstk;
		if ( 1 != sscanf(str,"%s%n",buffer,&help) ) 
		{
			DEBUG1("Invalid detach command.\n",0);return -EINVAL;
		}
		if (!(hlpstk=par_string_to_stack(buffer))) return -EINVAL;

		rc=detach_ev_stack(slf,hlpstk);
    		if (rc<0) return rc;

		for(rc++;rc<MAXSUBSTACK&&ss->substacks[rc];rc++)  /* Move the tail of the chain */
		{
			ss->vtnumber[rc-1]=ss->vtnumber[rc];
		}

		ss->vtnumber[rc-1]=-1;      /* Unset vt number. */
		return 0;
	}
	else if (strcmp(buffer,"setvt")==0 ||
		 strcmp(buffer,"=")==0 ) 
	{
#if 0
		evstack *hlpstk;
		int vtnum;	
		if ( 2 != sscanf(str,"%s%d%n",buffer,&vtnum,&help) ) 
		{
			DEBUG1("Invalid attach command.\n",0);return -EINVAL;
		}
		if (vtnum < 0 || !(hlpstk=par_string_to_stack(buffer))) return -EINVAL;
		rc=!!!search!!!_ev_stack(slf,hlpstk,NULL);
    		if (rc<0) return rc;
    		ss->vtnumber[rc]=vtnum;
    		return 0;
#else
		printk("Command not implemented.");
#endif
	}
	return -EINVAL;
}

static int vtswitch_read(evstack *slf,int lineno,int maxlen,char *str)
{
	struct vtswitch_state *ss=slf->private;
	if (lineno==0) 
	{
		sprintf(str,"# VT-Switcher %s consists of :",par_stack_to_string(slf)); return 0;
	} else
	if (lineno<=MAXSUBSTACK)
	{
		if (ss->substacks[lineno-1])
		{	sprintf(str,"+ %s %5d   #%s",par_stack_to_string(ss->substacks[lineno-1]),
				ss->vtnumber[lineno-1],ss->substacks[lineno-1]->name); 
			return 0;
		}
	}
	return -EINVAL;
}

                                                        
#ifdef MODULE
#define init_vtswitch init_module
#endif

int init_vtswitch(void)
{ 
	vtswitch_id=ev_register_class(CLASSNAME,&vtswitch_ops);
	printk("Registered %s.\n",CLASSNAME);
	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{ 
	ev_unregister_class(vtswitch_id);
	printk("Unregistered %s.\n",CLASSNAME);
}
#endif
