
/*
 *  Flying-GGIs
 *
 *  Another neat GGI demo...
 *
 *  Copyright (c) 1997-1998 Andrew Apted.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License.
 *
 *  USAGE: flying_ggis [--dim 320 200] [--gentime 100] [--speed 50]
 *                     [--lifetime 3000] [--maxsize 100 ]
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <sys/time.h>

#include <ggi/libggi.h>


#define fixed  sint32   /* 16.16 */

#define DEFAULT_WIDTH   320
#define DEFAULT_HEIGHT  200

#define DEFAULT_GENTIME   100   /* msec */
#define DEFAULT_LIFETIME  3000  /* msec */
#define DEFAULT_MAXSIZE   100   /* percent */
#define DEFAULT_SPEED     50 


typedef struct global_info
{
        int screen_width;
        int screen_height;
        int screen_diag;

        int banner_width;
        int banner_height;
        int banner_diag;

        int gen_time;   /* msec */
        int life_time;  /* msec */
        int speed;

        fixed max_size;   /* default 1.0 = full screen */

        ggi_visual_t *vis;
        ggi_color palette[256];

        uint8 *page;
        int pagelen;

} GlobalInfo;

static GlobalInfo GI;

static char *banner[46];


typedef struct texture
{
        struct texture *succ;
        struct texture *tail;

        fixed mid_x, mid_y;
        fixed size;
        fixed millis;
        fixed speed;

        uint8 color;
} Texture;

static Texture *texture_list;


static void banner_size(int *width, int *height)
{
        *width=0;

        for (*height=0; banner[*height] != NULL; (*height)++) {

                int len = strlen(banner[*height]);

                if (len > *width) {
                        *width = len;
		}
        }
}

static void setup_palette(void)
{
        int i;

        for (i=0; i < 256; i++) {
                GI.palette[i].r = ((i >> 5) & 0x7) << (16-3);
                GI.palette[i].g = ((i >> 2) & 0x7) << (16-3);
                GI.palette[i].b = ((i)      & 0x3) << (16-2);
        }

        ggiSetPaletteVec(GI.vis, 0, 256, GI.palette);
}

void update_frame(void)
{
        ggiPutBox(GI.vis, 0, 0, GI.screen_width, GI.screen_height, GI.page);
        ggiFlush(GI.vis);
}

static void init_textures(void)
{
        texture_list=NULL;
}

static void free_textures(void)
{
	Texture *t;

        while (texture_list != NULL) {
		
		t = texture_list;
		texture_list = t->succ;
		
		free(t);
	}
	
}

static void add_texture(int x, int y, uint8 col)
{
        Texture *t;

        t = (Texture *) malloc(sizeof(Texture));

        t->mid_x = x << 16;
        t->mid_y = y << 16;

        t->millis = 0;
        t->color  = col;
        t->speed  = 8 + (random() % GI.speed);

        t->succ = texture_list;
        texture_list = t;
}

static void render_texture(int width, int height, Texture *t)
{
        int x, y;
        int sx, sy, bx;
        int dx, dy;

        uint8 *dest;

        height <<= 16;
        width  <<= 16;

        dx = dy = t->size * GI.screen_diag / GI.banner_diag;

        bx = t->mid_x - (GI.banner_width  * dx / 2);
        sy = t->mid_y - (GI.banner_height * dy / 2);

        for (y=0; (banner[y] != NULL) && (sy < height); y++, sy += dy) {

                char *pos = banner[y];

                if (sy >= 0) {

                        dest = GI.page + ((sy>>16) * GI.screen_width);

                        for (x=0, sx=bx; (*pos != 0) && (sx < width); 
			     x++, pos++, sx += dx) {
			
                                if ((sx >= 0) && (*pos == '#'))
                                {
                                        dest[sx>>16] = t->color;
                                }
                        }
                }
        }
}

static void update_texture(Texture *t, Texture ***prev_ptr, int millis)
{
        t->millis += millis;

        t->size = t->millis * t->speed;

        if ((t->millis > GI.life_time) || (t->size > GI.max_size)) {
                
                /* remove texture */

                **prev_ptr = t->succ;

		free(t);

                return;
        }

        *prev_ptr = &t->succ;

        render_texture(GI.screen_width, GI.screen_height, t);

        if (0) {
                fixed ux = (t->mid_x>>15) - GI.screen_width;
                fixed uy = (t->mid_y>>15) - GI.screen_height;
        
                t->mid_x += ux * 100;
                t->mid_y += uy * 100;
        }
}

static void update_all_textures(int millis)
{
        Texture *cur;
        Texture **prev_ptr;

        cur = texture_list;
        prev_ptr = (Texture **) &texture_list;

        while (cur != NULL) {

                update_texture(cur, &prev_ptr, millis);

                cur = *prev_ptr;
        }
}

static void show_usage(char *progname)
{
	/* remove leading paths */

	if (strrchr(progname, '/') != NULL) {
		progname = strrchr(progname, '/') + 1;
	}

	printf("\nUSAGE: "
		"%s OPTIONS...\n\n"
		"Options:\n"
		"    -d, --dim       <width> <height>\n" 
		"    -g, --gentime   <generation time>   (ms)\n"
		"    -l, --lifetime  <life time>         (ms)\n"
		"    -s, --speed     <speed>\n"
		"    -m, --maxsize   <percentage>\n\n",
		progname);
}

void main(int argc, char **argv, char **envp)
{ 
        int gen_millis;

        struct timeval prev_time, cur_time;

	int i;


        /* initialize */

        srand(time(NULL));

        banner_size(&GI.banner_width, &GI.banner_height);

        GI.banner_diag = sqrt(GI.banner_width * GI.banner_width +
                              GI.banner_height * GI.banner_height);

        GI.screen_width  = DEFAULT_WIDTH;
        GI.screen_height = DEFAULT_HEIGHT;

        GI.gen_time = DEFAULT_GENTIME;

        GI.speed     = -1;
        GI.life_time = -1;
        GI.max_size  = -1;

        init_textures();

        
	/* handle arguments */

	for (i=1; i < argc; i++) {

		#define CMPOPT(x,s,l,n)  (((strcmp(x,s)==0) || \
				(strcmp(x,l)==0)) && ((i+n) < argc))

		if (CMPOPT(argv[i], "-h", "--help", 0)) {
		    
			show_usage(argv[0]);
			exit(0);
		}

		if (CMPOPT(argv[i], "-d", "--dim", 2)) {
		    
			i++; GI.screen_width  = atoi(argv[i]);
			i++; GI.screen_height = atoi(argv[i]);
			continue;
		}

		if (CMPOPT(argv[i], "-g", "--gentime", 1)) {

			i++; GI.gen_time = atoi(argv[i]);
			continue;
		}

		if (CMPOPT(argv[i], "-s", "--speed", 1)) {

			i++; GI.speed = atoi(argv[i]);
			continue;
		}

		if (CMPOPT(argv[i], "-l", "--lifetime", 1)) {

			i++; GI.life_time = atoi(argv[i]);
			continue;
		}

		if (CMPOPT(argv[i], "-m", "--maxsize", 1)) {

			i++; GI.max_size = atoi(argv[i]);
			continue;
		}

		#undef CMPOPT

		fprintf(stderr, "Unknown option '%s'\n", argv[i]);
		exit(1);
	}


	/* sort out parameters */

	if (GI.speed < 0) {
		GI.speed = DEFAULT_SPEED;
	}

	if (GI.life_time < 0) {
		GI.life_time = DEFAULT_LIFETIME;
	}

	if (GI.max_size < 0) {
		GI.max_size = DEFAULT_MAXSIZE;
	}

	GI.max_size = (GI.max_size << 16) / 100;


        /* setup graphics mode */

	if (ggiInit() != 0) {
		fprintf(stderr, "%s: unable to initialize libggi, exiting.\n",
			argv[0]);
		exit(1);
	}
        GI.vis = ggiOpen(NULL);

        if (GI.vis == NULL) {
                fprintf(stderr, "Failed to open visual.\n");
                exit(1);
        }

	ggiSetInfoFlags(GI.vis, GGIFLAG_ASYNC);
	
        ggiSetGraphMode(GI.vis, GI.screen_width, GI.screen_height,
                        GI.screen_width, GI.screen_height, GT_8BIT);
        
        setup_palette();
        

        GI.screen_diag = sqrt(GI.screen_width  * GI.screen_width +
                              GI.screen_height * GI.screen_height);

        GI.pagelen = GI.screen_width * GI.screen_height * 1;

        GI.page = malloc(GI.pagelen);


        /* main loop */

        gettimeofday(&prev_time, NULL);

        gen_millis=0;

        for(;;) {
                int millis;

                struct timeval tv = { 0, 10000 };

                if (ggiEventPoll(GI.vis, emKeyPress, &tv)) {
		
			ggi_event ev;
			int done=0;

			ggiEventRead(GI.vis, &ev, emKeyPress);

			if (ev.any.type == evKeyPress) {

				int key = ev.key.sym;

				if ((KTYP(U(key)) == KT_LATIN) ||
				    (KTYP(U(key)) == KT_LETTER))

				switch (KVAL(U(key)))
				{
					case '\033':  /* Escape */
					case 'q':
					case 'Q':
					case 'x':
					case 'X':
						done=1;
						break;
				}
			}

			if (done) {
				break;
			}
                }

                /* determine time lapse */

                gettimeofday(&cur_time, NULL);

                millis = (cur_time.tv_sec  - prev_time.tv_sec)  * 1000
                       + (cur_time.tv_usec - prev_time.tv_usec) / 1000;

                prev_time = cur_time;


                memset(GI.page, 0, GI.pagelen);

                update_all_textures(millis);

                update_frame();

                
                for (gen_millis += millis; 
		     gen_millis >  GI.gen_time;
                     gen_millis -= GI.gen_time) {

                        add_texture(random() % GI.screen_width,
                                    random() % GI.screen_height,
                                    random() % 256);
                }
        }

        free_textures();

        ggiClose(GI.vis);
        ggiExit();

        exit(0);
}


/* Image of GGI */

static char *banner[46] = {

"..................###########....................................###########...................................",
"..............###################...##.......................###################...##.......###################",
"............########.......###########.....................########.......###########...........###########....",
"..........#######.............########...................#######.............########............#########.....",
".........######................#######..................######................#######.............#######......",
"........######...................#####.................######...................#####.............#######......",
".......######.....................#####...............######.....................#####............#######......",
"......######......................#####..............######......................#####............#######......",
".....#######.......................####.............#######.......................####............#######......",
"....#######.........................###............#######.........................###............#######......",
"....######..........................###............######..........................###............#######......",
"...#######..........................###...........#######..........................###............#######......",
"...#######...........................##...........#######...........................##............#######......",
"..#######........................................#######..........................................#######......",
"..#######........................................#######..........................................#######......",
".########.......................................########..........................................#######......",
".#######........................................#######...........................................#######......",
".#######........................................#######...........................................#######......",
".#######........................................#######...........................................#######......",
"########.......................................########...........................................#######......",
"########.......................................########...........................................#######......",
"########..................###################..########..................###################......#######......",
"########......................###########......########......................###########..........#######......",
"########.......................#########.......########.......................#########...........#######......",
"########........................########.......########........................########...........#######......",
"########........................#######........########........................#######............#######......",
"########........................#######........########........................#######............#######......",
"#########.......................#######........#########.......................#######............#######......",
".########.......................#######.........########.......................#######............#######......",
".########.......................#######.........########.......................#######............#######......",
".########.......................#######.........########.......................#######............#######......",
"..########......................#######..........########......................#######............#######......",
"..########......................#######..........########......................#######............#######......",
"..#########.....................#######..........#########.....................#######............#######......",
"...########.....................#######...........########.....................#######............#######......",
"....########....................#######............########....................#######............#######......",
"....########....................#######............########....................#######............#######......",
".....########...................#######.............########...................#######............#######......",
"......########..................#######..............########..................#######............#######......",
".......########.................#######...............########.................#######............#######......",
"........########................#######................########................#######............#######......",
"..........########............#########..................########............#########...........#########.....",
"............#########......##########......................#########......##########............###########....",
"..............####################...........................####################...........###################",
"..................############...................................############..................................",

NULL
};

