/* 
   Graphics library for GGI. Events for X.

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#ifdef _AIX
#include <strings.h>
#include <sys/select.h>
#endif

#include "ggi-dl.h"
#include "Xvisual.h"
#include "../common/evqueue.inc"

#include <X11/cursorfont.h>

static uint32 translate_key(KeySym sym)
{
	switch(sym) { 
		case XK_VoidSymbol: return U(K_VOID);
		case XK_BackSpace : return U(8);
		case XK_Tab       : return U(9);
		case XK_Linefeed  : return U(K_ENTER);
		case XK_Clear	  : return U(11); /* ??? */
		case XK_Return    : return U(K_ENTER);
		case XK_Pause	  : break; /* ??? */
		case XK_Scroll_Lock: break;
		case XK_Sys_Req   : break;
		case XK_Escape  : return U(27);
		case XK_Delete	: return U(127);

		/* Japanese Language shifts not yet */

		/* Cursor control & motion */
		case XK_Home	: break; /* ??? */
		case XK_Left    : return U(K_LEFT);
		case XK_Up      : return U(K_UP);
		case XK_Right   : return U(K_RIGHT);
		case XK_Down    : return U(K_DOWN);
		case XK_Prior	: return U(K_PGUP);
		case XK_Next	: return U(K_PGDN);
		case XK_End	: break;
		case XK_Begin	: break;

		/* Misc Functions */
		case XK_Select	: return U(K_SELECT);
		case XK_Print	: break;
		case XK_Execute	: break;
		case XK_Insert	: return U(K_INSERT);
#ifdef K_UNDO
		case XK_Undo	: return U(K_UNDO);
#endif
		case XK_Redo	: break;
		case XK_Menu	: break;
		case XK_Find	: break;
		case XK_Cancel	: break;
		case XK_Help	: return U(K_HELP);
		case XK_Break	: break;
		case XK_Mode_switch : break;
		case XK_Num_Lock: break;

		/* Keypad Functions */

		case XK_KP_Space: break;
		case XK_KP_Tab	: break;
		case XK_KP_Enter: return U(K_PENTER);
		case XK_KP_F1	: break;
		case XK_KP_F2	: break;
		case XK_KP_F3	: break;
		case XK_KP_F4	: break;
		case XK_KP_Home	: return U(K_P7);
		case XK_KP_Left : return U(K_P4);
		case XK_KP_Up	: return U(K_P8);
		case XK_KP_Right: return U(K_P6);
		case XK_KP_Down : return U(K_P2);
		case XK_KP_Prior: return U(K_P9);
		case XK_KP_Next	: return U(K_P3);
		case XK_KP_End	: return U(K_P1);
		case XK_KP_Begin: break;
		case XK_KP_Insert: return U(K_P0);
		case XK_KP_Delete: return U(K_PDOT);
		case XK_KP_Equal: break;
		case XK_KP_Multiply: return U(K_PSTAR);
		case XK_KP_Add	: return U(K_PPLUS);
		case XK_KP_Separator: break;
		case XK_KP_Subtract: return U(K_PMINUS);
		case XK_KP_Decimal: return U(K_PDOT);
		case XK_KP_Divide: return U(K_PSLASH);

		case XK_KP_0	: return U(K_P0);
		case XK_KP_1	: return U(K_P1);
		case XK_KP_2	: return U(K_P2);
		case XK_KP_3	: return U(K_P3);
		case XK_KP_4	: return U(K_P4);
		case XK_KP_5	: return U(K_P5);
		case XK_KP_6	: return U(K_P6);
		case XK_KP_7	: return U(K_P7);
		case XK_KP_8	: return U(K_P8);
		case XK_KP_9	: return U(K_P9);

		/* function keys */
		case XK_F1	: return U(K_F1);
		case XK_F2	: return U(K_F2);
		case XK_F3	: return U(K_F3);
		case XK_F4	: return U(K_F4);
		case XK_F5	: return U(K_F5);
		case XK_F6	: return U(K_F6);
		case XK_F7	: return U(K_F7);
		case XK_F8	: return U(K_F8);
		case XK_F9	: return U(K_F9);
		case XK_F10	: return U(K_F10);
		case XK_F11	: return U(K_F11);
		case XK_F12	: return U(K_F12);
		case XK_F13	: return U(K_F13);
		case XK_F14	: return U(K_F14);
		case XK_F15	: return U(K_F15);
		case XK_F16	: return U(K_F16);
		case XK_F17	: return U(K_F17);
		case XK_F18	: return U(K_F18);
		case XK_F19	: return U(K_F19);
		case XK_F20	: return U(K_F20);
		case XK_F21	: return U(K_F21);
		case XK_F22	: return U(K_F22);
		case XK_F23	: return U(K_F23);
		case XK_F24	: return U(K_F24);
		case XK_F25	: return U(K_F25);
		case XK_F26	: return U(K_F26);
		case XK_F27	: return U(K_F27);
		case XK_F28	: return U(K_F28);
		case XK_F29	: return U(K_F29);
		case XK_F30	: return U(K_F30);
		case XK_F31	: return U(K_F31);
		case XK_F32	: return U(K_F32);
		case XK_F33	: return U(K_F33);
		case XK_F34	: return U(K_F34);
		case XK_F35	: return U(K_F35);
		
		/* Modifiers */

		case XK_Shift_L: return U(K_NORMAL_SHIFTL);
		case XK_Shift_R: return U(K_NORMAL_SHIFTR);
		case XK_Control_L: return U(K_NORMAL_CTRLL);
		case XK_Control_R: return U(K_NORMAL_CTRLR);
		case XK_Caps_Lock: return U(K_STICKY_SHIFT);
		case XK_Shift_Lock: return U(K_LOCKED_SHIFT);

		case XK_Meta_L: return U(K_NORMAL_ALT);
		case XK_Meta_R: return U(K_NORMAL_ALTGR);
		case XK_Alt_L: return U(K_NORMAL_ALT);
		case XK_Alt_R: return U(K_NORMAL_ALTGR);
		case XK_Super_L: break;
		case XK_Super_R: break;
		case XK_Hyper_L: break;
		case XK_Hyper_R: break;

		/* ISO 9995 Function and Modifier Keys missing */
		/* 3270 Terminal Keys missing */

#ifdef XK_LATIN2
		case XK_Aogonek		: return U(0x0104);
		case XK_breve		: return U(0x02D8);
		case XK_Lstroke		: return U(0x0141);
		case XK_Lcaron		: return U(0x013D);
		case XK_Sacute		: return U(0x015A);
		case XK_Scaron		: return U(0x0160);
		case XK_Scedilla	: return U(0x015E);
		case XK_Tcaron		: return U(0x0164);
		case XK_Zacute		: return U(0x0179);
		case XK_Zcaron		: return U(0x017D);
		case XK_Zabovedot	: return U(0x017B);
		case XK_aogonek		: return U(0x0105);
		case XK_ogonek		: return U(0x02DB);
		case XK_lstroke		: return U(0x0142);
		case XK_lcaron		: return U(0x013E);
		case XK_sacute		: return U(0x015B);
		case XK_caron		: return U(0x02C7);
		case XK_scaron		: return U(0x0161);
		case XK_scedilla	: return U(0x015F);
		case XK_tcaron		: return U(0x0165);
		case XK_zacute		: return U(0x017A);
		case XK_doubleacute	: return U(0x02DD);
		case XK_zcaron		: return U(0x017E);
		case XK_zabovedot	: return U(0x017C);
		case XK_Racute		: return U(0x0154);
		case XK_Abreve		: return U(0x0102);
		case XK_Lacute		: return U(0x0139);
		case XK_Cacute		: return U(0x0106);
		case XK_Ccaron		: return U(0x010C);
		case XK_Eogonek		: return U(0x0118);
		case XK_Ecaron		: return U(0x011A);
		case XK_Dcaron		: return U(0x010E);
		case XK_Dstroke		: return U(0x0110);
		case XK_Nacute		: return U(0x0143);
		case XK_Ncaron		: return U(0x0147);
		case XK_Odoubleacute	: return U(0x0150);
		case XK_Rcaron		: return U(0x0158);
		case XK_Uring		: return U(0x016E);
		case XK_Udoubleacute	: return U(0x0170);
		case XK_Tcedilla	: return U(0x0162);
		case XK_racute		: return U(0x0155);
		case XK_abreve		: return U(0x0103);
		case XK_lacute		: return U(0x013A);
		case XK_cacute		: return U(0x0107);
		case XK_ccaron		: return U(0x010D);
		case XK_eogonek		: return U(0x0119);
		case XK_ecaron		: return U(0x011B);
		case XK_dcaron		: return U(0x010F);
		case XK_dstroke		: return U(0x0111);
		case XK_nacute		: return U(0x0144);
		case XK_ncaron		: return U(0x0148);
		case XK_odoubleacute	: return U(0x0151);
		case XK_udoubleacute	: return U(0x0171);
		case XK_rcaron		: return U(0x0159);
		case XK_uring		: return U(0x016F);
		case XK_tcedilla	: return U(0x0163);
		case XK_abovedot	: return U(0x02D9);
#endif /* XK_LATIN2 */

#ifdef XK_LATIN3
		case XK_Hstroke		: return U(0x0126);
		case XK_Hcircumflex	: return U(0x0124);
		case XK_Iabovedot	: return U(0x0130);
		case XK_Gbreve		: return U(0x011E);
		case XK_Jcircumflex	: return U(0x0134);
		case XK_hstroke		: return U(0x0127);
		case XK_hcircumflex	: return U(0x0125);
		case XK_idotless	: return U(0x0131);
		case XK_gbreve		: return U(0x011F);
		case XK_jcircumflex	: return U(0x0135);
		case XK_Cabovedot	: return U(0x010A);
		case XK_Ccircumflex	: return U(0x0108);
		case XK_Gabovedot	: return U(0x0120);
		case XK_Gcircumflex	: return U(0x011C);
		case XK_Ubreve		: return U(0x016C);
		case XK_Scircumflex	: return U(0x015C);
		case XK_cabovedot	: return U(0x010B);
		case XK_ccircumflex	: return U(0x0109);
		case XK_gabovedot	: return U(0x0121);
		case XK_gcircumflex	: return U(0x011D);
		case XK_ubreve		: return U(0x016D);
		case XK_scircumflex	: return U(0x015D);
#endif /* XK_LATIN3 */

#ifdef XK_LATIN4
		case XK_kra		: return U(0x0138);
		case XK_Rcedilla	: return U(0x0156);
		case XK_Itilde		: return U(0x0128);
		case XK_Lcedilla	: return U(0x013B);
		case XK_Emacron		: return U(0x0112);
		case XK_Gcedilla	: return U(0x0122);
		case XK_Tslash		: return U(0x0166);
		case XK_rcedilla	: return U(0x0157);
		case XK_itilde		: return U(0x0129);
		case XK_lcedilla	: return U(0x013C);
		case XK_emacron		: return U(0x0113);
		case XK_gcedilla	: return U(0x0123);
		case XK_tslash		: return U(0x0167);
		case XK_ENG		: return U(0x014A);
		case XK_eng		: return U(0x014B);
		case XK_Amacron		: return U(0x0100);
		case XK_Iogonek		: return U(0x012E);
		case XK_Eabovedot	: return U(0x0116);
		case XK_Imacron		: return U(0x012A);
		case XK_Ncedilla	: return U(0x0145);
		case XK_Omacron		: return U(0x014C);
		case XK_Kcedilla	: return U(0x0136);
		case XK_Uogonek		: return U(0x0172);
		case XK_Utilde		: return U(0x0168);
		case XK_Umacron		: return U(0x016A);
		case XK_amacron		: return U(0x0101);
		case XK_iogonek		: return U(0x012F);
		case XK_eabovedot	: return U(0x0117);
		case XK_imacron		: return U(0x012B);
		case XK_ncedilla	: return U(0x0146);
		case XK_omacron		: return U(0x014D);
		case XK_kcedilla	: return U(0x0137);
		case XK_uogonek		: return U(0x0173);
		case XK_utilde		: return U(0x0169);
		case XK_umacron		: return U(0x016B);
#endif /* XK_LATIN4 */

#ifdef XK_KATAKANA
		case XK_overline		: break;
		case XK_kana_fullstop		: break;
		case XK_kana_openingbracket	: break;
		case XK_kana_closingbracket	: break;
		case XK_kana_comma		: break;
		case XK_kana_conjunctive	: return U(0x30FB);
		case XK_kana_WO			: return U(0x30F2);
		case XK_kana_a			: return U(0x30A1);
		case XK_kana_i			: return U(0x30A3);
		case XK_kana_u			: return U(0x30A5);
		case XK_kana_e			: return U(0x30A7);
		case XK_kana_o			: return U(0x30A9);
		case XK_kana_ya			: return U(0x30E3);
		case XK_kana_yu			: return U(0x30E5);
		case XK_kana_yo			: return U(0x30E7);
		case XK_kana_tsu		: return U(0x30C3);
		case XK_prolongedsound		: return U(0x30FC);
		case XK_kana_A			: return U(0x30A2);
		case XK_kana_I			: return U(0x30A4);
		case XK_kana_U			: return U(0x30A6);
		case XK_kana_E			: return U(0x30A8);
		case XK_kana_O			: return U(0x30AA);
		case XK_kana_KA			: return U(0x30AB);
		case XK_kana_KI			: return U(0x30AD);
		case XK_kana_KU			: return U(0x30AF);
		case XK_kana_KE			: return U(0x30B1);
		case XK_kana_KO			: return U(0x30B3);
		case XK_kana_SA			: return U(0x30B5);
		case XK_kana_SHI		: break;
		case XK_kana_SU			: return U(0x30B9);
		case XK_kana_SE			: return U(0x30BB);
		case XK_kana_SO			: return U(0x30BD);
		case XK_kana_TA			: return U(0x30BF);
		case XK_kana_CHI		: return U(0x30C1);
		case XK_kana_TSU		: return U(0x30C4);
		case XK_kana_TE			: return U(0x30C6);
		case XK_kana_TO			: return U(0x30C8);
		case XK_kana_NA			: return U(0x30CA);
		case XK_kana_NI			: return U(0x30CB);
		case XK_kana_NU			: return U(0x30CC);
		case XK_kana_NE			: return U(0x30CD);
		case XK_kana_NO			: return U(0x30CE);
		case XK_kana_HA			: return U(0x30CF);
		case XK_kana_HI			: return U(0x30D2);
		case XK_kana_FU			: return U(0x30D5);
		case XK_kana_HE			: return U(0x30D8);
		case XK_kana_HO			: return U(0x30DB);
		case XK_kana_MA			: return U(0x30DE);
		case XK_kana_MI			: return U(0x30DF);
		case XK_kana_MU			: return U(0x30E0);
		case XK_kana_ME			: return U(0x30E1);
		case XK_kana_MO			: return U(0x30E2);
		case XK_kana_YA			: return U(0x30E4);
		case XK_kana_YU			: return U(0x30E6);
		case XK_kana_YO			: return U(0x30E8);
		case XK_kana_RA			: return U(0x30E9);
		case XK_kana_RI			: return U(0x30EA);
		case XK_kana_RU			: return U(0x30EB);
		case XK_kana_RE			: return U(0x30EC);
		case XK_kana_RO			: return U(0x30ED);
		case XK_kana_WA			: return U(0x30EF);
		case XK_kana_N			: return U(0x30F3);
		case XK_voicedsound		: return U(0x309B);
		case XK_semivoicedsound		: return U(0x309C);
		/* case XK_kana_switch          0xFF7E  Alias for mode_switch */
#endif /* XK_KATAKANA */

#ifdef XK_ARABIC
		case XK_Arabic_comma		: return U(0x060C);
		case XK_Arabic_semicolon	: return U(0x061B);
		case XK_Arabic_question_mark	: return U(0x061F);
		case XK_Arabic_hamza		: return U(0x0621);
		case XK_Arabic_maddaonalef	: return U(0x0622);
		case XK_Arabic_hamzaonalef	: return U(0x0623);
		case XK_Arabic_hamzaonwaw	: return U(0x0624);
		case XK_Arabic_hamzaunderalef	: return U(0x0625);
		case XK_Arabic_hamzaonyeh	: return U(0x0626);
		case XK_Arabic_alef		: return U(0x0627);
		case XK_Arabic_beh		: return U(0x0628);
		case XK_Arabic_tehmarbuta	: return U(0x0629);
		case XK_Arabic_teh		: return U(0x062a);
		case XK_Arabic_theh		: return U(0x062b);
		case XK_Arabic_jeem		: return U(0x062c);
		case XK_Arabic_hah		: return U(0x062d);
		case XK_Arabic_khah		: return U(0x062e);
		case XK_Arabic_dal		: return U(0x062f);
		case XK_Arabic_thal		: return U(0x0630);
		case XK_Arabic_ra		: return U(0x0631);
		case XK_Arabic_zain		: return U(0x0632);
		case XK_Arabic_seen		: return U(0x0633);
		case XK_Arabic_sheen		: return U(0x0634);
		case XK_Arabic_sad		: return U(0x0635);
		case XK_Arabic_dad		: return U(0x0636);
		case XK_Arabic_tah		: return U(0x0637);
		case XK_Arabic_zah		: return U(0x0638);
		case XK_Arabic_ain		: return U(0x0639);
		case XK_Arabic_ghain		: return U(0x063A);
		case XK_Arabic_tatweel		: return U(0x0640);
		case XK_Arabic_feh		: return U(0x0641);
		case XK_Arabic_qaf		: return U(0x0642);
		case XK_Arabic_kaf		: return U(0x0643);
		case XK_Arabic_lam		: return U(0x0644);
		case XK_Arabic_meem		: return U(0x0645);
		case XK_Arabic_noon		: return U(0x0646);
		case XK_Arabic_ha		: return U(0x0647);
		case XK_Arabic_waw		: return U(0x0648);
		case XK_Arabic_alefmaksura	: return U(0x0649);
		case XK_Arabic_yeh		: return U(0x064A);
		case XK_Arabic_fathatan		: return U(0x064B);
		case XK_Arabic_dammatan		: return U(0x064C);
		case XK_Arabic_kasratan		: return U(0x064D);
		case XK_Arabic_fatha		: return U(0x064E);
		case XK_Arabic_damma		: return U(0x064F);
		case XK_Arabic_kasra		: return U(0x0650);
		case XK_Arabic_shadda		: return U(0x0651);
		case XK_Arabic_sukun		: return U(0x0652);
		/* case XK_Arabic_switch        0xFF7E  Alias for mode_switch */
#endif /* XK_ARABIC */

#ifdef XK_CYRILLIC
		case XK_Serbian_dje	: return U(0x0452);
		case XK_Macedonia_gje	: return U(0x0453);
		case XK_Cyrillic_io	: return U(0x0451);
		case XK_Ukrainian_ie	: return U(0x0458);
		case XK_Macedonia_dse	: return U(0x0455);
		case XK_Ukrainian_i	: return U(0x0456);
		case XK_Ukrainian_yi	: return U(0x0457);
		case XK_Cyrillic_je	: return U(0x0454);
		case XK_Cyrillic_lje	: return U(0x0459);
		case XK_Cyrillic_nje	: return U(0x045A);
		case XK_Serbian_tshe	: return U(0x045B);
		case XK_Macedonia_kje	: return U(0x045C);
		case XK_Byelorussian_shortu : return U(0x045E);
		case XK_Cyrillic_dzhe	: return U(0x045F);
		case XK_numerosign	: return U(0x2116);
		case XK_Serbian_DJE	: return U(0x0402);
		case XK_Macedonia_GJE	: return U(0x0403);
		case XK_Cyrillic_IO	: return U(0x042E);
		case XK_Ukrainian_IE	: return U(0x0404);
		case XK_Macedonia_DSE	: return U(0x0405);
		case XK_Ukrainian_I	: return U(0x0406);
		case XK_Ukrainian_YI	: return U(0x0407);
		case XK_Cyrillic_JE	: return U(0x0408);
		case XK_Cyrillic_LJE	: return U(0x0409);
		case XK_Cyrillic_NJE	: return U(0x040A);
		case XK_Serbian_TSHE	: return U(0x040B);
		case XK_Macedonia_KJE	: return U(0x040C);
		case XK_Byelorussian_SHORTU	: return U(0x040E);
		case XK_Cyrillic_DZHE	: return U(0x040F);
		case XK_Cyrillic_yu	: return U(0x044E);
		case XK_Cyrillic_a	: return U(0x0430);
		case XK_Cyrillic_be	: return U(0x0431);
		case XK_Cyrillic_tse	: return U(0x0446);
		case XK_Cyrillic_de	: return U(0x0434);
		case XK_Cyrillic_ie	: return U(0x0435);
		case XK_Cyrillic_ef	: return U(0x0444);
		case XK_Cyrillic_ghe	: return U(0x0433);
		case XK_Cyrillic_ha	: return U(0x0445);
		case XK_Cyrillic_i	: return U(0x0438);
		case XK_Cyrillic_shorti	: return U(0x0439);
		case XK_Cyrillic_ka	: return U(0x043A);
		case XK_Cyrillic_el	: return U(0x043B);
		case XK_Cyrillic_em	: return U(0x043C);
		case XK_Cyrillic_en	: return U(0x043D);
		case XK_Cyrillic_o	: return U(0x043E);
		case XK_Cyrillic_pe	: return U(0x043F);
		case XK_Cyrillic_ya	: return U(0x044F);
		case XK_Cyrillic_er	: return U(0x0440);
		case XK_Cyrillic_es	: return U(0x0441);
		case XK_Cyrillic_te	: return U(0x0442);
		case XK_Cyrillic_u	: return U(0x0443);
		case XK_Cyrillic_zhe	: return U(0x0436);
		case XK_Cyrillic_ve	: return U(0x0432);
		case XK_Cyrillic_softsign	: return U(0x044C);
		case XK_Cyrillic_yeru	: return U(0x044B);
		case XK_Cyrillic_ze	: return U(0x0437);
		case XK_Cyrillic_sha	: return U(0x0448);
		case XK_Cyrillic_e	: return U(0x044D);
		case XK_Cyrillic_shcha	: return U(0x0449);
		case XK_Cyrillic_che	: return U(0x0447);
		case XK_Cyrillic_hardsign	: return U(0x044A);

		case XK_Cyrillic_YU	: return U(0x0401);
		case XK_Cyrillic_A	: return U(0x0410);
		case XK_Cyrillic_BE	: return U(0x0411);
		case XK_Cyrillic_TSE	: return U(0x0426);
		case XK_Cyrillic_DE	: return U(0x0414);
		case XK_Cyrillic_IE	: return U(0x0415);
		case XK_Cyrillic_EF	: return U(0x0424);
		case XK_Cyrillic_GHE	: return U(0x0413);
		case XK_Cyrillic_HA	: return U(0x0425);
		case XK_Cyrillic_I	: return U(0x0418);
		case XK_Cyrillic_SHORTI	: return U(0x0419);
		case XK_Cyrillic_KA	: return U(0x041A);
		case XK_Cyrillic_EL	: return U(0x041B);
		case XK_Cyrillic_EM	: return U(0x041C);
		case XK_Cyrillic_EN	: return U(0x041D);
		case XK_Cyrillic_O	: return U(0x041E);
		case XK_Cyrillic_PE	: return U(0x041F);
		case XK_Cyrillic_YA	: return U(0x042F);
		case XK_Cyrillic_ER	: return U(0x0420);
		case XK_Cyrillic_ES	: return U(0x0421);
		case XK_Cyrillic_TE	: return U(0x0422);
		case XK_Cyrillic_U	: return U(0x0423);
		case XK_Cyrillic_ZHE	: return U(0x0416);
		case XK_Cyrillic_VE	: return U(0x0412);
		case XK_Cyrillic_SOFTSIGN	: return U(0x042C);
		case XK_Cyrillic_YERU	: return U(0x042B);
		case XK_Cyrillic_ZE	: return U(0x0417);
		case XK_Cyrillic_SHA	: return U(0x0428);
		case XK_Cyrillic_E	: return U(0x042D);
		case XK_Cyrillic_SHCHA	: return U(0x0429);
		case XK_Cyrillic_CHE	: return U(0x0427);
		case XK_Cyrillic_HARDSIGN	: return U(0x042A);
#endif

		/* Technical, Special missing*/

		case XK_Greek_ALPHAaccent	: return U(0x0386);
		case XK_Greek_EPSILONaccent	: return U(0x0388);
		case XK_Greek_ETAaccent		: return U(0x0389);
		case XK_Greek_IOTAaccent	: return U(0x038A);
		case XK_Greek_IOTAdiaeresis	: return U(0x03AA);
		case XK_Greek_OMICRONaccent	: return U(0x038C);
		case XK_Greek_UPSILONaccent	: return U(0x038E);
		case XK_Greek_UPSILONdieresis	: return U(0x03AB);
		case XK_Greek_OMEGAaccent	: return U(0x038F);
		case XK_Greek_accentdieresis	: return U(0x0385);
		case XK_Greek_horizbar		: return U(0x2015);
		case XK_Greek_alphaaccent	: return U(0x03AC);
		case XK_Greek_epsilonaccent	: return U(0x03AD);
		case XK_Greek_etaaccent		: return U(0x03AE);
		case XK_Greek_iotaaccent	: return U(0x03AF);
		case XK_Greek_iotadieresis	: return U(0x03CA);
		case XK_Greek_iotaaccentdieresis: return U(0x0390);
		case XK_Greek_omicronaccent	: return U(0x03CC);
		case XK_Greek_upsilonaccent	: return U(0x03CD);
		case XK_Greek_upsilondieresis	: return U(0x03CB);
		case XK_Greek_upsilonaccentdieresis: return U(0x03B0);
		case XK_Greek_omegaaccent	: return U(0x03CE);
		case XK_Greek_ALPHA		: return U(0x0391);
		case XK_Greek_BETA		: return U(0x0392);
		case XK_Greek_GAMMA		: return U(0x0393);
		case XK_Greek_DELTA		: return U(0x0394);
		case XK_Greek_EPSILON		: return U(0x0395);
		case XK_Greek_ZETA		: return U(0x0396);
		case XK_Greek_ETA		: return U(0x0397);
		case XK_Greek_THETA		: return U(0x0398);
		case XK_Greek_IOTA		: return U(0x0399);
		case XK_Greek_KAPPA		: return U(0x039A);
		case XK_Greek_LAMBDA		: return U(0x039B);
		case XK_Greek_MU		: return U(0x039C);
		case XK_Greek_NU		: return U(0x039D);
		case XK_Greek_XI		: return U(0x039E);
		case XK_Greek_OMICRON		: return U(0x039F);
		case XK_Greek_PI		: return U(0x03A0);
		case XK_Greek_RHO		: return U(0x03A1);
		case XK_Greek_SIGMA		: return U(0x03A3);
		case XK_Greek_TAU		: return U(0x03A4);
		case XK_Greek_UPSILON		: return U(0x03A5);
		case XK_Greek_PHI		: return U(0x03A6);
		case XK_Greek_CHI		: return U(0x03A7);
		case XK_Greek_PSI		: return U(0x03A8);
		case XK_Greek_OMEGA		: return U(0x03A9);
		case XK_Greek_alpha		: return U(0x03B1);
		case XK_Greek_beta		: return U(0x03B2);
		case XK_Greek_gamma		: return U(0x03B3);
		case XK_Greek_delta		: return U(0x03B4);
		case XK_Greek_epsilon		: return U(0x03B5);
		case XK_Greek_zeta		: return U(0x03B6);
		case XK_Greek_eta		: return U(0x03B7);
		case XK_Greek_theta		: return U(0x03B8);
		case XK_Greek_iota		: return U(0x03B9);
		case XK_Greek_kappa		: return U(0x03BA);
		case XK_Greek_lambda		: return U(0x03BB);
		case XK_Greek_mu		: return U(0x03BC);
		case XK_Greek_nu		: return U(0x03BD);
		case XK_Greek_xi		: return U(0x03BE);
		case XK_Greek_omicron		: return U(0x03BF);
		case XK_Greek_pi		: return U(0x03C0);
		case XK_Greek_rho		: return U(0x03C1);
		case XK_Greek_sigma		: return U(0x03C3);
		case XK_Greek_finalsmallsigma	: return U(0x03C2);
		case XK_Greek_tau		: return U(0x03C4);
		case XK_Greek_upsilon		: return U(0x03C5);
		case XK_Greek_phi		: return U(0x03C6);
		case XK_Greek_chi		: return U(0x03C7);
		case XK_Greek_psi		: return U(0x03C8);
		case XK_Greek_omega		: return U(0x03C9);
		/* case XK_Greek_switch         0xFF7E  Alias for mode_switch */

		/* Publishing, APL missing*/

#ifdef XK_HEBREW
		case XK_hebrew_doublelowline: return U(0x2017);
		case XK_hebrew_aleph	: return U(0x05D0);
		case XK_hebrew_bet     	: return U(0x05D1);
		case XK_hebrew_gimel	: return U(0x05D2);
		case XK_hebrew_dalet	: return U(0x05D3);
		case XK_hebrew_he	: return U(0x05D4);
		case XK_hebrew_waw	: return U(0x05D5);
		case XK_hebrew_zain	: return U(0x05D6);
		case XK_hebrew_chet	: return U(0x05D7);
		case XK_hebrew_tet	: return U(0x05D8);
		case XK_hebrew_yod	: return U(0x05D9);
		case XK_hebrew_finalkaph: return U(0x05DA);
		case XK_hebrew_kaph	: return U(0x05DB);
		case XK_hebrew_lamed	: return U(0x05DC);
		case XK_hebrew_finalmem	: return U(0x05DD);
		case XK_hebrew_mem	: return U(0x05DE);
		case XK_hebrew_finalnun	: return U(0x05DF);
		case XK_hebrew_nun	: return U(0x05E0);
		case XK_hebrew_samech	: return U(0x05E1);
		case XK_hebrew_ayin	: return U(0x05E2);
		case XK_hebrew_finalpe	: return U(0x05E3);
		case XK_hebrew_pe	: return U(0x05E4);
		case XK_hebrew_finalzade: return U(0x05E5);
		case XK_hebrew_zade	: return U(0x05E6);
		case XK_hebrew_qoph	: return U(0x05E7);
		case XK_hebrew_resh	: return U(0x05E8);
		case XK_hebrew_shin	: return U(0x05E9);
		case XK_hebrew_taw	: return U(0x05EA);
		/* case XK_Hebrew_switch        0xFF7E  Alias for mode_switch */
#endif
		/* Thai, Korean missing*/


		case XK_space   : return U(' ');
	}
	if ( (sym & 0xff00) == 0 ) /* Latin 1 - map through */
		return U(sym&0xff);
	return 0;
}

static uint32 translate_button(unsigned int button)
{
	switch(button) {
		case Button1:
			return 1;
		case Button2:
			return 4;
		case Button3:
			return 2;
		case Button4:
			return 8;
		case Button5:
			return 16;
	}
	return 0;
}

/************************** GGI Functions ***************************/
/* Event Handling */
ggi_event_mask GGIeventpoll(ggi_visual_t vis,ggi_event_mask mask,
			    struct timeval *t)
{
#define BUFFER_SIZE (sizeof(ggi_event)*4)
	
	
#define XlibWaitLock	while(xhook->XLibLock++!=0) \
	{xhook->XLibLock--;usleep(1000);}     /* FIXME use spinlocks here !*/
#define	XlibReleaseLock	xhook->XLibLock--
	
	KeySym relsyms[RELPTR_NUMKEYS] = RELPTR_KEYS;
	fd_set fds;
	int err,n;
	struct timeval tod,tnew;
	ggi_event_mask evmask;
	ggi_event ggievent;
	struct Xhooks *xhook=VIS_PRIVATE(vis);

DPRINT("GGIeventpoll(%p,0x%.8x",vis,mask);
if (t==NULL) {
	DPRINT(",NULL)\n");
} else {
	DPRINT(",{%d,%d})\n",t->tv_sec,t->tv_usec);
}

	evmask=_ggiEvQueueSeen(vis,mask);
	if (evmask!=0)  
		return evmask;

	gettimeofday(&tod,NULL);

	do {
		retry:	/* sorry .. I know ... */
		DPRINT("retry:");

		XlibWaitLock;
		n=XEventsQueued(xhook->display,QueuedAfterReading);
		XlibReleaseLock;

		if (! n ) {	/* No event read. Use select(); */

				if (t && ( t->tv_sec<0 ||
				           ( t->tv_sec==0 && t->tv_usec==0) ) )
					break;	/* Fast poll. */
				
				FD_ZERO(&fds);
				FD_SET(VIS_SELECT_FD(vis),&fds);
				err=select(VIS_SELECT_FD(vis)+1,&fds,NULL,NULL,t);

				/* Now update the time remaining. */
				if (t) {
					DPRINT("t before: %ld,%ld",t->tv_sec,t->tv_usec);
					gettimeofday(&tnew,NULL);
					tod.tv_sec-=tnew.tv_sec;tod.tv_usec-=tnew.tv_usec;
					if (tod.tv_usec>0) {tod.tv_usec-=1000000;tod.tv_sec++;}
					t->tv_sec +=tod.tv_sec ;t->tv_usec +=tod.tv_usec ;
					if (t->tv_usec<0) {t->tv_usec+=1000000;t->tv_sec--;}
					tod=tnew;
					DPRINT("t after: %ld,%ld",t->tv_sec,t->tv_usec);
				}
				
				if (err<0)	/* Something bad happened. Break out. */
				{	if (errno==EINTR)	/* Except if it is only the itimer ... */
						goto retry;
					else
						break;
				}
				if (FD_ISSET(VIS_SELECT_FD(vis),&fds)) 
					goto retry;	/* retry ... otherwise it was a timeout. */
		}

		while(n--) {
 			XEvent event;

			XlibWaitLock;
			XNextEvent(xhook->display,&event);
			XlibReleaseLock;

			ggievent.any.focus=0;
			ggievent.any.device=0;
			ggievent.any.time=0;	/* FIXME ! */
			ggievent.any.type=0;

			switch(event.type) { 
				case KeyPress:
				{
					int i;
					KeySym sym=XLookupKeysym(&event.xkey,0);
					DPRINT("KeyPress.\n");
					for (i=0; i < RELPTR_NUMKEYS; i++) {
						if (sym == relsyms[i]) {
							xhook->relptr_keymask &= ~(1<<i);
							break;
						}
					}
					if (xhook->relptr_keymask == 0) {
						if (xhook->relptr) {
							XUngrabPointer(xhook->display, CurrentTime);
							XUndefineCursor(xhook->display,
											xhook->window);
							xhook->relptr=0;
							DPRINT("Using absolute pointerevents.\n");
						} else {
							char emptybm[] = {0};
							Pixmap crsrpix;
							XColor nocol;
							Cursor mycrsr;
							crsrpix=XCreateBitmapFromData(xhook->display,
													   xhook->window,
													   emptybm, 1, 1);
							mycrsr=XCreatePixmapCursor(xhook->display,
												crsrpix, crsrpix,
												&nocol, &nocol,
												0, 0);
							XFreePixmap(xhook->display, crsrpix);
							XGrabPointer(xhook->display, xhook->window, 1, 0,
										 GrabModeAsync,
										 GrabModeAsync,
										 xhook->window,
										 mycrsr,
										 CurrentTime);
							XWarpPointer(xhook->display, None,
										 xhook->window,
										 0, 0, 0, 0,
										 xhook->viswidth/2, xhook->visheight/2);
							xhook->relptr=1;
							xhook->oldmousex=xhook->viswidth/2;
							xhook->oldmousey=xhook->visheight/2;
							DPRINT("Using relative pointerevents.\n");
						}
					}
					ggievent.any.type=evKeyPress;
					XlibWaitLock;
					ggievent.key.sym=translate_key(sym);
					XlibReleaseLock;
					ggievent.key.code=event.xkey.keycode;
					break;
				}
				case KeyRelease: 
				{
					int i;
					KeySym sym=XLookupKeysym(&event.xkey,0);
					for (i=0; i < RELPTR_NUMKEYS; i++) {
						if (sym == relsyms[i]) {
							xhook->relptr_keymask |= (1<<i);
							break;
						}
					}
					ggievent.any.type=evKeyRelease;
					XlibWaitLock;
					ggievent.key.sym=translate_key(sym);
					XlibReleaseLock;
					ggievent.key.code=event.xkey.keycode;
					DPRINT("KeyRelease.\n"); break;
				}
				case ButtonPress:
					ggievent.any.type=evPtrButtonPress;
					ggievent.pbutton.state=event.xbutton.state;
					ggievent.pbutton.button=translate_button(event.xbutton.button);
					DPRINT("ButtonPress %x,%x.\n",event.xbutton.state,ggievent.pbutton.button); 
					break;
				case ButtonRelease:
					ggievent.any.type=evPtrButtonRelease;
					ggievent.pbutton.state=event.xbutton.state;
					ggievent.pbutton.button=translate_button(event.xbutton.button);
					DPRINT("ButtonRelease %x,%x.\n",event.xbutton.state,ggievent.pbutton.button); 
					break;
				case MotionNotify:
					if (xhook->relptr) {
						int realmove=0;
						if (!event.xmotion.send_event) {
							ggievent.pmove.x =
								event.xmotion.x-xhook->oldmousex;
							ggievent.pmove.y =
								event.xmotion.y-xhook->oldmousey;
							ggievent.any.type=evPtrRelative;
							realmove=1;
#undef ABS
#define ABS(a) (((a)<0) ? -(a) : (a) )
							if (ABS(xhook->viswidth/2 - event.xmotion.x)
								> 3 * xhook->viswidth / 8
								|| ABS(xhook->visheight/2 - event.xmotion.y)
								> 3 * xhook->visheight / 8) {
#undef ABS							
								XEvent event;
								event.type = MotionNotify;
								event.xmotion.display = xhook->display;
								event.xmotion.window = xhook->window;
								event.xmotion.x = xhook->viswidth/2;
								event.xmotion.y = xhook->visheight/2;
								XSendEvent(xhook->display, xhook->window, False, PointerMotionMask, &event);
								XWarpPointer(xhook->display, None,
											 xhook->window,
											 0, 0, 0, 0,
											 xhook->viswidth/2, xhook->visheight/2);
							}
						}
						xhook->oldmousex=event.xmotion.x;
						xhook->oldmousey=event.xmotion.y;
						if (!realmove)
							goto dont_queue_this_event;
					} else {
						ggievent.any.type=evPtrAbsolute;
						ggievent.pmove.x=event.xmotion.x;
						ggievent.pmove.y=event.xmotion.y;
					}
					DPRINT("MouseMove %d,%d.\n",
						   event.xmotion.x,event.xmotion.y); 
					break;
				default: 
					DPRINT("Other Event (%d).\n",event.type); break;
			}      
			switch(ggievent.any.type) {
				case evKeyPress:
				case evKeyRelease:
					ggievent.any.size=sizeof(ggi_key_event);
					ggievent.key.effect=0;
					ggievent.key.normal=0;
					ggievent.key.locked=0;
					ggievent.key.sticky=0;
					break;
				case evPtrButtonPress:
				case evPtrButtonRelease:
					ggievent.any.size=sizeof(ggi_pbutton_event);
					ggievent.pbutton.effect=0;
					ggievent.pbutton.normal=0;
					ggievent.pbutton.locked=0;
					ggievent.pbutton.sticky=0;
					break;
				case evPtrAbsolute:
					ggievent.any.size=sizeof(ggi_pmove_event);
					ggievent.pmove.effect=0;
					ggievent.pmove.normal=0;
					ggievent.pmove.locked=0;
					ggievent.pmove.sticky=0;
					break;
			}
			if (ggievent.any.type) _ggiEvQueueAdd(vis,&ggievent);
			evmask = _ggiEvQueueSeen(vis,mask);
			dont_queue_this_event:
		}
	} while (evmask==0);

	return evmask;
}

int GGIeventread(ggi_visual_t vis,ggi_event *ev,ggi_event_mask mask)
{
	/* Block if we don't have anything queued... */
	GGIeventpoll(vis,mask,NULL);
	return _ggiEvQueueRelease(vis,ev,mask);
}
