/* 
   Graphics library for GGI. Mode management.

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <sys/time.h>
#include <stdlib.h>
#include <unistd.h>

#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <signal.h>
#include <string.h>

#include "ggi-dl.h"
#include "Xvisual.h"
#include "../common/X-mode.inc"

void GGIgcchanged(ggi_visual *vis, int mask)
{
	if ((mask & GGI_GCCHANGED_CLIP)) {
		XRectangle xrect;
		xrect.x = VIS_GC(vis)->cliptl.x;
		xrect.y = VIS_GC(vis)->cliptl.y;
		xrect.width = VIS_GC(vis)->clipbr.x-VIS_GC(vis)->cliptl.x;
		xrect.height = VIS_GC(vis)->clipbr.y-VIS_GC(vis)->cliptl.y;
		XSetClipRectangles(XLIB_PRIV(vis)->display, XLIB_PRIV(vis)->gc,
						   0, 0, &xrect, 1, Unsorted);
		XSetClipRectangles(XLIB_PRIV(vis)->display, XLIB_PRIV(vis)->tempgc,
						   0, 0, &xrect, 1, Unsorted);
	}
	if ((mask & GGI_GCCHANGED_FG)) {
		XSetForeground(XLIB_PRIV(vis)->display, XLIB_PRIV(vis)->gc,
					   VIS_GC(vis)->fg_color);
	}
	if ((mask & GGI_GCCHANGED_BG)) {
		XSetBackground(XLIB_PRIV(vis)->display, XLIB_PRIV(vis)->gc,
					   VIS_GC(vis)->fg_color);
	}
}

static int _GGIdomode(ggi_visual *vis)
{
	int err;
	
	switch(VIS_MODE(vis)->graphtype)
		{
		case GT_8BIT : err=8;break;
		case GT_15BIT: err=15;break;
		case GT_16BIT: err=16;break;
		case GT_24BIT: err=24;break;
		case GT_32BIT: err=32;break;
		default: return -EINVAL;
		}

	_ggiZapMode(vis,~GGI_DL_OPDISPLAY);

	_ggiSetupMode(vis);

	err=(_ggiOpenDL(vis,"generic-stubs","")==NULL);
	if (err) {
		fprintf(stderr,"display-X: Can't open the \"generic-stubs\" "
				"library.\n");
	}

		/* Override the color operations for X if we are 8 bit palette. */
	if (XLIB_PRIV(vis)->colormap) { 
		free(XLIB_PRIV(vis)->colormap);XLIB_PRIV(vis)->colormap=NULL;
	}

	if (VIS_MODE(vis)->graphtype==GT_8BIT) {
		 
		int x;
		XLIB_PRIV(vis)->nocols=256;
		XLIB_PRIV(vis)->colormap=malloc(256*sizeof(ggi_color));
		for(x=0;x<XLIB_PRIV(vis)->nocols;x++) {
			XLIB_PRIV(vis)->colormap[x].r=((x&0xe0)>>5)*0x2492;
			XLIB_PRIV(vis)->colormap[x].g=((x&0x1c)>>2)*0x2492;
			XLIB_PRIV(vis)->colormap[x].b=((x&0x03)   )*0x5555;
		} 
		vis->opcolor->mapcolor=GGImapcolor;
		vis->opcolor->setpalvec=GGIsetpalvec;
		vis->opcolor->getpalvec=GGIgetpalvec;
			/* This is just to trigger the initial setup. */
		GGIsetpalvec(vis,0,256,XLIB_PRIV(vis)->colormap);
	}
	/* color */
	vis->opcolor->unmappixel=GGIunmappixel;

	vis->opgc->gcchanged=GGIgcchanged;
	 
		/* pixels */
	vis->op2d->drawpixel=GGIdrawpixel;
	vis->op2d->_drawpixel=GGIdrawpixel;
	vis->op2d->putpixel=GGIputpixel;
	vis->op2d->getpixel=GGIgetpixel;

		/* lines */
	vis->op2d->drawline=GGIdrawline;
	vis->op2d->_drawhline=GGIdrawhline;
	vis->op2d->puthline=GGIputhline;
	vis->op2d->drawvline=GGIdrawvline;
	vis->op2d->_drawvline=GGIdrawvline;
	vis->op2d->putvline=GGIputvline;
	 
		/* boxes */
	vis->op2d->drawbox=GGIdrawbox;
	vis->op2d->putbox=GGIputbox;
	vis->op2d->copybox=GGIcopybox;
	vis->op2d->getbox=GGIgetbox;
	vis->op2d->fillscreen=GGIfillscreen;
		/* text */
/*	Not operational yet. */
/*	vis->op2d->putc=GGIputc;
	vis->op2d->puts=GGIputs;*/

	return EOK;
}

int GGIsetmode(ggi_visual *vis,ggi_mode *tm)
{ 
	struct Xhooks *xhook=VIS_PRIVATE(vis);
	XSizeHints hint;
	unsigned int fg, bg;
	XVisualInfo vinfo;
	int rc;
	
	if ((rc=GGIcheckmode(vis,tm))) return rc;
	
	xhook=VIS_PRIVATE(vis);
	_GGIbasiccheck(vis,tm,&vinfo);
	
	if (xhook->XLibLock++!=0) 
		{xhook->XLibLock--;return -EAGAIN;}	/* FIXME use spinlocks here !*/

	xhook->viswidth=tm->visible.x;
	xhook->visheight=tm->visible.y;

		/* Fill in hint structure. Minimal window until we know better. */
	hint.x      = hint.y      = 0;
	hint.width  = tm->visible.x;
	hint.height = tm->visible.y;
	hint.flags  = PSize;
  
		/* Get some colors */
	bg = WhitePixel (xhook->display, xhook->screen);
	fg = BlackPixel (xhook->display, xhook->screen);
	DPRINT("X-lib has colors.\n");
  
	if (xhook->window) 
		XDestroyWindow(xhook->display,xhook->window);

	if (VIS_FB_LINEAR(vis)!=NULL) {
		VIS_FB_LINEAR(vis)=NULL;
		VIS_FB_LINEAR_SIZE(vis)=0;
	}


	xhook->window = XCreateSimpleWindow (xhook->display,
										 RootWindow (xhook->display,xhook->screen),
										 hint.x, hint.y,
										 hint.width, hint.height,
										 2, fg, bg);
  
	DPRINT("X-lib has window.\n");

		/* Tell other applications about this window */
	XSetStandardProperties (xhook->display, xhook->window, "GGI-on-X", "GGI-on-X", None, NULL, 0, &hint);
	 
	if (!xhook->gc)
		xhook->gc = XCreateGC(xhook->display, xhook->window, 0, 0);
	if (!xhook->tempgc)
		xhook->tempgc = XCreateGC(xhook->display, xhook->window, 0, 0);
	DPRINT("X-lib GCs allocated.\n");

	XResizeWindow(xhook->display, xhook->window, tm->visible.x, tm->visible.y);
	DPRINT("X: Resize.\n");

	DPRINT("X: About to map.\n");
		/* Map window. */
	XSelectInput (xhook->display, xhook->window, ExposureMask);
	XMapRaised(xhook->display, xhook->window);
	XFlush(xhook->display);
		/* Wait for window to become mapped */
	{
		XEvent        event;
		XNextEvent (xhook->display, &event);
	}
	DPRINT("X: Window Mapped.\n");

		/* Select input events to listen for */
	XSelectInput(xhook->display, xhook->window, 
				 KeyPressMask|KeyReleaseMask|
				 ButtonPressMask|ButtonReleaseMask|
				 PointerMotionMask);

		/* Turn on backing store */
	{
		XSetWindowAttributes attrib;
		attrib.backing_store=Always;
		XChangeWindowAttributes(xhook->display, xhook->window,
								CWBackingStore, &attrib);
	}

	if (xhook->cmap)
		XFreeColormap(xhook->display,xhook->cmap);
	 
	if (tm->graphtype==GT_8BIT) {
		xhook->cmap = XCreateColormap(xhook->display, xhook->window, 
									  vinfo.visual, 
									  AllocAll);
		 
		xhook->cmap_first=256;
		xhook->cmap_last=0;
		 
		DPRINT("X-lib colormap allocated %x.\n",xhook->cmap);
	} else {
		xhook->cmap = XCreateColormap(xhook->display, xhook->window, 
									  vinfo.visual, 
									  AllocNone);
	}

	GGIflush(vis);

	DPRINT("X: Flush.\n");

	memcpy(VIS_MODE(vis),tm,sizeof(ggi_mode));

	xhook->XLibLock--;	/* FIXME Spinlocks ... */

	return _GGIdomode(vis);
}

int GGIflush(ggi_visual *vis)
{
	struct Xhooks *xhook=(struct Xhooks *)VIS_PRIVATE(vis);
	
	if (xhook->XLibLock++!=0) 
		{xhook->XLibLock--;return -EAGAIN;}	/* FIXME use spinlocks here !*/
	
	if (xhook->cmap && xhook->cmap_first<xhook->cmap_last) {
		
		int x;
		XColor xcol;
		for(x=xhook->cmap_first;x<xhook->cmap_last;x++)
			{
				xcol.red  =XLIB_PRIV(vis)->colormap[x].r;
				xcol.green=XLIB_PRIV(vis)->colormap[x].g;
				xcol.blue =XLIB_PRIV(vis)->colormap[x].b;
				xcol.pixel=x;
				xcol.flags= DoRed | DoGreen | DoBlue ;
				XStoreColor(xhook->display, xhook->cmap,&xcol);
			}
		xhook->cmap_first=256;
		xhook->cmap_last=0;
		XSetWindowColormap(xhook->display, xhook->window, xhook->cmap);
	}
	
	XFlush(xhook->display);
	xhook->XLibLock--;	/* FIXME Spinlocks ... */
	
	return EOK;
}
