/* 
   Common code for the implementation of SYNC mode on targets which are
   inherently ASYNC (e.g. X) and require manual flushes of the framebuffer.

   Copyright (C) 1998 Steve Cheng <steve@ggi-project.org>
   // and whoever wrote the original code

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* 
   There are currently three methods of doing the above:

   MANSYNC_CHILD: A child process, periodically signalling the parent to flush
   MANSYNC_PTHREAD: A new thread which periodically flushes the framebuffer
   MANSYNC_ALARM: (for completeness) An alarm is set up to signal flushes

   In all cases, target must provide the MANSYNC_DOFLUSH macro which
   expands to the target-specific flush implementation. See display-X or
   display-tile for examples.

   Various other macros such as MANSYNC_ISASYNC should be set to e.g. 
       (((Struct Xhooks *)VIS_PRIVATE(vis))->isasync)
   where vis is supplied by the various _mansync* functions.

   Also, be sure to set vis->opdisplay->flush to GGIflush.

   The target itself is responsible for locking its flush code.

   Convenience _mansync* functions are provided in case the target needs to
   temporarily disable mansync during mode setting, etc.

*/

#ifndef MANSYNC_INC
#define MANSYNC_INC

#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <pthread.h>

#include "ggi-dl.h"

/* Required defines */
#ifndef MANSYNC_DOFLUSH
	#error Target must provide MANSYNC_DOFLUSH macro!
#endif
#ifndef MANSYNC_ISASYNC
	#error Target must provide MANSYNC_ISASYNC!
#endif
#ifndef MANSYNC_SYNCFPS
	#define MANSYNC_FPS	20
#endif





#ifdef MANSYNC_CHILD

#ifndef MANSYNC_CHILDPID
	#error Target must provide MANSYNC_CHILDPID!
#endif
#ifndef MANSYNC_OLDHANDLER
	#error Target must provide MANSYNC_OLDHANDLER!
#endif

#ifndef MANSYNC_SYNCSIGNAL
	#define MANSYNC_SIGNAL	SIGPROF
#endif

void _mansync_handler(int unused)
{ 
	signal(MANSYNC_SIGNAL,_mansync_handler);
	MANSYNC_DOFLUSH;
}

void _mansync_start(ggi_visual *vis)
{
        switch (MANSYNC_CHILDPID = fork()) {
	        case 0: {
			int ppid = getppid();
			for(;;) {
				usleep(1000000/MANSYNC_FPS);
				if (kill(ppid, MANSYNC_SIGNAL) != 0)
					_exit(0);
			}
			} break;
	        case -1:
			break;
		default:
			MANSYNC_OLDHANDLER = signal(MANSYNC_SIGNAL,_mansync_handler);
			MANSYNC_ISASYNC = 0;
			break;
	}
}

void _mansync_stop(ggi_visual *vis)
{
	signal(MANSYNC_SIGNAL, MANSYNC_OLDHANDLER);
	kill(MANSYNC_CHILDPID, SIGKILL);
	waitpid(MANSYNC_CHILDPID, NULL, 0);
	MANSYNC_ISASYNC= 1;
}

void _mansync_ignore(ggi_visual *vis)
{
	signal(MANSYNC_SIGNAL, SIG_IGN);
}

void _mansync_cont(ggi_visual *vis)
{
	signal(MANSYNC_SIGNAL, _mansync_handler);
}

#endif /* MANSYNC_CHILD */





#ifdef MANSYNC_PTHREAD


void _mansync_thread(ggi_visual *vis)
{
	while(!(VIS_FLAGS(vis) & GGIFLAG_ASYNC)) {
		usleep(1000000/MANSYNC_FPS);
		MANSYNC_DOFLUSH;
	}
}

void _mansync_start(ggi_visual *vis)
{
	pthread_t p;
	pthread_create(&p, NULL, _mansync_thread, vis);
	MANSYNC_ISASYNC = 0;
}

void _mansync_stop(ggi_visual *vis)
{
	/* Thread should die automatically */
	MANSYNC_ISASYNC = 1;
}

void _mansync_ignore(ggi_visual *vis)
{
}

void _mansync_cont(ggi_visual *vis)
{
}

#endif	/* MANSYNC_PTHREAD */





#ifdef MANSYNC_ALARM

#ifdef _AIX
int setitimer (int Which, const struct itimerval *value, struct itimerval *ovalue);
#endif

void _mansync_handler(int unused)
{ 
	struct itimerval tval={{0,00000},{0,200000}};
	signal(SIGALRM, _mansync_handler);
	MANSYNC_DOFLUSH;
	setitimer(ITIMER_REAL,&tval,NULL);
}

void _mansync_start(ggi_visual *vis)
{
	signal(SIGALRM, _mansync_handler);
	alarm(1);
	MANSYNC_ISASYNC = 0;
}

void _mansync_stop(ggi_visual *vis)
{
	alarm(0);
	signal(SIGALRM, SIG_DFL);
	MANSYNC_ISASYNC = 1;
}

void _mansync_ignore(ggi_visual *vis)
{
	signal(SIGALRM, SIG_IGN);
}

void _mansync_cont(ggi_visual *vis)
{
	signal(SIGALRM, _mansync_handler);
}

#endif	/* MANSYNC_ALARM */

int GGIflush(ggi_visual *vis)
{
	if (MANSYNC_ISASYNC && !(VIS_FLAGS(vis) & GGIFLAG_ASYNC))
		_mansync_start(vis);

	else if (!MANSYNC_ISASYNC && (VIS_FLAGS(vis) & GGIFLAG_ASYNC))
		_mansync_stop(vis);

	return MANSYNC_DOFLUSH;
}

#endif	/* MANSYNC_INC armor */
