/* 
   Graphics library for GGI. Mode management.

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#include "ggi-dl.h"
#include "internal.h"
#include "debug.h"
#include <ggi/ggi_commands.h>

int GGIsetorigin(ggi_visual *vis,int x,int y)
{
	ggi_coord where;

	CHECKXY(vis,x,y);

	where.x=x;
	where.y=y;

	return ioctl(VIS_FD(vis),CHIP_SETVISFRAME,&where);
}

int GGIsetsplitline(ggi_visual *vis,int y)
{
	if (y<0 || y>VIS_MODE(vis)->visible.y) return -1;

	return ioctl(VIS_FD(vis),CHIP_SETSPLITLINE,y);
}

/*
 * _Attempt_ to get the default framebuffer.. 
 */
static void _GGIgetmmio(ggi_visual *vis)
{
	int size=VIS_MODE(vis)->virt.x*VIS_MODE(vis)->virt.y;

	if (VIS_FB_LINEAR(vis)!=NULL) {
		munmap(VIS_FB_LINEAR(vis),VIS_FB_LINEAR_SIZE(vis));
		VIS_FB_LINEAR(vis)=NULL;
		VIS_FB_LINEAR_SIZE(vis)=0;
	}

	size=_ggiSetupMode(vis);
	DPRINT("Calculated size=%d bytes\n",size);

	if (size <= 0) 
		return;

	VIS_FB_LINEAR_SIZE(vis)=size;	
	VIS_FB_LINEAR(vis)=mmap(NULL,
                     size,
                     PROT_READ|PROT_WRITE,
                     MAP_SHARED,
                     VIS_FD(vis),
	             MMAP_TYPE_MMIO|MMAP_PER_REGION_TYPE|MMAP_FRAMEBUFFER);

	DPRINT("Linear FB=%p\n",VIS_FB_LINEAR(vis));
	if (VIS_FB_LINEAR(vis) == (void *)(-1)) {
		VIS_FB_LINEAR(vis)=NULL;
		VIS_FB_LINEAR_SIZE(vis)=0;
	}
}

static int _GGIdomode(ggi_visual *vis)
{
	int err;
	ggi_suggest sug;

	_GGIgetmmio(vis);

	err=(_ggiOpenDL(vis,"generic-stubs","")==NULL);
	if (err) {
		fprintf(stderr,"display-KGI: Can't load the \"generic-stubs\" "
			       "library\n");
	}
		
	vis->op2d->setsplitline=GGIsetsplitline;
	vis->op2d->setorigin=GGIsetorigin;

	sug.id=0;
	do {
		err=ioctl(VIS_FD(vis),GRAPHICS_GETSUGGEST,&sug);
		if (err!=0) {
			fprintf(stderr,"display-KGI: Failed getting suggestion %d\n",
				       sug.id);
			perror("display-KGI");
			return -1;	/* Error */
		}

		DPRINT("display-KGI - attempting %s (%s)\n",sug.name,sug.args);
		err=(_ggiOpenDL(vis,sug.name,sug.args)==NULL);
		if (err) {
			fprintf(stderr,"display-KGI: Can't find an appropriate "
				       "library for %s (%s)\n",
					sug.name,sug.args);
			return err;
		} else {
			DPRINT("Success in loading %s (%s)\n",sug.name,sug.args);
		}
	} while (sug.id!=0);

	return EOK;
}

int GGIsetmode(ggi_visual *vis,ggi_mode *tm)
{ 
	int err;

	if (vis==NULL) {
		DPRINT("Visual==NULL\n");
		return -1;
	}
	
	/* Temporary */
	if (VIS_FB_LINEAR(vis)!=NULL) {	/* Unmap mem - it might be invalid after mode-change */
		munmap(VIS_FB_LINEAR(vis),VIS_FB_LINEAR_SIZE(vis));
		VIS_FB_LINEAR(vis)=NULL;
	}
	_ggiZapMode(vis,~GGI_DL_OPDISPLAY);

	err=ioctl(VIS_FD(vis),DRIVER_SETGRAPHMODE,tm);
	if (err) {
		DPRINT("%d=ioctl(%d,DRIVER_SETGRAPHMODE,%p)\n",err,VIS_FD(vis),tm);
		return err;
	}
	memcpy(VIS_MODE(vis),tm,sizeof(ggi_mode));

	return _GGIdomode(vis);
}

/**********************************/
/* check any mode (text/graphics) */
/**********************************/
int GGIcheckmode(ggi_visual *vis,ggi_mode *tm)
{
	if (vis==NULL)
		return -1;
	
	return ioctl(VIS_FD(vis),DRIVER_TESTMODE,tm);
}

/************************/
/* get the current mode */
/************************/
int GGIgetmode(ggi_visual *vis,ggi_mode *tm)
{
	DPRINT("In GGIgetmode(%p,%p)\n",vis,tm);
	if (vis==NULL)
		return -1;
	
	return ioctl(VIS_FD(vis),DRIVER_GETGRAPHMODE,tm);
}
