/* ------------------------------------------------------------------------ 
 *	mode.c  - part of the TrueEmu target of GGI.
 * ------------------------------------------------------------------------
 *
 *	Copyright (c) 1998  Andrew Apted  <andrew@ggi-project.org>
 *
 *	This library is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU Library General Public
 *	License as published by the Free Software Foundation; either
 *	version 2, or (at your option) any later version.
 *
 *	This library is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Library General Public License for more details.
 *
 *	You should have received a copy of the GNU Library General
 *	Public License along with this library; see the file
 *	COPYING.LGPL.  If not, write to the Free Software Foundation,
 *	675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#include <ggi/ggi_commands.h>

#include "ggi-dl.h"
#include "internal.h"
#include "debug.h"

#include "trueemu.h"


/*
 * Attempt to get the default framebuffer.
 */

static int do_getmmap(ggi_visual *vis)
{
	int size;

	if (VIS_FB_LINEAR(vis) != NULL) {
	
		trueemuClose(vis);
		
		free(VIS_FB_LINEAR(vis));
		VIS_FB_LINEAR(vis) = NULL;
		VIS_FB_LINEAR_SIZE(vis) = 0;
	}


	/* setup defaults for FB width, height, bpp etc. */

	size = _ggiSetupMode(vis);

	if (size > 0) {
		VIS_FB_LINEAR(vis) = malloc(size);

		if (VIS_FB_LINEAR(vis) == NULL) {
			DPRINT("display-trueemu: Failed to mmap a "
				"%dK area.\n", size/1024);
			return -1;
		}
		
		VIS_FB_LINEAR_SIZE(vis) = size;	
	}

	return EOK;
}


static int do_loadstub(ggi_visual *vis, ggi_suggest *sug)
{
	if (_ggiOpenDL(vis, sug->name, sug->args) == NULL) {

		fprintf(stderr, "display-trueemu: Unable to open the "
			"library %s (%s).\n", sug->name, sug->args);
		return -1;
	}

	DPRINT("display-trueemu: Loaded %s library.\n", sug->name);

	return EOK;
}


static void nuke_directbuffers(ggi_visual *vis)
{
	while (VIS_DB(vis)) {

		ggi_directbuffer *db = VIS_DB(vis);

		VIS_DB(vis) = db->next;

		free(db);
	}
}


static int do_setmode(ggi_visual *vis)
{
	TrueemuHook *th = VIS_PRIVATE(vis);

	ggi_suggest sug;
	int err;

	_ggiZapMode(vis, ~GGI_DL_OPDISPLAY);


	/* get mmap */

	if ((err = do_getmmap(vis)) != EOK) {
		return err;
	}
	

	/* load libraries */

	strcpy(sug.name, "generic-stubs");
	strcpy(sug.args, "");

	if ((err = do_loadstub(vis, &sug)) != EOK) {
		return err;
	}

	sprintf(sug.name, "generic-linear-%d", 
		(VIS_MODE(vis)->graphtype == GT_32BIT) ? 32 : 24);
	strcpy(sug.args, "");

	if ((err = do_loadstub(vis, &sug)) != EOK) {
		return err;
	}

	/* Backup the current 2D operations, and override them with our
	 * own (which then call on these backups to do the work).
	 */

	th->mem_op2d = malloc(sizeof(struct ggi_visual_op2d));

	*th->mem_op2d = *vis->op2d;

	vis->op2d->drawpixel=GGIdrawpixel;
	vis->op2d->drawline=GGIdrawline;
	vis->op2d->drawhline=GGIdrawhline;
	vis->op2d->drawvline=GGIdrawvline;

	vis->op2d->putc=GGIputc;
	vis->op2d->puts=GGIputs;

	vis->op2d->putpixel=GGIputpixel;
	vis->op2d->puthline=GGIputhline;
	vis->op2d->putvline=GGIputvline;
	vis->op2d->putbox=GGIputbox;

	vis->op2d->drawbox=GGIdrawbox;
	vis->op2d->copybox=GGIcopybox;
	vis->op2d->crossblit=GGIcrossblit;
	vis->op2d->fillscreen=GGIfillscreen;

	vis->op2d->setsplitline=GGIsetsplitline;
	vis->op2d->setorigin=GGIsetorigin;

	/* Nuke the directbuffer -- this forces applications to use
	 * ggiPutBox() which allows us to optimize our things better.
	 */

	nuke_directbuffers(vis);
	 
	DPRINT("display-trueemu: Attempting to setmode on parent "
		"visual...\n");

	return trueemuOpen(vis);
}


int GGIsetmode(ggi_visual *vis, ggi_mode *mode)
{ 
	int err;

	if ((vis == NULL) || (mode == NULL) || (VIS_MODE(vis) == NULL)) {
		DPRINT("display-trueemu: vis/mode == NULL\n");
		return -1;
	}
	
	DPRINT("display-trueemu: TE_setmode %dx%d (gt=%d)\n",
		mode->visible.x, mode->visible.y, mode->graphtype);

	if ((err = GGIcheckmode(vis, mode)) != EOK) {
		return err;
	}

	memcpy(VIS_MODE(vis), mode, sizeof(ggi_mode));
	
	if ((err = do_setmode(vis)) != EOK) {
		return err;
	}

	DPRINT("display-trueemu: TE_setmode succeeded.\n", vis, mode);

	return EOK;
}


int GGIcheckmode(ggi_visual *vis, ggi_mode *mode)
{
	if ((vis == NULL) || (mode == NULL)) {
		DPRINT("display-trueemu: vis/mode == NULL\n");
		return -1;
	}

	DPRINT("display-trueemu: TE_checkmode %dx%d (gt=%d)\n",
		mode->visible.x, mode->visible.y, mode->graphtype);

	if ((mode->graphtype != GT_24BIT) &&
	    (mode->graphtype != GT_32BIT)) {

		DPRINT("display-trueemu: TE_checkmode: graphtype "
			"not supported.\n");
		return -1;
	}

	return EOK;
}


int GGIgetmode(ggi_visual *vis, ggi_mode *mode)
{
	if ((vis == NULL) || (mode == NULL) || (VIS_MODE(vis) == NULL)) {
		DPRINT("display-trueemu: vis/mode == NULL\n");
		return -1;
	}
	
	DPRINT("display-trueemu: TE_getmode.\n", vis, mode);

	memcpy(mode, VIS_MODE(vis), sizeof(ggi_mode));

	return EOK;
}


int GGIflush(ggi_visual *vis)
{
	return trueemuFlush(vis);
}
