/* ------------------------------------------------------------------------
 *	TrueEmu - 24 bit emulation target for GGI
 * ------------------------------------------------------------------------
 *
 * 	Copyright (c) 1998  Andrew Apted  <andrew@ggi-project.org>
 *
 *	This library is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU Library General Public
 *	License as published by the Free Software Foundation; either
 *	version 2, or (at your option) any later version.
 * 
 *	This library is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Library General Public License for more details.
 *
 *	You should have received a copy of the GNU Library General
 *	Public License along with this library; see the file
 *	COPYING.LGPL.  If not, write to the Free Software Foundation,
 *	675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ------------------------------------------------------------------------
 */

#include <ggi/maintainers.h>
#define MAINTAINER  Andrew_Apted

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>

#include <ggi/ggi.h>
#include <ggi/libggi.h>

#include "ggi-dl.h"
#include "internal.h"
#include "debug.h"

#include "trueemu.h"


/**************************************************
 ***
 ***  Various Defines
 ***
 **************************************************/


#define NUM_PASTELS	21

#define CUBE_LEVELS	6
#define PASTEL_LEVELS	12	/* not including black */

#define C16_GREYS	6	/* not including black */
#define C16_BROWNS	2
#define C16_GREENS	2

#define NO_F  -1,-1	/* means: not this face */


/***
 ***  Lookup Tables
 ***/


typedef struct pastel
{
	ggi_color color;

	TrueColor RF_min, RF_max;	/* area on red face */
	TrueColor GF_min, GF_max;	/* area on green face */
	TrueColor BF_min, BF_max;	/* area on blue face */

} Pastel;


static Pastel pastel_array[NUM_PASTELS] =
{
	/* extra colors */

	{ T2GGI(0xffbc7c), 0xffa055,0xffe0a0, NO_F, NO_F },
	{ T2GGI(0xffdcb0), 0xffa0a0,0xffe8d8, NO_F, NO_F },

	/* standard 19 points on the outer RGB cube face */

	{ T2GGI(0x0000ff), NO_F, NO_F, 0x0000ff,0x5555ff },
	{ T2GGI(0x0080ff), NO_F, NO_F, 0x0055ff,0x55aaff },
	{ T2GGI(0x8000ff), NO_F, NO_F, 0x5500ff,0xaa55ff },
	{ T2GGI(0x8080ff), NO_F, NO_F, 0x5555ff,0xaaaaff },
	
	{ T2GGI(0x00ff00), NO_F, 0x00ff00,0x55ff55, NO_F },
	{ T2GGI(0x00ff80), NO_F, 0x00ff55,0x55ffaa, NO_F },
	{ T2GGI(0x80ff00), NO_F, 0x55ff00,0xaaff55, NO_F },
	{ T2GGI(0x80ff80), NO_F, 0x55ff55,0xaaffaa, NO_F },
	{ T2GGI(0x00ffff), NO_F, 0x00ffaa,0x55ffff, 0x00aaff,0x55ffff },
	{ T2GGI(0x80ffff), NO_F, 0x55ffaa,0xaaffff, 0x55aaff,0xaaffff },

	{ T2GGI(0xff0000), 0xff0000,0xff5555, NO_F, NO_F },
	{ T2GGI(0xff0080), 0xff0055,0xff55aa, NO_F, NO_F },
	{ T2GGI(0xff8000), 0xff5500,0xffaa55, NO_F, NO_F },
	{ T2GGI(0xff8080), 0xff5555,0xffaaaa, NO_F, NO_F },
	{ T2GGI(0xff00ff), 0xff00aa,0xff55ff, NO_F, 0xaa00ff,0xff55ff },
	{ T2GGI(0xff80ff), 0xff55aa,0xffaaff, NO_F, 0xaa55ff,0xffaaff },
	{ T2GGI(0xffff00), 0xffaa00,0xffff55, 0xaaff00,0xffff55, NO_F },
	{ T2GGI(0xffff8c), 0xffaa55,0xffffaa, 0xaaff55,0xffffc7, NO_F },
	{ T2GGI(0xffffff), 0xffaaaa,0xffffff, 0xaaffc7,0xffffff, 
			   0xaaaaff,0xffffff }
};


static ggi_color col16_palette[16] =
{
	T2GGI(0x000000),	/* standard 2x2x2 cube */
	T2GGI(0x0000ff),
	T2GGI(0x00ff00),
	T2GGI(0x00ffff),
	T2GGI(0xff0000),
	T2GGI(0xff00ff),
	T2GGI(0xffff00),
	T2GGI(0xffffff),

	T2GGI(0x2a2a2a),	/* five extra greys */
	T2GGI(0x555555),
	T2GGI(0x7f7f7f),
	T2GGI(0xaaaaaa),
	T2GGI(0xd5d5d5),

	T2GGI(0x007f00),	/* extra green */
	
	T2GGI(0x7f5839),	/* extra brown & orange */
	T2GGI(0xffb172)
};

static int col16_greys [C16_GREYS+1]  = { 0,8,9,10,11,12,7 };
static int col16_browns[C16_BROWNS+1] = { 0,14,15 };
static int col16_greens[C16_GREENS+1] = { 0,13,2 };


static ggi_color black = { 0,0,0 };
static ggi_color white = { 0xffff,0xffff,0xffff };


/**************************************************
 ***
 ***  Internal functions 
 ***
 **************************************************/


static int calc_default_mode(int mode, int graphtype, int width)
{
	int default_dither  = TRUEEMU_M_DITHER_4;
	int default_palette = TRUEEMU_M_RGB;
	
	switch (graphtype)
	{
		case GT_32BIT:
		case GT_24BIT:
		{
			default_dither = TRUEEMU_M_DITHER_0;
			break;
		}

		case GT_16BIT:
		case GT_15BIT:
		{
			if (width < 512) {
				default_dither = TRUEEMU_M_DITHER_2;
			}
			break;
		}

		case GT_8BIT:
		{
			if (width < 512) {
				default_dither = TRUEEMU_M_DITHER_2;
			}

			default_palette = TRUEEMU_M_CUBE;
			break;
		}

		case GT_4BIT:
		{
			default_palette = TRUEEMU_M_PASTEL;
			break;
		}
		
		default:
			INTERNAL_ERROR;
	}

	if ((mode & TE_DITHER_MASK) == 0) {
		mode |= default_dither;
	}
	
	if ((mode & TE_PALETTE_MASK) == 0) {
		mode |= default_palette;
	}

	return mode;
}


/*** 
 ***  Palette routines 
 ***/


static void load_332_palette(ggi_color *colormap)
{
	int r, g, b;

	for (r = 0; r < 8; r++)
	for (g = 0; g < 8; g++)
	for (b = 0; b < 4; b++) {

		int index = (r << 5) | (g << 2) | b;

		ggi_color col;

		col.r = r * 0xffff / (8-1);
		col.g = g * 0xffff / (8-1);
		col.b = b * 0xffff / (4-1);

		colormap[index] = col;
	}
}


static void load_cube_palette(ggi_color *colormap)
{
	int r, g, b;

	for (r = 0; r < CUBE_LEVELS; r++)
	for (g = 0; g < CUBE_LEVELS; g++)
	for (b = 0; b < CUBE_LEVELS; b++) {

		int index = (r * CUBE_LEVELS + g) * CUBE_LEVELS + b;

		ggi_color col;

		col.r = r * 0xffff / (CUBE_LEVELS-1);
		col.g = g * 0xffff / (CUBE_LEVELS-1);
		col.b = b * 0xffff / (CUBE_LEVELS-1);

		colormap[index] = col;
	}
}


static void load_pastel_palette(ggi_color *colormap)
{
	int index;
	int pastel;


	/* color 0 is black */

	colormap[0] = black;


	for (pastel = 0; pastel < NUM_PASTELS; pastel++)
	{
		int pl;

		int tr = pastel_array[pastel].color.r;
		int tg = pastel_array[pastel].color.g;
		int tb = pastel_array[pastel].color.b;

		index = 1 + pastel * PASTEL_LEVELS;

		for (pl=1; pl <= PASTEL_LEVELS; pl++) {

			ggi_color col;

			col.r = pl * tr / PASTEL_LEVELS;
			col.g = pl * tg / PASTEL_LEVELS;
			col.b = pl * tb / PASTEL_LEVELS;

			colormap[index++] = col;
		}
	}
}


static void load_col16_palette(ggi_color *colormap)
{
	int index;

	for (index=0; index < 16; index++) {
		colormap[index] = col16_palette[index];
	}
}


static void load_121_palette(ggi_color *colormap)
{
	int r, g, b;

	for (r = 0; r < 2; r++)
	for (g = 0; g < 4; g++)
	for (b = 0; b < 2; b++) {

		int index = (r << 3) | (g << 1) | b;

		ggi_color col;

		col.r = r * 0xffff / (2-1);
		col.g = g * 0xffff / (4-1);
		col.b = b * 0xffff / (2-1);

		colormap[index] = col;
	}
}


static void setup_palette(TrueemuHook *th)
{
	switch (th->parent_gt)
	{
		case GT_8BIT:
		{
			if ((th->mode & TRUEEMU_M_PASTEL) ||
			    (th->mode & TRUEEMU_M_PASTEL_2)) {
				load_pastel_palette(th->colormap);

			} else if (th->mode & TRUEEMU_M_CUBE) {
				load_cube_palette(th->colormap);

			} else {
				load_332_palette(th->colormap);
			}

			ggiSetPaletteVec(th->parent, 0, 256, th->colormap);
			ggiFlush(th->parent);
			return;
		}

		case GT_4BIT:
		{
			if ((th->mode & TRUEEMU_M_PASTEL) ||
			    (th->mode & TRUEEMU_M_PASTEL_2)) {
				load_col16_palette(th->colormap);
			} else {
				load_121_palette(th->colormap);
			}

			ggiSetPaletteVec(th->parent, 0, 16, th->colormap);
			ggiFlush(th->parent);
			return;
		}
	}
	
	/* no palette for other graphtypes */
}


/***
 ***  Dithering calculations 
 ***/


static int lookup_pastel(int r, int g, int b)
{
	int tr, tg, tb;
	int pastel;

	int max = MAX(r, MAX(g, b));


	if (max == 0) {

		/* Completely black, therefore it doesn't matter which
		 * pastel we choose.  Just choose the first one.
		 */

		return 0;
	}


	/* project color onto the RGB cube face */

	tr = (r * 0xff) / max;
	tg = (g * 0xff) / max;
	tb = (b * 0xff) / max;


	/* find match in pastel array */

	for (pastel=0; pastel < NUM_PASTELS; pastel++) {

		Pastel *p = & pastel_array[pastel];


		if ((tr == 0xff) && ! (p->RF_min < 0)) {

			/* try the red face */

			if ((TC_GREEN(p->RF_min) <= tg) &&
				(tg <= TC_GREEN(p->RF_max)) &&
				(TC_BLUE(p->RF_min) <= tb) &&
				(tb <= TC_BLUE(p->RF_max))) {

				return pastel;
			}
		}

		if ((tg == 0xff) && ! (p->GF_min < 0)) {

			/* try the green face */

			if ((TC_RED(p->GF_min) <= tr) &&
				(tr <= TC_RED(p->GF_max)) &&
				(TC_BLUE(p->GF_min) <= tb) &&
				(tb <= TC_BLUE(p->GF_max))) {

				return pastel;
			}
		}

		if ((tb == 0xff) && ! (p->BF_min < 0)) {

			/* try the blue face */

			if ((TC_RED(p->BF_min) <= tr) &&
				(tr <= TC_RED(p->BF_max)) &&
				(TC_GREEN(p->BF_min) <= tg) &&
				(tg <= TC_GREEN(p->BF_max))) {

				return pastel;
			}
		}
	}


	/* Oops, pastel array has a hole in it */

	ggiPanic("Pastel array has a hole in it.\n");
	return 0;
}


static void calc_hi16_dither(TrueemuHook *th, int shift)
{
	int i, n;

	int num = 1 << shift;
	int R_bands = 1 + num * (32-1);
	int G_bands = 1 + num * (64-1);
	int B_bands = 1 + num * (32-1);

	
	/* allocate lookup tables */
		
	if (th->R == NULL) {
		
		th->R = malloc(256 * 4 * 2);
		th->G = malloc(256 * 4 * 2);
		th->B = malloc(256 * 4 * 2);

		if ((th->R == NULL) || (th->G == NULL) || (th->B == NULL)) {
			ggiPanic("No memory for hi16 dither.\n");
			return;
		}
	}


	for (i=0; i < 256; i++)		/* intensity */
	for (n=0; n < num; n++) {
	 
	 	int hr = (((i * R_bands) >> 8) + n) >> shift;
	 	int hg = (((i * G_bands) >> 8) + n) >> shift;
	 	int hb = (((i * B_bands) >> 8) + n) >> shift;

		th->R[i][n] = (hr << 11);
		th->G[i][n] = (hg << 5);
		th->B[i][n] = (hb);
	}
}


static void calc_hi15_dither(TrueemuHook *th, int shift)
{
	int i, n;

	int num = 1 << shift;
	int bands = 1 + num * (32-1);

	
	/* allocate lookup tables */
		
	if (th->R == NULL) {
		
		th->R = malloc(256 * 4 * 2);
		th->G = malloc(256 * 4 * 2);
		th->B = malloc(256 * 4 * 2);

		if ((th->R == NULL) || (th->G == NULL) || (th->B == NULL)) {
			ggiPanic("No memory for hi15 dither.\n");
			return;
		}
	}


	for (i=0; i < 256; i++)		/* intensity */
	for (n=0; n < num; n++) {
	 
	 	int h = (((i * bands) >> 8) + n) >> shift;

		th->R[i][n] = (h << 10);
		th->G[i][n] = (h << 5);
		th->B[i][n] = (h << 0);
	}
}


static void calc_332_dither(TrueemuHook *th, int shift)
{
	int r, g, b, n;

	int num = 1 << shift;
	int R_bands = 1 + num * (8-1);
	int G_bands = 1 + num * (8-1);
	int B_bands = 1 + num * (4-1);

	
	/* allocate lookup table */
		
	if (th->T == NULL) {
		
		th->T = malloc(32768 * 4);

		if (th->T == NULL) {
			ggiPanic("No memory for 3:3:2 dither.\n");
			return;
		}
	}


	for (r=0; r < 32; r++)
	for (g=0; g < 32; g++)
	for (b=0; b < 32; b++)

	for (n=0; n < num; n++) {

		int cr = (((r * R_bands) >> 5) + n) >> shift;
		int cg = (((g * G_bands) >> 5) + n) >> shift;
		int cb = (((b * B_bands) >> 5) + n) >> shift;

		th->T[(r << 10) | (g << 5) | b][n] =
			(cr << 5) | (cg << 2) | cb;
	}
}


static void calc_cube_dither(TrueemuHook *th, int shift)
{
	int r, g, b, n;

	int num = 1 << shift;
	int bands = 1 + num * (CUBE_LEVELS-1);

	
	/* allocate lookup table */
		
	if (th->T == NULL) {
		
		th->T = malloc(32768 * 4);

		if (th->T == NULL) {
			ggiPanic("No memory for cube dither.\n");
			return;
		}
	}


	for (r=0; r < 32; r++)
	for (g=0; g < 32; g++)
	for (b=0; b < 32; b++)

	for (n=0; n < num; n++) {

		int cr = (((r * bands) >> 5) + n) >> shift;
		int cg = (((g * bands) >> 5) + n) >> shift;
		int cb = (((b * bands) >> 5) + n) >> shift;

		th->T[(r << 10) | (g << 5) | b][n] =
			(cr * CUBE_LEVELS + cg) * CUBE_LEVELS + cb;
	}
}


static void calc_pastel_dither(TrueemuHook *th, int shift)
{
	int r, g, b, n;

	int num = 1 << shift;
	int bands = 1 + num * PASTEL_LEVELS;

	/* Note: black acts as an extra pastel level */


	/* allocate lookup table */
		
	if (th->T == NULL) {
		
		th->T = malloc(32768 * 4);

		if (th->T == NULL) {
			ggiPanic("No memory for pastel dither.\n");
			return;
		}
	}


	for (r=0; r < 32; r++)
	for (g=0; g < 32; g++)
	for (b=0; b < 32; b++) {

		int index = (r << 10) | (g << 5) | b;

		int tr = r * 0xff / 31;
		int tg = g * 0xff / 31;
		int tb = b * 0xff / 31;

		int pastel = lookup_pastel(tr, tg, tb);
		int max = MAX(r, MAX(g, b));

		for (n=0; n < num; n++) {

			int pl = (((max * bands) >> 5) + n) >> shift;

			th->T[index][n] = (pl == 0) ? 0 : 
				1 + (pastel * PASTEL_LEVELS) + (pl - 1);
		}
	}
}


static void calc_col16_dither(TrueemuHook *th, int shift)
{
	int r, g, b, n;

	int num = 1 << shift;

	int cube_bands  = 1 + num;
	int grey_bands  = 1 + num * C16_GREYS;
	int brown_bands = 1 + num * C16_BROWNS;
	int green_bands = 1 + num * C16_GREENS;


	/* allocate lookup table */
		
	if (th->T == NULL) {
		
		th->T = malloc(32768 * 4);

		if (th->T == NULL) {
			ggiPanic("No memory for col16 dither.\n");
			return;
		}
	}


	for (r=0; r < 32; r++)
	for (g=0; g < 32; g++)
	for (b=0; b < 32; b++)

	for (n=0; n < num; n++) {

		int index = (r << 10) | (g << 5) | b;
		int max = MAX(r, MAX(g, b));

		int tr,tg,tb;
		int cr,cg,cb;
		int pl;

		if (max == 0) {
			th->T[index][n] = 0;
			continue;
		}

		tr = (r * 0xff) / max;
		tb = (b * 0xff) / max;
		tg = (g * 0xff) / max;


		/* handle greens */

		if ((tg == 0xff) && (tr <= 0xe0) && (tb <= 0xe0)) {

			pl = (((max * green_bands) >> 5) + n) >> shift;
					
			th->T[index][n] = col16_greens[pl];
			continue;
		}

		
		/* handle browns */

		if ((tr == 0xff) && (tb <= 0xd8) &&
			(0x90 <= tg) && (tg <= 0xe8)) {

			pl = (((max * brown_bands) >> 5) + n) >> shift;
					
			th->T[index][n] = col16_browns[pl];
			continue;
		}
		
		
		/* handle greys */

		if (((tr == 0xff) && (tb >= 0xb0) && (tg >= 0xb0)) ||
		    ((tb == 0xff) && (tr >= 0xb0) && (tg >= 0xb0)) ||
		    ((tg == 0xff) && (tr >= 0xb0) && (tb >= 0xb0))) {

			pl = (((max * grey_bands) >> 5) + n) >> shift;
					
			th->T[index][n] = col16_greys[pl];
			continue;
		}
		
		
		/* handle all other colors */

		cr = (((r * cube_bands) >> 5) + n) >> shift;
		cg = (((g * cube_bands) >> 5) + n) >> shift;
		cb = (((b * cube_bands) >> 5) + n) >> shift;

		th->T[index][n] = ((cr*2) + cg)*2 + cb;
	}
}


static void calc_121_dither(TrueemuHook *th, int shift)
{
	int r, g, b, n;

	int num = 1 << shift;
	int R_bands = 1 + num * (2-1);
	int G_bands = 1 + num * (4-1);
	int B_bands = 1 + num * (2-1);

	
	/* allocate lookup table */
		
	if (th->T == NULL) {
		
		th->T = malloc(32768 * 4);

		if (th->T == NULL) {
			ggiPanic("No memory for 1:2:1 dither.\n");
			return;
		}
	}


	for (r=0; r < 32; r++)
	for (g=0; g < 32; g++)
	for (b=0; b < 32; b++)

	for (n=0; n < num; n++) {

		int cr = (((r * R_bands) >> 5) + n) >> shift;
		int cg = (((g * G_bands) >> 5) + n) >> shift;
		int cb = (((b * B_bands) >> 5) + n) >> shift;

		th->T[(r << 10) | (g << 5) | b][n] =
			(cr << 3) | (cg << 1) | cb;
	}
}


static void setup_dithering(TrueemuHook *th)
{
	TrueemuBlits *B;

	int shift=0;


	if (th->child_gt == GT_32BIT) {
		B = & blit32_table;
	} else {
		B = & blit24_table;
	}

	if (th->mode & TRUEEMU_M_DITHER_2) {
		shift=1;
	} else if (th->mode & TRUEEMU_M_DITHER_4) {
		shift=2;
	}

	switch (th->parent_gt)
	{
		case GT_32BIT:
		{
			th->blitter_even = B->blitter_b32_d0;
			th->blitter_odd  = B->blitter_b32_d0;
			return;
		}

		case GT_24BIT:
		{
			th->blitter_even = B->blitter_b24_d0;
			th->blitter_odd  = B->blitter_b24_d0;
			return;
		}

		case GT_15BIT:
		case GT_16BIT:
		{
			if (shift == 1) {
				th->blitter_even = B->blitter_b16_d2_ev;
				th->blitter_odd  = B->blitter_b16_d2_od;
				
			} else if (shift == 2) {
				th->blitter_even = B->blitter_b16_d4_ev;
				th->blitter_odd  = B->blitter_b16_d4_od;
				
			} else {
				th->blitter_even = B->blitter_b16_d0;
				th->blitter_odd  = B->blitter_b16_d0;
			}

			if (th->parent_gt == GT_16BIT) {
				calc_hi16_dither(th, shift);
			} else {
				calc_hi15_dither(th, shift);
			}
			return;
		}

		case GT_8BIT:
		{
			if (shift == 1) {
				th->blitter_even = B->blitter_b8_d2_ev;
				th->blitter_odd  = B->blitter_b8_d2_od;

			} else if (shift == 2) {
				th->blitter_even = B->blitter_b8_d4_ev;
				th->blitter_odd  = B->blitter_b8_d4_od;

			} else {
				th->blitter_even = B->blitter_b8_d0;
				th->blitter_odd  = B->blitter_b8_d0;
			}

			if ((th->mode & TRUEEMU_M_PASTEL) ||
			    (th->mode & TRUEEMU_M_PASTEL_2)) {
				calc_pastel_dither(th, shift);

			} else if (th->mode & TRUEEMU_M_CUBE) {
				calc_cube_dither(th, shift);

			} else {
				calc_332_dither(th, shift);
			}
			return;
		}

		case GT_4BIT:
		{
			if (shift == 1) {
				th->blitter_even = B->blitter_b4_d2_ev;
				th->blitter_odd  = B->blitter_b4_d2_od;

			} else if (shift == 2) {
				th->blitter_even = B->blitter_b4_d4_ev;
				th->blitter_odd  = B->blitter_b4_d4_od;
			
			} else {
				th->blitter_even = B->blitter_b4_d0;
				th->blitter_odd  = B->blitter_b4_d0;
			}

			if ((th->mode & TRUEEMU_M_PASTEL) ||
			    (th->mode & TRUEEMU_M_PASTEL_2)) {
				calc_col16_dither(th, shift);
			} else {
				calc_121_dither(th, shift);
			}
			return;
		}
	}

	INTERNAL_ERROR;
}


/**************************************************
 ***
 ***  Exported functions 
 ***
 **************************************************/


int trueemuUpdate(ggi_visual *vis, int x, int y, int w, int h)
{
	TrueemuHook *th = VIS_PRIVATE(vis);

	
	/* round down x if odd */

	if (x & 1) {
		x--; w++;
	}


	/* do transfer, one horizontal line at a time */

	for (; h > 0; h--, y++) {

		ggiGetHLine(vis, x, y, w, th->src_buf);

		if (y & 1) {
			th->blitter_odd(th,  th->dest_buf, th->src_buf, w);
		} else {
			th->blitter_even(th, th->dest_buf, th->src_buf, w);
		}

		ggiPutHLine(th->parent, x, y, w, th->dest_buf);
	}

	return 0;
}


int trueemuFlush(ggi_visual *vis)
{
	TrueemuHook *th = VIS_PRIVATE(vis);

	int sx = MAX(th->modified_min.x, vis->gc->cliptl.x);
	int sy = MAX(th->modified_min.y, vis->gc->cliptl.y);

	int ex = MIN(th->modified_max.x, vis->gc->clipbr.x-1);
	int ey = MIN(th->modified_max.y, vis->gc->clipbr.y-1);

	int err = EOK;


	/* clear 'dirty region' */

	th->modified_min.x = th->width;
	th->modified_min.y = th->height;
	th->modified_max.x = 0;
	th->modified_max.y = 0;

	
	if ((sx <= ex) && (sy <= ey)) {
		err = trueemuUpdate(vis, sx, sy, ex-sx+1, ey-sy+1);
	}

	ggiFlush(th->parent);

	return err;
}


static ggi_visual *_sync_vis = NULL;

static void do_flush_SYNC(int unused)
{
	signal(SYNCSIGNAL, SIG_IGN);

	if (_sync_vis != NULL) {
		trueemuFlush(_sync_vis);
	}

	signal(SYNCSIGNAL, do_flush_SYNC);

#ifdef NO_GAME_OF_PING_PONG
	if (_sync_vis != NULL) {

		TrueemuHook *th = VIS_PRIVATE(_sync_vis);

		kill(th->sync_process, SYNCSIGNAL);
	}
#endif
}


int trueemuOpen(ggi_visual *vis)
{
	TrueemuHook *th = VIS_PRIVATE(vis);

	int bufsize;

	th->child_gt = VIS_MODE(vis)->graphtype;
	th->width    = VIS_MODE(vis)->visible.x;
	th->height   = VIS_MODE(vis)->visible.y;

	th->mode = calc_default_mode(th->mode, th->parent_gt, th->width);

	th->colormap = NULL;
	th->src_buf  = NULL;
	th->dest_buf = NULL;

	th->R = th->G = th->B = NULL;
	th->T = NULL;


	/* allocate colormap */

	switch (th->parent_gt)
	{
		case GT_32BIT: case GT_24BIT:
		case GT_16BIT: case GT_15BIT:
		{
			break;
		}

		case GT_8BIT: case GT_4BIT:
		{
			int num_cols = 256;
			
			if (th->parent_gt == GT_4BIT) num_cols = 16;
			if (th->parent_gt == GT_1BIT) num_cols = 2;

			th->colormap = (ggi_color *)
				malloc(num_cols * sizeof(ggi_color));

			if (th->colormap == NULL) {
				ggiPanic("No memory for colormap.\n");
				return -2;
			}

			break;
		}

		default:
			INTERNAL_ERROR;
			return -2;
	}
		

	/* !!! Going straight through to the parent when 
	 * parent_gt == child_gt isn't implemented yet.
	 */


	/* allocate buffers */

	bufsize = th->width * 4;

	th->src_buf  = malloc(bufsize);
	th->dest_buf = malloc(bufsize);

	if ((th->src_buf == NULL) || (th->dest_buf == NULL)) {
		ggiPanic("No memory for buffers.\n");
		return -2;
	}


	/* set the parent mode */

	if (ggiSetGraphMode(th->parent, th->width, th->height, 
		th->width, th->height, th->parent_gt) < 0) {

		DPRINT("Couldn't set child graphic mode.\n");
		return -1;
	}


	/* now sort out the dither/palette mode */

	setup_dithering(th);
	setup_palette(th);


	/* clear 'dirty region' */

	th->modified_min.x = th->width;
	th->modified_min.y = th->height;
	th->modified_max.x = 0;
	th->modified_max.y = 0;

	
	/* enable sync_process */

	_sync_vis = vis;
	signal(SYNCSIGNAL, do_flush_SYNC);

	/* serve the first volley :-) */
	kill(th->sync_process, SYNCSIGNAL);

	return EOK;
}


int trueemuClose(ggi_visual *vis)
{
	TrueemuHook *th = VIS_PRIVATE(vis);

	signal(SYNCSIGNAL, SIG_IGN);
	_sync_vis = NULL;


	if (th->colormap != NULL) {
		free(th->colormap);
	}

	if (th->src_buf != NULL) {
		free(th->src_buf);
	}

	if (th->dest_buf != NULL) {
		free(th->dest_buf);
	}

	if (th->R != NULL) free(th->R);
	if (th->G != NULL) free(th->G);
	if (th->B != NULL) free(th->B);
	if (th->T != NULL) free(th->T);


	return EOK;
}


int trueemuNewMode(ggi_visual *vis)
{
	TrueemuHook *th = VIS_PRIVATE(vis);

	th->mode = calc_default_mode(th->mode, th->parent_gt, th->width);

	/* For palettized modes, we first blank the parent visual to
	 * prevent a nasty colored image during the transition.  It is
	 * a happy coincidence that black is always at color number 0.
	 */

	if ((th->parent_gt == GT_8BIT) || (th->parent_gt == GT_4BIT)) {

		ggiSetGCForeground(th->parent, 0);
		ggiFillscreen(th->parent);
		ggiFlush(th->parent);
	}

	setup_dithering(th);
	setup_palette(th);
	
	UPDATE_MOD(th, 0, 0, th->width-1, th->height-1);

	return trueemuFlush(vis);
}
