/* ------------------------------------------------------------------------
 *	TrueEmu - 24 bit emulation target for GGI
 * ------------------------------------------------------------------------
 *
 * 	Copyright (c) 1998  Andrew Apted  <andrew@ggi-project.org>
 *
 *	This library is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU Library General Public
 *	License as published by the Free Software Foundation; either
 *	version 2, or (at your option) any later version.
 * 
 *	This library is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Library General Public License for more details.
 *
 *	You should have received a copy of the GNU Library General
 *	Public License along with this library; see the file
 *	COPYING.LGPL.  If not, write to the Free Software Foundation,
 *	675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ------------------------------------------------------------------------
 */

#ifndef _LIBTRUEEMU_H
#define _LIBTRUEEMU_H

#include <unistd.h>

#include <ggi/ggi.h>
#include <ggi/libggi.h>


/**************************************************
 **
 **  Trueemu private defines
 **
 **************************************************/


#define TrueColor	sint32

#define TC_RED(tc)	(((tc) >> 16) & 0xff)
#define TC_GREEN(tc)	(((tc) >>  8) & 0xff)
#define TC_BLUE(tc)	(((tc)      ) & 0xff)

#define T2GGI(tc)  { TC_RED(tc) << 8, TC_GREEN(tc) << 8, TC_BLUE(tc) << 8 }


#define BLITFUNC(def)	void def(struct trueemu_hook *th,  \
				void *dest_raw, uint8 *src, int width)

#ifndef NULL
#define NULL  ((void *) 0L)
#endif

#ifndef MIN
#define MIN(a, b)  ((a) < (b) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a, b)  ((a) > (b) ? (a) : (b))
#endif

#ifndef ABS
#define ABS(n)  (((n) < 0) ? -(n) : (n))
#endif


/* Info for SYNC helper process */

#define SYNCFPS		10	/* must be > 1 */
#define SYNCSIGNAL	SIGPROF


#define UPDATE_MOD(th,x1,y1,x2,y2)  \
	if ((x1) < th->modified_min.x) th->modified_min.x = (x1);  \
	if ((y1) < th->modified_min.y) th->modified_min.y = (y1);  \
	if ((x2) > th->modified_max.x) th->modified_max.x = (x2);  \
	if ((y2) > th->modified_max.y) th->modified_max.y = (y2);


/**************************************************** 
 **
 **  TrueEmu private data
 **
 ****************************************************/


struct trueemu_hook;  /* declare forward */

typedef struct trueemu_blits
{
	BLITFUNC( (* blitter_b32_d0) );

	BLITFUNC( (* blitter_b24_d0) );
	
	BLITFUNC( (* blitter_b16_d0) );
	BLITFUNC( (* blitter_b16_d2_ev) );
	BLITFUNC( (* blitter_b16_d2_od) );
	BLITFUNC( (* blitter_b16_d4_ev) );
	BLITFUNC( (* blitter_b16_d4_od) );
	
	BLITFUNC( (* blitter_b8_d0) );
	BLITFUNC( (* blitter_b8_d2_ev) );
	BLITFUNC( (* blitter_b8_d2_od) );
	BLITFUNC( (* blitter_b8_d4_ev) );
	BLITFUNC( (* blitter_b8_d4_od) );
	
	BLITFUNC( (* blitter_b4_d0) );
	BLITFUNC( (* blitter_b4_d2_ev) );
	BLITFUNC( (* blitter_b4_d2_od) );
	BLITFUNC( (* blitter_b4_d4_ev) );
	BLITFUNC( (* blitter_b4_d4_od) );

} TrueemuBlits;


typedef struct trueemu_hook
{
	ggi_visual_t parent;

	int flags;

	int parent_gt;
	int child_gt;

	int mode;	/* dither & palette mode */
	int width, height;

	pid_t sync_process;
	
	ggi_color *colormap;

	/* 2D operations on memory buffer */

	struct ggi_visual_op2d *mem_op2d;

	ggi_coord modified_min;
	ggi_coord modified_max;
	
	/* blit information */
	
	BLITFUNC( (* blitter_even) );
	BLITFUNC( (* blitter_odd) );

	void *src_buf;
	void *dest_buf;
	
	/* hicolor dither tables */

	uint16 (* R)[4];	/* actually [256][4] */
	uint16 (* G)[4];
	uint16 (* B)[4];

	/* palette dither table */

	uint8 (* T)[4];		/* actually [32768][4] */
	
} TrueemuHook;


/* modes */

#define TRUEEMU_M_DITHER_0	0x0001
#define TRUEEMU_M_DITHER_2	0x0002
#define TRUEEMU_M_DITHER_4	0x0004

#define TRUEEMU_M_RGB		0x0010	 /* 'r' */
#define TRUEEMU_M_CUBE		0x0020	 /* 'c' */
#define TRUEEMU_M_PASTEL	0x0040	 /* 'p' */
#define TRUEEMU_M_PASTEL_2	0x0080	 /* 'q' */

#define TE_DITHER_MASK  \
	(TRUEEMU_M_DITHER_0 | TRUEEMU_M_DITHER_2 | TRUEEMU_M_DITHER_4)

#define TE_PALETTE_MASK  \
	(TRUEEMU_M_RGB | TRUEEMU_M_CUBE | TRUEEMU_M_PASTEL |  \
	 TRUEEMU_M_PASTEL_2)

/* The meaning of the various modes:
 *
 *	Dithering:
 *
 *	TRUEEMU_M_DITHER_0 : Don't dither if possible.
 *	TRUEEMU_M_DITHER_2 : Use a two pixel dither if possible.
 *	TRUEEMU_M_DITHER_4 : Use a four pixel dither if possible.
 *
 *	two pixel:	four pixel:
 *
 *	--  o-  oo	--  o-  o-  oo  oo
 *	--  -o  oo	--  --  -o  -o  oo
 *
 *	Palette:
 *
 *	TRUEEMU_M_RGB    : Use a whole number of bits for red, green and
 *			   blue.  For an 8 bit mode, this is 3:3:2.  For
 *			   a 4 bit mode, this is 1:2:1.
 *	TRUEEMU_M_CUBE   : Use the same number of reds, greens and
 *			   blues.  For an 8 bit mode, this is 6x6x6.
 *	TRUEEMU_M_PASTEL : Use a shaded pastel scheme.
 */


/* flags */

#define TRUEEMU_F_PASSTHRU  0x0001    /* parent depth == child depth */

#define TRUEEMU_F_CTRL	0x0010	    /* is control key pressed ? */
#define TRUEEMU_F_ALT	0x0020	    /* is alt key pressed ? */


/* special graphtype */

#define TRUEEMU_GT_AUTO  -1


/* blitter tables */

extern TrueemuBlits blit24_table;
extern TrueemuBlits blit32_table;


/****************************************************
 **
 **  TrueEmu private functions
 **
 ****************************************************/
 
  
extern int trueemuOpen(ggi_visual *vis);
extern int trueemuClose(ggi_visual *vis);
extern int trueemuUpdate(ggi_visual*vis, int x, int y, int w, int h);
extern int trueemuFlush(ggi_visual*vis);
extern int trueemuNewMode(ggi_visual *vis);
  

#endif  /* _LIBTRUEEMU_H */
