/* 
 *  Graphics context manipulation
 *
 *  Copyright (C) 1998 Thomas Tanner. See CREDITS for details.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "internal.h"
#include "util.h"

/*----------------------------------------------------------------------*/

/* function table updating */

void	GGI2DregisterUpdateFunc(ggi_visual_t vis, updatefunc f, int mask)
{
	ggi2d_data	*data = GGI2D_DATA(vis);
	
	if (data->modules == MAX_MODULES) {
		DPRINT("Error: too much modules loaded on %p\n",vis);
		return;
	}
	data->module[data->modules].update = f;
	data->module[data->modules].change_mask = mask;
	data->change_mask |= mask;
	data->modules++;
}

void	GGI2DinitGC(ggi_visual_t vis)
{
	const ggi_info *info;
  
	info = ggiGetInfo(vis);
	GGI2D_GC(vis).arcmode = GGI_ARC_SECTOR;
	GGI2D_GC(vis).polymode = GGI_POLY_EVENODD;
	GGI2D_GC_DASHSIZE(vis) = 0;
	GGI2D_GC_APPEND(vis) = GGI_OFF;
	GGI2D_GC_ANTIALIAS(vis) = GGI_OFF;
	GGI2D_GC(vis).drawcol = 0; 
	GGI2D_GC(vis).fillcol = 0; 
	GGI2D_GC_TEXTURED(vis) = GGI_OFF;
	GGI2D_GC_CLIP_X1(vis) = 0; 
	GGI2D_GC_CLIP_Y1(vis) = 0; 
	GGI2D_GC_CLIP_X2(vis) = info->fb.width; 
	GGI2D_GC_CLIP_Y2(vis) = info->fb.height; 
	GGI2D_GC_OP(vis) = GGI_SET; 
	GGI2D_GC_ALPHA(vis) = GGI_MAX_ALPHA; 
	GGI2D_ACCELCHANGES(vis) = GGI2D_UPD_ALL; 
}

void	GGI2Dupdate(ggi_visual_t vis, int change)
{
	int	i;
	ggi2d_data	*data = GGI2D_DATA(vis);
	
	if (change & data->change_mask) {
		for (i = 0;  i < data->modules; i++)
			if (change & data->module[i].change_mask)
				data->module[i].update(vis, change);
	}
	data->gc.accel_changes |= change; 
}

/*----------------------------------------------------------------------*/

/* clipping */

void	ggiSetClip(ggi_visual_t vis, int x1, int y1, int x2, int y2)
{
	int	tmp;
	const ggi_info *info;
  
	_ggiLock(vis->mutex); 
	info = ggiGetInfo(vis);
	if (x1 > x2) xchg(x1, x2, tmp);
	if (y1 > y2) xchg(y1, y2, tmp);
	if (x1 < 0) x1 = 0;
	if (y1 < 0) y1 = 0;
	if (x2 > info->fb.width) x2 = info->fb.width;
	if (y2 > info->fb.height) y2 = info->fb.height;
	GGI2D_GC_CLIP_X1(vis) = x1; 
	GGI2D_GC_CLIP_Y1(vis) = y1; 
	GGI2D_GC_CLIP_X2(vis) = x2; 
	GGI2D_GC_CLIP_Y2(vis) = y2; 
	DPRINT("changing clip rect to ((%d,%d),(%d,%d))\n", x1, y1, x2, y2);
	GGI2Dupdate(vis, GGI2D_UPD_CLIP);    
	_ggiUnlock(vis->mutex); 
}

ggi_bool ggiPointVisible(ggi_visual_t vis, int x, int y)
{
	return CLIP(vis,x,y);
}


ggi_bool ggiRectVisible(ggi_visual_t vis, int x1, int y1, int x2, int y2)
{
	int cx1, cy1, cx2, cy2, tmp;
	
	if (x1 > x2) xchg(x1, x2, tmp);
	if (y1 > y2) xchg(y1, y2, tmp);
	cx1 = GGI2D_GC_CLIP_X1(vis);
	cy1 = GGI2D_GC_CLIP_Y1(vis);
	cx2 = GGI2D_GC_CLIP_X2(vis);
	cy2 = GGI2D_GC_CLIP_Y2(vis);
	return (x1 < cx2 && y1 < cy2 && x2 > cx1 && y2 > cy1);
}

/*----------------------------------------------------------------------*/

/* manipulation of the GC */

void	ggi2dReset(ggi_visual_t vis)
{
	DPRINT("reset of visual %p\n", vis);
	GGI2DinitGC(vis);
	GGI2Dupdate(vis, GGI2D_UPD_ALL);	
}


void	ggiSetArcMode(ggi_visual_t vis, ggi_arcmode mode)
{
	_ggiLock(vis->mutex); 
	if (GGI2D_GC(vis).arcmode != mode) {
		GGI2D_GC(vis).arcmode = mode;
		DPRINT("changing arcmode to %d\n", (int)mode);
		GGI2Dupdate(vis, GGI2D_UPD_ARCMODE);
	}
	_ggiUnlock(vis->mutex); 
}


ggi_arcmode ggiGetArcMode(ggi_visual_t vis)
{
	return GGI2D_GC(vis).arcmode;
}


void	ggiSetPolyMode(ggi_visual_t vis, ggi_polymode mode)
{
	_ggiLock(vis->mutex); 
	if (GGI2D_GC(vis).polymode != mode) {
		GGI2D_GC(vis).polymode = mode;
		DPRINT("changing polymode to %d\n", (int)mode);
		GGI2Dupdate(vis, GGI2D_UPD_POLYMODE);
	}
	_ggiUnlock(vis->mutex); 
}


ggi_polymode ggiGetPolyMode(ggi_visual_t vis)
{
	return GGI2D_GC(vis).polymode;
}


void	ggiSetLineDash(ggi_visual_t vis, uint dash[], uint size)
{
	_ggiLock(vis->mutex); 
	if (GGI2D_GC_DASHSIZE(vis) * size == 0) { 
		GGI2D_GC_DASH(vis) = dash;
		GGI2D_GC_DASHSIZE(vis) = size;
		DPRINT("changing line dash (size %d)\n", size);
		GGI2Dupdate(vis, GGI2D_UPD_LINEMODE);
	} else
	if (GGI2D_GC_DASHSIZE(vis) != size ||
	    memcmp(GGI2D_GC_DASH(vis), dash, size)) {
		GGI2D_GC_DASH(vis) = dash;
		GGI2D_GC_DASHSIZE(vis) = size;
		DPRINT("changing line dash (size %d)\n", size);
		GGI2Dupdate(vis, GGI2D_UPD_LINE);
	}
	_ggiUnlock(vis->mutex); 
}


void	ggiGetLineDash(ggi_visual_t vis, uint dash[], uint *size)
{
	if ((*size = GGI2D_GC_DASHSIZE(vis))) 
		dash = GGI2D_GC_DASH(vis);
	else  
		dash = NULL;
}


void	ggiSetAppendMode(ggi_visual_t vis, ggi_bool append)
{
	_ggiLock(vis->mutex); 
	if ((GGI2D_GC_APPEND(vis)) != (append)) {
		GGI2D_GC_APPEND(vis) = append;
		DPRINT("changing append mode to %d\n", append);
		GGI2Dupdate(vis, GGI2D_UPD_APPENDMODE);
	}
	_ggiUnlock(vis->mutex); 
}


ggi_bool ggiGetAppendMode(ggi_visual_t vis)
{
	return GGI2D_GC_APPEND(vis);
}


void	ggiSetAntialias(ggi_visual_t vis, ggi_bool antialias)
{
	_ggiLock(vis->mutex); 
	if ((GGI2D_GC_ANTIALIAS(vis)) != (antialias)) {
		GGI2D_GC_ANTIALIAS(vis) = antialias;
		DPRINT("changing antialiasing to %d\n", antialias);
		GGI2Dupdate(vis, GGI2D_UPD_ANTIALIAS);
	}
	_ggiUnlock(vis->mutex); 
}


ggi_bool ggiGetAntialias(ggi_visual_t vis)
{
	return GGI2D_GC_ANTIALIAS(vis);
}


void	ggiSetDrawColor(ggi_visual_t vis, ggi_pixel color)
{
	_ggiLock(vis->mutex); 
	if (GGI2D_GC_DRAWCOL(vis) != color) {
		GGI2D_GC_DRAWCOL(vis) = color;
		DPRINT("changing drawcolor to %d\n", color);
		GGI2Dupdate(vis, GGI2D_UPD_DRAWCOL);
	}
	_ggiUnlock(vis->mutex); 
}


ggi_pixel	ggiGetDrawColor(ggi_visual_t vis)
{
	return GGI2D_GC_DRAWCOL(vis);
}


void	ggiSetFillColor(ggi_visual_t vis, ggi_pixel color)
{
	_ggiLock(vis->mutex); 
	if (GGI2D_GC_TEXTURED(vis)) {
	/* FIXME: release texture */
		GGI2D_GC_TEXTURED(vis) = GGI_OFF;
		GGI2D_GC_FILLCOL(vis) = color;
		DPRINT("changing fillcolor to %d\n", color);
		GGI2Dupdate(vis, GGI2D_UPD_TEXTUREMODE);	
	} else
	if (GGI2D_GC_FILLCOL(vis) != color) {
		GGI2D_GC_FILLCOL(vis) = color;
		DPRINT("changing fillcolor to %d\n", color);
		GGI2Dupdate(vis, GGI2D_UPD_FILLCOL);	
	}
	_ggiUnlock(vis->mutex); 
}


ggi_pixel	ggiGetFillColor(ggi_visual_t vis)
{
	return GGI2D_GC_FILLCOL(vis);
}


void	ggiSetFillTexture(ggi_visual_t vis, int refx, int refy, 
			ggi_visual_t texture)
{
	const ggi_info *info;
  
	info = ggiGetInfo(vis);
	if (GGI2D_GC_TEXTURED(vis)) {
	/* FIXME: release old texture */
	}	
	/* FIXME: copy texture */
	GGI2D_GC_TEXTURE(vis) = texture;
	refx %= info->fb.width;
	refy %= info->fb.height;
	GGI2D_GC_TEXREF(vis).x = refx;
	GGI2D_GC_TEXREF(vis).y = refy;
	DPRINT("changing texture to (%d,%d) %p\n", refx, refy, texture);
	if (!GGI2D_GC_TEXTURED(vis)) {
			GGI2D_GC_TEXTURED(vis) = GGI_ON;
		GGI2Dupdate(vis, GGI2D_UPD_TEXTUREMODE);	
	} else
		GGI2Dupdate(vis, GGI2D_UPD_TEXTURE);
	_ggiUnlock(vis->mutex); 
}


ggi_visual_t ggiGetFillTexture(ggi_visual_t vis, int *refx, int *refy)
{
	if (!GGI2D_GC_TEXTURED(vis)) {
		*refx = *refy = 0;
		return NULL;
	} else {
		*refx = GGI2D_GC_TEXREF(vis).x;
		*refy = GGI2D_GC_TEXREF(vis).y;
		return GGI2D_GC_TEXTURE(vis);
	}
}


void	ggiSetOperator(ggi_visual_t vis, ggi_operator operator)
{
	_ggiLock(vis->mutex); 
	if (GGI2D_GC_OP(vis) != operator) {
		GGI2D_GC_OP(vis) = operator;
		DPRINT("changing operator to %d\n", (int)operator);
		GGI2Dupdate(vis, GGI2D_UPD_OPERATOR);	
	}
	_ggiUnlock(vis->mutex); 
}


ggi_operator ggiGetOperator(ggi_visual_t vis)
{
	return GGI2D_GC_OP(vis);
}


void	ggiSetAlpha(ggi_visual_t vis, ggi_alpha alpha)
{
	_ggiLock(vis->mutex); 
	if (GGI2D_GC_ALPHA(vis) != alpha) {
		DPRINT("changing alpha value to %d\n", alpha);
		if (max(alpha,GGI2D_GC_ALPHA(vis)) == GGI_MAX_ALPHA) {
			GGI2D_GC_ALPHA(vis) = alpha;
			GGI2Dupdate(vis, GGI2D_UPD_ALPHAMODE);	
		} else {
			GGI2D_GC_ALPHA(vis) = alpha;
			GGI2Dupdate(vis, GGI2D_UPD_ALPHA);	
		}
	}
	_ggiUnlock(vis->mutex); 
}


ggi_alpha ggiGetAlpha(ggi_visual_t vis)
{
	return GGI2D_GC_ALPHA(vis);
}

